﻿//	Copyright (c) 2012, Michael Kunz. All rights reserved.
//	http://managedcuda.codeplex.com
//
//	This file is part of ManagedCuda.
//
//	ManagedCuda is free software: you can redistribute it and/or modify
//	it under the terms of the GNU Lesser General Public License as 
//	published by the Free Software Foundation, either version 2.1 of the 
//	License, or (at your option) any later version.
//
//	ManagedCuda is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//	GNU Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//	MA 02110-1301  USA, http://www.gnu.org/licenses/.


using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using ManagedCuda.BasicTypes;

namespace ManagedCuda.NPP
{
	/// <summary>
	/// C# Wrapper-Methods for NPP functions defined in npp.h, nppversion.h, nppcore.h, nppi.h, npps.h, nppdefs.h
	/// </summary>
	public static class NPPNativeMethods
	{
		//unfortunately Nvidia provides different dll-names for x86 and x64. Use preprocessor macro to switch names:
#if _x64
		internal const string NPPC_API_DLL_NAME = "nppc64_55";
		internal const string NPPI_API_DLL_NAME = "nppi64_55";
		internal const string NPPS_API_DLL_NAME = "npps64_55";
#else
		internal const string NPPC_API_DLL_NAME = "nppc32_55";
		internal const string NPPI_API_DLL_NAME = "nppi32_55";
		internal const string NPPS_API_DLL_NAME = "npps32_55";
#endif

		/// <summary>
		/// nppcore.h
		/// </summary>
		public static class NPPCore
		{
			[DllImport(NPPC_API_DLL_NAME, EntryPoint = "nppGetLibVersion", CallingConvention = CallingConvention.Cdecl)]
			private static extern IntPtr nppGetLibVersionInternal();

			/// <summary>
			/// Get the NPP library version.
			/// </summary>
			/// <returns>A struct containing separate values for major and minor revision and build number.</returns>
			public static NppLibraryVersion nppGetLibVersion()
			{
				IntPtr ptr = nppGetLibVersionInternal();
				NPP.NppLibraryVersion version;
				version = (NPP.NppLibraryVersion)Marshal.PtrToStructure(ptr, typeof(NPP.NppLibraryVersion));
				return version;
			}

			/// <summary>
			/// What CUDA compute model is supported by the default CUDA device? <para/>
			/// Before trying to call any NPP functions, the user should make a call this function to ensure that the current machine has a CUDA capable device.
			/// </summary>
			/// <returns>An enum value representing if a CUDA capable device was found and what level of compute capabilities it supports.</returns>
			[DllImport(NPPC_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			public static extern GpuComputeCapability nppGetGpuComputeCapability();

			/// <summary>
			/// Get the number of Streaming Multiprocessors (SM) on the default CUDA device.
			/// </summary>
			/// <returns>Number of SMs of the default CUDA device.</returns>
			[DllImport(NPPC_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			public static extern int nppGetGpuNumSMs();

			/// <summary>
			/// Get the maximum number of threads per block on the default CUDA device.
			/// </summary>
			/// <returns>Maximum number of threads per block on the default CUDA device.</returns>
			[DllImport(NPPC_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			public static extern int nppGetMaxThreadsPerBlock();

			/// <summary>
			/// Get the maximum number of threads per SM for the active GPU
			/// </summary>
			/// <returns>Maximum number of threads per SM for the active GPU.</returns>
			[DllImport(NPPC_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			public static extern int nppGetMaxThreadsPerSM();


			[DllImport(NPPC_API_DLL_NAME, EntryPoint = "nppGetGpuName", CallingConvention = CallingConvention.Cdecl)]
			private static extern IntPtr nppGetGpuNameInternal();

			/// <summary>
			/// Get the name of the default CUDA device.
			/// </summary>
			/// <returns>Name string of the graphics-card/compute device in a system.</returns>
			public static string nppGetGpuName()
			{
				IntPtr ptr = nppGetGpuNameInternal();
				string gpuName;
				gpuName = Marshal.PtrToStringAnsi(ptr);
				return gpuName;
			}

			/// <summary>
			/// Get the NPP CUDA stream. <para/>
			/// NPP enables concurrent device tasks via a global stream state varible.
			/// The NPP stream by default is set to stream 0, i.e. non-concurrent mode.
			/// A user can set the NPP stream to any valid CUDA stream. All CUDA commands
			/// issued by NPP (e.g. kernels launched by the NPP library) are then
			/// issed to that NPP stream.
			/// </summary>
			/// <returns></returns>
			[DllImport(NPPC_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			public static extern CUstream nppGetStream();

			/// <summary>
			/// Set the NPP CUDA stream.
			/// </summary>
			[DllImport(NPPC_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			public static extern void nppSetStream(CUstream hStream);
		}

		/// <summary>
		/// nppi.h
		/// </summary>
		public static class NPPi
		{
			/// <summary>
			/// Image-Memory Allocation <para/>
			/// ImageAllocator methods for 2D arrays of data. The allocators have width and height parameters
			/// to specify the size of the image data being allocated. They return a pointer to the
			/// newly created memory and return the numbers of bytes between successive lines. 
			/// <para/>
			/// If the memory allocation failed due to lack of free device memory or device memory fragmentation
			/// the routine returns 0.
			/// <para/>
			/// All allocators return memory with line strides that are 
			/// beneficial for performance. It is not mandatory to use these allocators. Any valid CUDA device-memory
			/// pointers can be used by the NPP primitives and there are no restrictions on line strides.
			/// </summary>
			public static class MemAlloc
			{		
				/// <summary>
				/// 8-bit unsigned image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_8u_C1(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 2 channel 8-bit unsigned image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_8u_C2(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 3 channel 8-bit unsigned image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_8u_C3(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 4 channel 8-bit unsigned image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_8u_C4(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 16-bit unsigned image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16u_C1(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 2 channel 16-bit unsigned image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16u_C2(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 3 channel 16-bit unsigned image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16u_C3(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 4 channel 16-bit unsigned image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16u_C4(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 16-bit signed image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16s_C1(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 2 channel 16-bit signed image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16s_C2(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 4 channel 16-bit signed image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16s_C4(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 1 channel 16-bit signed complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16sc_C1(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 2 channel 16-bit signed complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16sc_C2(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 3 channel 16-bit signed complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16sc_C3(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 4 channel 16-bit signed complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_16sc_C4(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 32-bit signed image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32s_C1(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 3 channel 32-bit signed image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32s_C3(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 4 channel 32-bit signed image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32s_C4(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 32-bit integer complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32sc_C1(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 2 channel 32-bit integer complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32sc_C2(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 3 channel 32-bit integer complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32sc_C3(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 4 channel 32-bit integer complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32sc_C4(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 32-bit floating point image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32f_C1(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 2 channel 32-bit floating point image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32f_C2(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 3 channel 32-bit floating point image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32f_C3(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 4 channel 32-bit floating point image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32f_C4(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 32-bit float complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32fc_C1(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 2 channel 32-bit float complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32fc_C2(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 3 channel 32-bit float complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32fc_C3(int nWidthPixels, int nHeightPixels, ref int pStepBytes);


				/// <summary>
				/// 4 channel 32-bit float complex image memory allocator.
				/// </summary>
				/// <param name="nWidthPixels">Image width.</param>
				/// <param name="nHeightPixels">Image height.</param>
				/// <param name="pStepBytes">Line Step.</param>
				/// <returns>
				/// Pointer to new image data.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppiMalloc_32fc_C4(int nWidthPixels, int nHeightPixels, ref int pStepBytes);

				/// <summary>
				/// Free method for any 2D allocated memory.<para/>
				/// This method should be used to free memory allocated with any of the nppiMalloc_&lt;modifier&gt; methods.
				/// </summary>
				/// <param name="pData">A pointer to memory allocated using nppiMalloc_&lt;modifier&gt;.</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern void nppiFree(CUdeviceptr pData);
			}

			
			/// <summary>
			/// Set methods for images of various types. Images are passed to these primitives via a pointer
			/// to the image data (first pixel in the ROI) and a step-width, i.e. the number of bytes between
			/// successive lines. The size of the area to be set (region-of-interest, ROI) is specified via
			/// a Size struct. <para/>
			/// In addition to the image data and ROI, all methods have a parameter to specify the value being
			/// set. In case of single channel images this is a single value, in case of multi-channel, an 
			/// array of values is passed. 
			/// </summary>
			public static class MemSet
			{
				/// <summary>
				/// 8-bit image set.
				/// </summary>
				/// <param name="nValue">The pixel value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8s_C1R(sbyte nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 8-bit two-channel image set.
				/// </summary>
				/// <param name="aValue">The pixel value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8s_C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 2)] sbyte[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 8-bit three-channel image set.
				/// </summary>
				/// <param name="aValue">The pixel value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8s_C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] sbyte[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 8-bit four-channel image set.
				/// </summary>
				/// <param name="aValue">The pixel value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8s_C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] sbyte[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 8-bit four-channel image set ignoring alpha channel.
				/// </summary>
				/// <param name="aValue">The pixel value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8s_AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] sbyte[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 8-bit unsigned image set.
				/// </summary>
				/// <param name="nValue">The pixel value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_C1R(byte nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 8-bit unsigned image set. 
				/// The 8-bit mask image affects setting of the respective pixels in the destination image.
				/// If the mask value is zero (0) the pixel is not set, if the mask is non-zero, the corresponding
				/// destination pixel is set to specified value.
				/// </summary>
				/// <param name="nValue">The pixel value to be set.</param>
				/// <param name="pDst">Pointer Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_C1MR(byte nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 8-bit unsigned image set.
				/// </summary>
				/// <param name="aValues">Four-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 8-bit unsigned image set.
				/// </summary>
				/// <param name="aValues">Four-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_C4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 8-bit unsigned image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 8-bit unsigned image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_AC4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 8-bit unsigned image set affecting only single channel.
				/// For RGBA images, this method allows setting of a single of the four (RGBA) values 
				/// without changing the contents of the other three channels. The channel is selected
				/// via the pDst pointer. The pointer needs to point to the actual first value to be set,
				/// e.g. in order to set the R-channel (first channel), one would pass pDst unmodified, since
				/// its value actually points to the r channel. If one wanted to modify the B channel (second
				/// channel), one would pass pDst + 2 to the function.
				/// </summary>
				/// <param name="nValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_C4CR(byte nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit unsigned image set.
				/// </summary>
				/// <param name="nValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_C1R(ushort nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 16-bit unsigned image set.
				/// </summary>
				/// <param name="nValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_C1MR(ushort nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 2 channel 16-bit unsigned image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 2)] ushort[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit unsigned image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 16-bit unsigned image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_C4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 16-bit unsigned image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 16-bit unsigned image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_AC4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 16-bit unsigned image set affecting only single channel.
				/// For RGBA images, this method allows setting of a single of the four (RGBA) values 
				/// without changing the contents of the other three channels. The channel is selected
				/// via the pDst pointer. The pointer needs to point to the actual first value to be set,
				/// e.g. in order to set the R-channel (first channel), one would pass pDst unmodified, since
				/// its value actually points to the r channel. If one wanted to modify the B channel (second
				/// channel), one would pass pDst + 2 to the function.
				/// </summary>
				/// <param name="nValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_C4CR(ushort nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit image set.
				/// </summary>
				/// <param name="nValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_C1R(short nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 16-bit image set.
				/// </summary>
				/// <param name="nValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_C1MR(short nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 2 channel 16-bit image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 2)] short[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 16-bit image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_C4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 16-bit image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 16-bit image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_AC4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 16-bit unsigned image set affecting only single channel.
				/// For RGBA images, this method allows setting of a single of the four (RGBA) values 
				/// without changing the contents of the other three channels. The channel is selected
				/// via the pDst pointer. The pointer needs to point to the actual first value to be set,
				/// e.g. in order to set the R-channel (first channel), one would pass pDst unmodified, since
				/// its value actually points to the r channel. If one wanted to modify the B channel (second
				/// channel), one would pass pDst + 2 to the function.
				/// </summary>
				/// <param name="nValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_C4CR(short nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit complex integer image set.
				/// </summary>
				/// <param name="oValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16sc_C1R(Npp16sc oValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit complex integer two-channel image set.
				/// </summary>
				/// <param name="aValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16sc_C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 2)] Npp16sc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit complex integer three-channel image set.
				/// </summary>
				/// <param name="aValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16sc_C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit complex integer four-channel image set ignoring alpha.
				/// </summary>
				/// <param name="aValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16sc_AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit complex integer four-channel image set.
				/// </summary>
				/// <param name="aValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16sc_C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit image set.
				/// </summary>
				/// <param name="nValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_C1R(int nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 32-bit image set.
				/// </summary>
				/// <param name="nValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_C1MR(int nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 32-bit image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 32-bit image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_C4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 16-bit image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 16-bit image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_AC4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 32-bit unsigned image set affecting only single channel.
				/// For RGBA images, this method allows setting of a single of the four (RGBA) values 
				/// without changing the contents of the other three channels. The channel is selected
				/// via the pDst pointer. The pointer needs to point to the actual first value to be set,
				/// e.g. in order to set the R-channel (first channel), one would pass pDst unmodified, since
				/// its value actually points to the r channel. If one wanted to modify the B channel (second
				/// channel), one would pass pDst + 2 to the function.
				/// </summary>
				/// <param name="nValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_C4CR(int nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit complex integer image set.
				/// </summary>
				/// <param name="oValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32sc_C1R(Npp32sc oValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Two channel 32-bit complex integer image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32sc_C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 2)] Npp32sc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit complex integer image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32sc_C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit complex integer image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32sc_C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit complex integer four-channel image set ignoring alpha.
				/// </summary>
				/// <param name="aValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32sc_AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit floating point image set.
				/// </summary>
				/// <param name="nValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_C1R(float nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 32-bit floating point image set.
				/// </summary>
				/// <param name="nValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_C1MR(float nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 32-bit floating point image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 32-bit floating point image set.
				/// </summary>
				/// <param name="aValues">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_C4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 32-bit floating point image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 4 channel 32-bit floating point image set method, not affecting Alpha channel.
				/// For RGBA images, this method allows setting of the RGB values without changing the contents
				/// of the alpha-channel (fourth channel).
				/// </summary>
				/// <param name="aValues">Three-channel array containing the pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the mask image. This is a single channel 8-bit unsigned int image.</param>
				/// <param name="nMaskStep">Number of bytes between line starts of successive lines in the mask image.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_AC4MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aValues, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 4 channel 32-bit floating point image set affecting only single channel.
				/// For RGBA images, this method allows setting of a single of the four (RGBA) values 
				/// without changing the contents of the other three channels. The channel is selected
				/// via the pDst pointer. The pointer needs to point to the actual first value to be set,
				/// e.g. in order to set the R-channel (first channel), one would pass pDst unmodified, since
				/// its value actually points to the r channel. If one wanted to modify the B channel (second
				/// channel), one would pass pDst + 2 to the function.
				/// </summary>
				/// <param name="nValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_C4CR(float nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit complex image set.
				/// </summary>
				/// <param name="oValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32fc_C1R(Npp32fc oValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Two channel 32-bit complex image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32fc_C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 2)] Npp32fc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit complex image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32fc_C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit complex image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32fc_C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit complex four-channel image set ignoring alpha.
				/// </summary>
				/// <param name="aValue">New pixel value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32fc_AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);



				#region new in Cuda 5.5
				/// <summary>
				/// 3 channel 8-bit unsigned image set.
				/// </summary>
				/// <param name="aValue">The pixel value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 16-bit unsigned image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 16-bit image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 16-bit unsigned image set affecting only single channel.
				/// For RGB images, this method allows setting of a single of the three (RGB) values 
				/// without changing the contents of the other three channels. The channel is selected
				/// via the pDst pointer. The pointer needs to point to the actual first value to be set,
				/// e.g. in order to set the R-channel (first channel), one would pass pDst unmodified, since
				/// its value actually points to the r channel. If one wanted to modify the B channel (second
				/// channel), one would pass pDst + 2 to the function.
				/// </summary>
				/// <param name="nValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_C3CR(short nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 32-bit image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 32-bit signed image set affecting only single channel.
				/// For RGB images, this method allows setting of a single of the three (RGB) values 
				/// without changing the contents of the other three channels. The channel is selected
				/// via the pDst pointer. The pointer needs to point to the actual first value to be set,
				/// e.g. in order to set the R-channel (first channel), one would pass pDst unmodified, since
				/// its value actually points to the r channel. If one wanted to modify the B channel (second
				/// channel), one would pass pDst + 2 to the function.
				/// </summary>
				/// <param name="nValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_C3CR(int nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 32-bit floating point image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 32-bit floating point image set affecting only single channel.
				/// For RGB images, this method allows setting of a single of the three (RGB) values 
				/// without changing the contents of the other three channels. The channel is selected
				/// via the pDst pointer. The pointer needs to point to the actual first value to be set,
				/// e.g. in order to set the R-channel (first channel), one would pass pDst unmodified, since
				/// its value actually points to the r channel. If one wanted to modify the B channel (second
				/// channel), one would pass pDst + 2 to the function.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_C3CR(float aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Masked 3 channel 8-bit unsigned image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_C3MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// Masked 3 channel 16-bit unsigned image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_C3MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// Masked 3 channel 16-bit image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16s_C3MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// Masked 3 channel 32-bit image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32s_C3MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// Masked 3 channel 32-bit floating point image set.
				/// </summary>
				/// <param name="aValue">The pixel-value to be set.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_32f_C3MR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// 3 channel 8-bit unsigned image set affecting only single channel.
				/// </summary>
				/// <param name="nValue">The pixel-value to be set.</param>
				/// <param name="pDst">\ref select_destination_pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_8u_C3CR(byte nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 16-bit unsigned image set affecting only single channel.
				/// </summary>
				/// <param name="nValue">The pixel-value to be set.</param>
				/// <param name="pDst">\ref select_destination_pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSet_16u_C3CR(ushort nValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);



				#endregion
			}

			/// <summary>
			/// Copy methods for images of various types. Images are passed to these primitives via a pointer
			/// to the image data (first pixel in the ROI) and a step-width, i.e. the number of bytes between
			/// successive lines. <para/>
			/// The size of the area to be copied (region-of-interest, ROI) is specified via
			/// a Size struct. 
			/// </summary>
			public static class MemCopy
			{
				/// <summary>
				/// 8-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Two-channel 8-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8s_C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 8-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit image copy, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned image copy, not affecting Alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit unsigned image copy, not affecting Alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit image copy, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16sc_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Two-channel 16-bit complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16sc_C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16sc_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16sc_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit complex image copy, ignoring alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16sc_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 32-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 32-bit image copy, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32sc_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Two-channel 32-bit complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32sc_C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32sc_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32sc_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit complex image copy, ignoring alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32sc_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit floating point image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 32-bit floating point image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 32-bit floating point image copy, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit floating-point complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32fc_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Two-channel 32-bit floating-point complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32fc_C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit floating-point complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32fc_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating-point complex image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32fc_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating-point complex image copy, ignoring alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32fc_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// \ref masked_operation 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation three channel 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C3MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 8-bit unsigned image copy, ignoring alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_AC4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation three channel 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C3MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 16-bit unsigned image copy, ignoring alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_AC4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation 16-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation three channel 16-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C3MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 16-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 16-bit signed image copy, ignoring alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_AC4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation 32-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation three channel 32-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C3MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 32-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 32-bit signed image copy, ignoring alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_AC4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation 32-bit float image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation three channel 32-bit float image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C3MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 32-bit float image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// \ref masked_operation four channel 32-bit float image copy, ignoring alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_AC4MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, int nMaskStep);


				/// <summary>
				/// Select-channel 8-bit unsigned image copy for three-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C3CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Select-channel 8-bit unsigned image copy for four-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C4CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Select-channel 16-bit signed image copy for three-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C3CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Select-channel 16-bit signed image copy for four-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C4CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Select-channel 16-bit unsigned image copy for three-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C3CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Select-channel 16-bit unsigned image copy for four-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C4CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Select-channel 32-bit signed image copy for three-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C3CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Select-channel 32-bit signed image copy for four-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C4CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Select-channel 32-bit float image copy for three-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C3CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Select-channel 32-bit float image copy for four-channel images.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Select-Channel Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C4CR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel to single-channel 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel to single-channel 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel to single-channel 16-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel to single-channel 16-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel to single-channel 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel to single-channel 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel to single-channel 32-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel to single-channel 32-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel to single-channel 32-bit float image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel to single-channel 32-bit float image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to three-channel 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to four-channel 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to three-channel 16-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to four-channel 16-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to three-channel 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to four-channel 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to three-channel 32-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to four-channel 32-bit signed image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to three-channel 32-bit float image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel to four-channel 32-bit float image copy.
				/// </summary>
				/// <param name="pSrc">Select-Channel Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 8-bit unsigned packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C4P4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit signed packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit signed packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C4P4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit unsigned packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C4P4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit signed packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit signed packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C4P4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit float packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit float packed to planar image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aDst">Planar Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C4P4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] aDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 8-bit unsigned planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_P4C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit unsigned planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_P4C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit signed planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit signed planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_P4C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit signed planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit signed planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_P4C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit float planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit float planar to packed image copy.
				/// </summary>
				/// <param name="aSrc">Planar Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_P4C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] aSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				#region new in Cuda 5.5
				/// <summary>
				/// Three channel 8-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit unsigned image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit floating point image copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopy_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);



				#endregion
			}
			
			/// <summary>
			/// Convert bit-depth up and down.<para/>
			/// The integer conversion methods do not involve any scaling. Conversions that reduce bit-depth saturate
			/// values exceeding the reduced range to the range's maximum/minimum value.
			/// When converting from floating-point values to integer values, a rounding mode can be specified. After rounding
			/// to integer values the values get saturated to the destination data type's range.
			/// </summary>
			public static class BitDepthConversion
			{
				/// <summary>
				/// 8-bit unsigned to 16-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit unsigned to 8-bit unsigned conversion.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned to 16-bit unsigned  conversion.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned to 16-bit unsigned  conversion.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit unsigned to 8-bit unsigned conversion.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned to 16-bit unsigned conversion, not affecting Alpha.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit unsigned to 8-bit unsigned conversion, not affecting Alpha.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 8-bit unsigned to 16-bit signed conversion.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit signed to 8-bit unsigned conversion.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned to 16-bit signed conversion.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit signed to 8-bit unsignedconversion, not affecting Alpha.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned to 16-bit signed conversion, not affecting Alpha.
				/// For detailed documentation see nppiConverte_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit signed to 8-bit unsigned conversion, not affecting Alpha.
				/// For detailed documentation see nppiConverte_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit singedto 32-bit floating point conversion.
				/// For detailed documentation see nppiConverte_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit floating point to 16-bit conversion.
				/// For detailed documentation see nppiConverte_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// 8-bit unsigned to 32-bit floating point conversion.
				/// For detailed documentation see nppiConverte_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit unsigned to 32-bit floating point conversion.
				/// For detailed documentation see nppiConverte_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit floating point to 16-bit unsigned conversion.
				/// For detailed documentation see nppiConverte_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// 32-bit floating point to 8-bit unsigned conversion.
				/// For detailed documentation see nppiConverte_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// 16-bit unsigned to 32-bit signed conversion.
				/// For detailed documentation see nppiConverte_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 16-bit to 32-bit  conversion.
				/// For detailed documentation see nppiConvert_8u16u_C1R().
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// Three channel 8-bit unsigned to 16-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 32-bit signed conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 32-bit floating-point conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit signed to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit signed to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit signed to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit signed to 32-bit signed conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit signed to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit signed to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit signed to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit signed to 32-bit floating-point conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit unsigned to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit unsigned to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit unsigned to 32-bit signed conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit unsigned to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit unsigned to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit unsigned to 32-bit floating-point conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed to 32-bit signed conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed to 32-bit floating-point conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit signed to 8-bit unsigned conversion with saturation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s8u_C1Rs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit unsigned conversion with saturation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s16u_C1Rs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit signed to 32-bit unsigned conversion with saturation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8s32u_C1Rs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit signed to 16-bit unsigned conversion with saturation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s16u_C1Rs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit signed to 32-bit unsigned conversion with saturation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s32u_C1Rs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit unsigned to 32-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u32u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit signed to 32-bit unsigned conversion with saturation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s32u_C1Rs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit signed to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit unsigned to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32u32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit unsigned to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit signed to 8-bit unsigned conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit signed to 8-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s8s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit signed to 8-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s8s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit signed to 8-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s8s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit signed to 8-bit signed conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s8s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);
				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_8u8s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u8s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16s8s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_16u16s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32u8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32u8s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32u16u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32u16s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32u32s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s16u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Convert
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Round mode</param>
				/// <param name="nScaleFactor">scaling factor</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32s16s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary>
				/// Three channel 32-bit floating point to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Four channel 32-bit floating point to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Four channel 32-bit floating point to 8-bit unsigned conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Single channel 32-bit floating point to 8-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Three channel 32-bit floating point to 8-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Four channel 32-bit floating point to 8-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Four channel 32-bit floating point to 8-bit signed conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Three channel 32-bit floating point to 16-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Four channel 32-bit floating point to 16-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Four channel 32-bit floating point to 16-bit unsigned conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Three channel 32-bit floating point to 16-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Four channel 32-bit floating point to 16-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Four channel 32-bit floating point to 16-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode);


				/// <summary>
				/// Single channel 32-bit floating point to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Single channel 32-bit floating point to 8-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f8s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Single channel 32-bit floating point to 16-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Single channel 32-bit floating point to 16-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f16s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Single channel 32-bit floating point to 32-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f32u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);


				/// <summary>
				/// Single channel 32-bit floating point to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eRoundMode">Flag specifying how fractional float values are rounded to integer values.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiConvert_32f32s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode eRoundMode, int nScaleFactor);




			}
			
			/// <summary>
			/// Methods for copying images and padding borders with a constant, user-specifiable color.
			/// </summary>
			public static class CopyConstBorder
			{ 		
				/// <summary>
				/// 8-bit unsigned image copy width constant border color.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The height of the border at the bottom of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <param name="nValue">The pixel value to be set for border pixels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, byte nValue);


				/// <summary>
				/// 4channel 8-bit unsigned image copy with constant border color.
				/// See nppiCopyConstBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aValue);


				/// <summary>
				/// 4 channel 8-bit unsigned image copy with constant border color.
				/// See nppiCopyConstBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGB values of the border pixels. Because this method does not</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aValue);


				/// <summary>
				/// See nppiCopyConstBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="nValue">Border luminance value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_32s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, int nValue);

				#region new in Cuda 5.5
				/// <summary>
				/// 3 channel 8-bit unsigned integer image copy with constant border color.
				/// See nppiCopyConstBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aValue);


				/// <summary>
				/// 1 channel 16-bit unsigned integer image copy with constant border color.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <param name="nValue">The pixel value to be set for border pixels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, ushort nValue);


				/// <summary>
				/// 3 channel 16-bit unsigned integer image copy with constant border color.
				/// See nppiCopyConstBorder_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aValue);


				/// <summary>
				/// 4 channel 16-bit unsigned integer image copy with constant border color.
				/// See nppiCopyConstBorder_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aValue);


				/// <summary>
				/// 4 channel 16-bit unsigned integer image copy with constant border color with alpha channel unaffected.
				/// See nppiCopyConstBorder_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGB values of the border pixels. Because this method does not</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aValue);


				/// <summary>
				/// 1 channel 16-bit signed integer image copy with constant border color.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <param name="nValue">The pixel value to be set for border pixels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, short nValue);


				/// <summary>
				/// 3 channel 16-bit signed integer image copy with constant border color.
				/// See nppiCopyConstBorder_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aValue);


				/// <summary>
				/// 4 channel 16-bit signed integer image copy with constant border color.
				/// See nppiCopyConstBorder_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aValue);


				/// <summary>
				/// 4 channel 16-bit signed integer image copy with constant border color with alpha channel unaffected.
				/// See nppiCopyConstBorder_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGB values of the border pixels. Because this method does not</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aValue);


				/// <summary>
				/// 3 channel 32-bit signed integer image copy with constant border color.
				/// See nppiCopyConstBorder_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_32s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aValue);


				/// <summary>
				/// 4 channel 32-bit signed integer image copy with constant border color.
				/// See nppiCopyConstBorder_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_32s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aValue);


				/// <summary>
				/// 4 channel 32-bit signed integer image copy with constant border color with alpha channel unaffected.
				/// See nppiCopyConstBorder_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGB values of the border pixels. Because this method does not</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_32s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aValue);


				/// <summary>
				/// 1 channel 32-bit floating point image copy with constant border color.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <param name="nValue">The pixel value to be set for border pixels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, float nValue);


				/// <summary>
				/// 3 channel 32-bit floating point image copy with constant border color.
				/// See nppiCopyConstBorder_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aValue);


				/// <summary>
				/// 4 channel 32-bit floating point image copy with constant border color.
				/// See nppiCopyConstBorder_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGBA values of the border pixels to be set.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aValue);


				/// <summary>
				/// 4 channel 32-bit floating point image copy with constant border color with alpha channel unaffected.
				/// See nppiCopyConstBorder_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <param name="aValue">Vector of the RGB values of the border pixels. Because this method does not</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyConstBorder_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aValue);



				#endregion
			}
			
			/// <summary>
			/// Methods for transposing images of various types. Like matrix transpose, 
			/// image transpose is a mirror along the image's diagonal 
			/// (upper-left to lower-right corner).
			/// </summary>
			public static class Transpose
			{ 
				/// <summary>
				/// 8-bit image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI);

				#region new in Cuda 5.5
				/// <summary>
				/// 3 channel 8-bit unsigned int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 4 channel 8-bit unsigned int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 1 channel 16-bit unsigned int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 3 channel 16-bit unsigned int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 4 channel 16-bit unsigned int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 1 channel 16-bit signed int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 3 channel 16-bit signed int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 4 channel 16-bit signed int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 1 channel 32-bit signed int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 3 channel 32-bit signed int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 4 channel 32-bit signed int image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 1 channel 32-bit floating point image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 3 channel 32-bit floating point image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);


				/// <summary>
				/// 4 channel 32-bit floating point image transpose.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Pointer to the destination ROI.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiTranspose_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSrcROI);



				#endregion
			}

			/// <summary>
			/// Methods for exchanging the color channels of an image. <para/>
			/// The methods support arbitrary permutations of the original channels, 
			/// including replication.
			/// </summary>
			public static class SwapChannel
			{ 
				/// <summary>
				/// 4 channel 8-bit unsigned swap channels, in-place.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_8u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] aDstOrder);

				#region new in Cuda 5.5
				/// <summary>
				/// 3 channel 8-bit unsigned integer source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 8-bit unsigned integer in place image.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 8-bit unsigned integer source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_8u_C4C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 8-bit unsigned integer source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 8-bit unsigned integer source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <param name="nValue">(V) Single channel constant value that can be replicated in one or more of the 4 destination channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_8u_C3C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder, byte nValue);


				/// <summary>
				/// 4 channel 8-bit unsigned integer source image to 4 channel destination image with destination alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 16-bit unsigned integer source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 16-bit unsigned integer in place image.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 16-bit unsigned integer source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16u_C4C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 16-bit unsigned integer source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 16-bit unsigned integer in place image.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 16-bit unsigned integer source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <param name="nValue">(V) Single channel constant value that can be replicated in one or more of the 4 destination channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16u_C3C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder, ushort nValue);


				/// <summary>
				/// 4 channel 16-bit unsigned integer source image to 4 channel destination image with destination alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 16-bit signed integer source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 16-bit signed integer in place image.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 16-bit signed integer source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16s_C4C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 16-bit signed integer source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 16-bit signed integer in place image.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16s_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 16-bit signed integer source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <param name="nValue">(V) Single channel constant value that can be replicated in one or more of the 4 destination channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16s_C3C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder, short nValue);


				/// <summary>
				/// 4 channel 16-bit signed integer source image to 4 channel destination image with destination alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 32-bit signed integer source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 32-bit signed integer in place image.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 32-bit signed integer source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32s_C4C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 32-bit signed integer source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 32-bit signed integer in place image.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32s_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 32-bit signed integer source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <param name="nValue">(V) Single channel constant value that can be replicated in one or more of the 4 destination channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32s_C3C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder, int nValue);


				/// <summary>
				/// 4 channel 32-bit signed integer source image to 4 channel destination image with destination alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 32-bit floating point source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 32-bit floating point in place image.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">oSizeROI Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 32-bit floating point source image to 3 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32f_C4C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 32-bit floating point source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder);


				/// <summary>
				/// 4 channel 32-bit floating point in place image.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32f_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder);


				/// <summary>
				/// 3 channel 32-bit floating point source image to 4 channel destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <param name="nValue">(V) Single channel constant value that can be replicated in one or more of the 4 destination channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32f_C3C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aDstOrder, float nValue);


				/// <summary>
				/// 4 channel 32-bit floating point source image to 4 channel destination image with destination alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDstOrder">Host memory integer array describing how channel values are permutated. The n-th entry</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSwapChannels_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstOrder);



				#endregion
			}
						
			/// <summary>
			/// Adds a constant value to each pixel of an image.
			/// </summary>
			public static class AddConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_8u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, byte nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_8u_C1IRSfs(byte nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_8u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel 8-bit unsigned char in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_8u_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_8u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_8u_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_8u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_8u_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, ushort nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16u_C1IRSfs(ushort nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16u_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16u_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16u_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, short nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16s_C1IRSfs(short nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16s_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16s_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16s_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, Npp16sc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16sc_C1IRSfs(Npp16sc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16sc_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16sc_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_16sc_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, int nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32s_C1IRSfs(int nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32s_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32s_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32s_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, Npp32sc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32sc_C1IRSfs(Npp32sc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32sc_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel with unmodified alpha image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32sc_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32sc_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image add constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, float nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image add constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32f_C1IR(float nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image add constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel in place image add constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32f_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha image add constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha in place image add constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32f_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image add constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image add constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32f_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image add constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32fc_C1R(CUdeviceptr pSrc1, int nSrc1Step, Npp32fc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image add constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32fc_C1IR(Npp32fc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image add constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32fc_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image add constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32fc_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel with unmodified alpha image add constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32fc_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel with unmodified alpha in place image add constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32fc_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image add constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32fc_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image add constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddC_32fc_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Multiplies each pixel of an image by a constant value.
			/// </summary>
			public static class MulConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_8u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, byte nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_8u_C1IRSfs(byte nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_8u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel 8-bit unsigned char in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_8u_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_8u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_8u_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_8u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_8u_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, ushort nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16u_C1IRSfs(ushort nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16u_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16u_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16u_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, short nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16s_C1IRSfs(short nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16s_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16s_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16s_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, Npp16sc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16sc_C1IRSfs(Npp16sc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16sc_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16sc_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_16sc_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, int nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32s_C1IRSfs(int nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32s_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32s_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32s_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, Npp32sc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32sc_C1IRSfs(Npp32sc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32sc_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel with unmodified alpha image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32sc_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image multiply by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32sc_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image multiply by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, float nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image multiply by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32f_C1IR(float nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image multiply by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel in place image multiply by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32f_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha image multiply by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha in place image multiply by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32f_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image multiply by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image multiply by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32f_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image multiply by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32fc_C1R(CUdeviceptr pSrc1, int nSrc1Step, Npp32fc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image multiply by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32fc_C1IR(Npp32fc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image multiply by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32fc_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image multiply by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32fc_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel with unmodified alpha image multiply by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32fc_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel with unmodified alpha in place image multiply by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32fc_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image multiply by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32fc_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image multiply by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulC_32fc_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Multiplies each pixel of an image by a constant value then scales the result by the maximum value for the data bit width.
			/// </summary>
			public static class MulConstScale
			{ 		
				/// <summary>
				/// One 8-bit unsigned char channel image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, byte nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_8u_C1IR(byte nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel 8-bit unsigned char in place image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_8u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image multiply by constant, scale and scale by max bit width value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_8u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_8u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, ushort nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_16u_C1IR(ushort nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_16u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_16u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image multiply by constant and scale by max bit width value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulCScale_16u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Subtracts a constant value from each pixel of an image.
			/// </summary>
			public static class SubConst
			{		
				/// <summary>
				/// One 8-bit unsigned char channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_8u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, byte nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_8u_C1IRSfs(byte nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_8u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel 8-bit unsigned char in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_8u_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_8u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_8u_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_8u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_8u_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, ushort nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16u_C1IRSfs(ushort nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16u_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16u_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16u_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, short nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16s_C1IRSfs(short nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16s_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16s_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16s_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, Npp16sc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16sc_C1IRSfs(Npp16sc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16sc_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16sc_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_16sc_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, int nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32s_C1IRSfs(int nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32s_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32s_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32s_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, Npp32sc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32sc_C1IRSfs(Npp32sc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32sc_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel with unmodified alpha image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32sc_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32sc_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image subtract constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, float nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image subtract constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32f_C1IR(float nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image subtract constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel in place image subtract constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32f_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha image subtract constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha in place image subtract constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32f_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image subtract constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image subtract constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32f_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image subtract constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32fc_C1R(CUdeviceptr pSrc1, int nSrc1Step, Npp32fc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image subtract constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32fc_C1IR(Npp32fc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image subtract constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32fc_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image subtract constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32fc_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel with unmodified alpha image subtract constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32fc_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel with unmodified alpha in place image subtract constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32fc_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image subtract constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32fc_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image subtract constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSubC_32fc_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Divides each pixel of an image by a constant value.
			/// </summary>
			public static class DivConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_8u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, byte nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_8u_C1IRSfs(byte nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_8u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel 8-bit unsigned char in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_8u_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_8u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_8u_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_8u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_8u_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, ushort nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16u_C1IRSfs(ushort nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16u_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16u_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16u_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, short nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16s_C1IRSfs(short nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16s_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16s_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16s_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, Npp16sc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16sc_C1IRSfs(Npp16sc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16sc_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16sc_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_16sc_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp16sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, int nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32s_C1IRSfs(int nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32s_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32s_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32s_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, Npp32sc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32sc_C1IRSfs(Npp32sc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32sc_C3IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel with unmodified alpha image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32sc_AC4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed complex integer (32-bit real, 32-bit imaginary) channel in place image divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32sc_C4IRSfs([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32sc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image divided by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, float nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image divided by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32f_C1IR(float nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image divided by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel in place image divided by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32f_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha image divided by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha in place image divided by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32f_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image divided by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image divided by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32f_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image divided by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32fc_C1R(CUdeviceptr pSrc1, int nSrc1Step, Npp32fc nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image divided by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32fc_C1IR(Npp32fc nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image divided by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32fc_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image divided by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32fc_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel with unmodified alpha image divided by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32fc_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel with unmodified alpha in place image divided by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32fc_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel image divided by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32fc_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit complex floating point (32-bit floating point real, 32-bit floating point imaginary) channel in place image divided by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDivC_32fc_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] Npp32fc[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Determines absolute difference between each pixel of an image and a constant value.
			/// </summary>
			public static class AbsDiffConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image absolute difference with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbsDiffC_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nConstant);


				/// <summary>
				/// One 16-bit unsigned short channel image absolute difference with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbsDiffC_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, ushort nConstant);


				/// <summary>
				/// One 32-bit floating point channel image absolute difference with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbsDiffC_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nConstant);
			}
						
			/// <summary>
			/// Pixel by pixel addition of two images.
			/// </summary>
			public static class Add
			{
				/// <summary>
				/// One 8-bit unsigned char channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_8u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_8u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_8u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_8u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_8u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_8u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_8u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_8u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16s_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16s_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16s_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16sc_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16sc_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16sc_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_16sc_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Note: This function is to be deprecated in future NPP releases, use the function above with a scale factor of 0 instead. 
				/// 32-bit image add.
				/// Add the pixel values of corresponding pixels in the ROI and write them to the output image.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32s_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32s_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32s_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32sc_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32sc_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32sc_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image addition, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32sc_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image addition.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image addition.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image addition.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image addition.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32f_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha image addition.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha in place image addition.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32f_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image addition.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image addition.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32f_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image addition.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32fc_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image addition.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32fc_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image addition.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32fc_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image addition.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32fc_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha image addition.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32fc_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image addition.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32fc_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image addition.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32fc_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image addition.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAdd_32fc_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel addition of squared pixels from source image to floating point pixel values of destination image.
			/// </summary>
			public static class AddSquare
			{		
				/// <summary>
				/// One 8-bit unsigned char channel image squared then added to in place floating point destination image using filter mask (updates destination when mask is non-zero).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddSquare_8u32f_C1IMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel image squared then added to in place floating point destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddSquare_8u32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image squared then added to in place floating point destination image using filter mask (updates destination when mask is non-zero).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddSquare_16u32f_C1IMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image squared then added to in place floating point destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddSquare_16u32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel image squared then added to in place floating point destination image using filter mask (updates destination when mask is non-zero).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddSquare_32f_C1IMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel image squared then added to in place floating point destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddSquare_32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel addition of product of pixels from two source images to floating point pixel values of destination image.
			/// </summary>
			public static class AddProduct
			{ 		
				/// <summary>
				/// One 8-bit unsigned char channel image product added to in place floating point destination image using filter mask (updates destination when mask is non-zero).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddProduct_8u32f_C1IMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel image product added to in place floating point destination image.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddProduct_8u32f_C1IR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image product added to in place floating point destination image using filter mask (updates destination when mask is non-zero).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddProduct_16u32f_C1IMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image product added to in place floating point destination image.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddProduct_16u32f_C1IR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel image product added to in place floating point destination image using filter mask (updates destination when mask is non-zero).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddProduct_32f_C1IMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel image product added to in place floating point destination image.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddProduct_32f_C1IR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel addition of alpha weighted pixel values from a source image to floating point pixel values of destination image.
			/// </summary>
			public static class AddWeighted
			{ 		
				/// <summary>
				/// One 8-bit unsigned char channel alpha weighted image added to in place floating point destination image using filter mask (updates destination when mask is non-zero).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAlpha">Alpha weight to be applied to source image pixels (0.0F to 1.0F)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddWeighted_8u32f_C1IMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nAlpha);


				/// <summary>
				/// One 8-bit unsigned char channel alpha weighted image added to in place floating point destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAlpha">Alpha weight to be applied to source image pixels (0.0F to 1.0F)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddWeighted_8u32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nAlpha);


				/// <summary>
				/// One 16-bit unsigned short channel alpha weighted image added to in place floating point destination image using filter mask (updates destination when mask is non-zero).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAlpha">Alpha weight to be applied to source image pixels (0.0F to 1.0F)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddWeighted_16u32f_C1IMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nAlpha);


				/// <summary>
				/// One 16-bit unsigned short channel alpha weighted image added to in place floating point destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAlpha">Alpha weight to be applied to source image pixels (0.0F to 1.0F)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddWeighted_16u32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nAlpha);


				/// <summary>
				/// One 32-bit floating point channel alpha weighted image added to in place floating point destination image using filter mask (updates destination when mask is non-zero).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">Mask-Image Pointer.</param>
				/// <param name="nMaskStep">Mask-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAlpha">Alpha weight to be applied to source image pixels (0.0F to 1.0F)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddWeighted_32f_C1IMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nAlpha);


				/// <summary>
				/// One 32-bit floating point channel alpha weighted image added to in place floating point destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAlpha">Alpha weight to be applied to source image pixels (0.0F to 1.0F)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAddWeighted_32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nAlpha);
			}
						
			/// <summary>
			/// Pixel by pixel multiply of two images.
			/// </summary>
			public static class Mul
			{
				/// <summary>
				/// One 8-bit unsigned char channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_8u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_8u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_8u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_8u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_8u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_8u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_8u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_8u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16s_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16s_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16s_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16sc_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16sc_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16sc_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_16sc_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Note: This function is to be deprecated in future NPP releases, use the function above with a scale factor of 0 instead.
				/// 1 channel 32-bit image multiplication.
				/// Multiply corresponding pixels in ROI.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32s_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32s_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32s_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32sc_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32sc_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32sc_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image multiplication, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32sc_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image multiplication.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image multiplication.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image multiplication.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image multiplication.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32f_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha image multiplication.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha in place image multiplication.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32f_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image multiplication.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image multiplication.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32f_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image multiplication.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32fc_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image multiplication.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32fc_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image multiplication.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32fc_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image multiplication.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32fc_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha image multiplication.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32fc_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image multiplication.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32fc_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image multiplication.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32fc_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image multiplication.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMul_32fc_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel multiplies each pixel of two images then scales the result by the maximum value for the data bit width.
			/// </summary>
			public static class MulScale
			{
				/// <summary>
				/// One 8-bit unsigned char channel image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_8u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_8u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_8u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_8u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_16u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_16u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_16u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image multiplication then scale by maximum value for pixel bit width.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMulScale_16u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel subtraction of two images.
			/// </summary>
			public static class Sub
			{
				/// <summary>
				/// One 8-bit unsigned char channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_8u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_8u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_8u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_8u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_8u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_8u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_8u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_8u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16s_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16s_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16s_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16sc_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16sc_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16sc_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_16sc_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Note: This function is to be deprecated in future NPP releases, use the function above with a scale factor of 0 instead. 
				/// 32-bit image subtraction.
				/// Subtract pSrc1's pixels from corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32s_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32s_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32s_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32sc_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32sc_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32sc_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32sc_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image subtraction.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image subtraction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image subtraction.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image subtraction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32f_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha image subtraction.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha in place image subtraction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32f_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image subtraction.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image subtraction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32f_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image subtraction.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32fc_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image subtraction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32fc_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image subtraction.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32fc_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image subtraction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32fc_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha image subtraction.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32fc_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image subtraction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32fc_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image subtraction.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32fc_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image subtraction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSub_32fc_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				///// <summary>
				///// Four 32-bit signed integer channel image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				///// </summary>
				///// <param name="pSrc1">Source-Image Pointer.</param>
				///// <param name="nSrc1Step">Source-Image Line Step.</param>
				///// <param name="pSrc2">Source-Image Pointer.</param>
				///// <param name="nSrc2Step">Source-Image Line Step.</param>
				///// <param name="pDst">Destination-Image Pointer.</param>
				///// <param name="nDstStep">Destination-Image Line Step.</param>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiSub_32s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				///// <summary>
				///// Four 32-bit signed integer channel in place image subtraction, scale by 2^(-nScaleFactor), then clamp to saturated value.
				///// </summary>
				///// <param name="pSrc">Source-Image Pointer.</param>
				///// <param name="nSrcStep">Source-Image Line Step.</param>
				///// <param name="pSrcDst">In-Place Image Pointer.</param>
				///// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiSub_32s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);

			}
						
			/// <summary>
			/// Pixel by pixel division of two images.
			/// </summary>
			public static class Div
			{
				/// <summary>
				/// One 8-bit unsigned char channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_8u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_8u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_8u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_8u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_8u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel with unmodified alpha in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_8u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_8u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_8u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel with unmodified alpha in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16s_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16s_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel with unmodified alpha in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16s_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16sc_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16sc_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel with unmodified alpha in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16sc_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short complex number (16-bit real, 16-bit imaginary) channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_16sc_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Note: This function is to be deprecated in future NPP releases, use the function above with a scale factor of 0 instead. 
				/// 32-bit image division.
				/// Divide pixels in pSrc2 by pSrc1's pixels.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32s_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32s_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel with unmodified alpha in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32s_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32sc_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32sc_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32sc_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32sc_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32sc_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32sc_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32sc_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 32-bit signed integer complex number (32-bit real, 32-bit imaginary) channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32sc_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image division.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image division.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image division.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image division.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32f_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha image division.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel with unmodified alpha in place image division.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32f_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image division.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image division.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32f_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image division.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32fc_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image division.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32fc_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image division.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32fc_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image division.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32fc_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha image division.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32fc_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel with unmodified alpha in place image division.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32fc_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel image division.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32fc_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point complex number (32-bit real, 32-bit imaginary) channel in place image division.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_32fc_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);



			}
						
			/// <summary>
			/// Pixel by pixel division of two images using result rounding modes.
			/// </summary>
			public static class DivRound
			{
				/// <summary>
				/// One 8-bit unsigned char channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_8u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_8u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_8u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_8u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image division with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_8u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image division with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_8u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_8u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_8u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16u_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16u_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16u_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16u_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image division with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16u_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image division with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16u_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16u_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16u_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16s_C1RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16s_C1IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16s_C3RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16s_C3IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image division with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16s_AC4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image division with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16s_AC4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16s_C4RSfs(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image division, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rndMode">Result Rounding mode to be used (NPP_RND_ZERO, NPP_RND_NEAR, or NP_RND_FINANCIAL)</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDiv_Round_16s_C4IRSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, NppRoundMode rndMode, int nScaleFactor);
			}
						
			/// <summary>
			/// Absolute value of each pixel value in an image.
			/// </summary>
			public static class Abs
			{
				/// <summary>
				/// One 16-bit signed short channel image absolute value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit signed short channel in place image absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit signed short channel image absolute value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit signed short channel in place image absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit signed short channel image absolute value with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit signed short channel in place image absolute value with unmodified alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit signed short channel image absolute value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit signed short channel in place image absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_16s_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel image absolute value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image absolute value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel in place image absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image absolute value with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image absolute value with unmodified alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image absolute value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbs_32f_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel absolute difference between two images.
			/// </summary>
			public static class AbsDiff
			{
				/// <summary>
				/// One 8-bit unsigned char channel absolute difference of image1 minus image2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbsDiff_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel absolute difference of image1 minus image2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbsDiff_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel absolute difference of image1 minus image2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbsDiff_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// Three 8-bit unsigned char channels absolute difference of image1 minus image2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbsDiff_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channels absolute difference of image1 minus image2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAbsDiff_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}
						
			/// <summary>
			/// Square each pixel in an image.
			/// </summary>
			public static class Sqr
			{
				/// <summary>
				/// One 8-bit unsigned char channel image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_8u_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_8u_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image squared with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image squared with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_8u_AC4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_8u_C4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16u_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16u_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image squared with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image squared with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16u_AC4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16u_C4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16s_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16s_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel in place image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16s_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image squared with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16s_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image squared with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16s_AC4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16s_C4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image squared, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_16s_C4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image squared.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image squared.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image squared.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel in place image squared.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image squared with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image squared with unmodified alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image squared.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image squared.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqr_32f_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel square root of each pixel in an image.
			/// </summary>
			public static class Sqrt
			{
				/// <summary>
				/// One 8-bit unsigned char channel image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_8u_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_8u_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image square root with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image square root with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_8u_AC4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_8u_C4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16u_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16u_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image square root with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image square root with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16u_AC4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16u_C4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16s_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16s_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel in place image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16s_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image square root with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16s_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image square root with unmodified alpha, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16s_AC4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16s_C4RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Four 16-bit signed short channel in place image square root, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_16s_C4IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image square root.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image square root.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image square root.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel in place image square root.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image square root with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image square root with unmodified alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel image square root.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit floating point channel in place image square root.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrt_32f_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel natural logarithm of each pixel in an image.
			/// </summary>
			public static class Ln
			{
				/// <summary>
				/// One 8-bit unsigned char channel image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_8u_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_8u_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_16u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_16u_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_16u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_16u_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_16s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_16s_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_16s_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel in place image natural logarithm, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_16s_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image natural logarithm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image natural logarithm.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image natural logarithm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel in place image natural logarithm.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLn_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Exponential value of each pixel in an image.
			/// </summary>
			public static class Exp
			{ 		
				/// <summary>
				/// One 8-bit unsigned char channel image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 8-bit unsigned char channel in place image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_8u_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_8u_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_16u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit unsigned short channel in place image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_16u_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_16u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_16u_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_16s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 16-bit signed short channel in place image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_16s_C1IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_16s_C3RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// Three 16-bit signed short channel in place image exponential, scale by 2^(-nScaleFactor), then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_16s_C3IRSfs(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, int nScaleFactor);


				/// <summary>
				/// One 32-bit floating point channel image exponential.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit floating point channel in place image exponential.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel image exponential.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit floating point channel in place image exponential.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiExp_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel logical and of an image with a constant.
			/// </summary>
			public static class AndConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image logical and with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, byte nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image logical and with constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_8u_C1IR(byte nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image logical and with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image logical and with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_8u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical and with constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical and with constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_8u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical and with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical and with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_8u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image logical and with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, ushort nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image logical and with constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_16u_C1IR(ushort nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image logical and with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image logical and with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_16u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical and with constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical and with constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_16u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical and with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical and with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_16u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel image logical and with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, int nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image logical and with constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_32s_C1IR(int nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel image logical and with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_32s_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel in place image logical and with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_32s_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical and with constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_32s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical and with constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_32s_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical and with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_32s_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical and with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAndC_32s_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel logical or of an image with a constant.
			/// </summary>
			public static class OrConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image logical or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, byte nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image logical or with constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_8u_C1IR(byte nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image logical or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image logical or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_8u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical or with constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical or with constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_8u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_8u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image logical or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, ushort nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image logical or with constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_16u_C1IR(ushort nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image logical or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image logical or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_16u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical or with constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical or with constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_16u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_16u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel image logical or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, int nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image logical or with constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_32s_C1IR(int nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel image logical or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_32s_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel in place image logical or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_32s_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical or with constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_32s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical or with constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_32s_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_32s_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOrC_32s_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel logical exclusive or of an image with a constant.
			/// </summary>
			public static class XorConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image logical exclusive or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, byte nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image logical exclusive or with constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_8u_C1IR(byte nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image logical exclusive or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image logical exclusive or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_8u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical exclusive or with constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical exclusive or with constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_8u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical exclusive or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical exclusive or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_8u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image logical exclusive or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, ushort nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image logical exclusive or with constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_16u_C1IR(ushort nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image logical exclusive or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image logical exclusive or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_16u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical exclusive or with constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical exclusive or with constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_16u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical exclusive or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical exclusive or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_16u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel image logical exclusive or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, int nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image logical exclusive or with constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_32s_C1IR(int nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel image logical exclusive or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_32s_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel in place image logical exclusive or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_32s_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical exclusive or with constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_32s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical exclusive or with constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_32s_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical exclusive or with constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_32s_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical exclusive or with constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXorC_32s_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel right shift of an image by a constant value.
			/// </summary>
			public static class RightShiftConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, uint nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image right shift by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8u_C1IR(uint nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit signed char channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8s_C1R(CUdeviceptr pSrc1, int nSrc1Step, uint nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit signed char channel in place image right shift by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8s_C1IR(uint nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit signed char channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8s_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit signed char channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8s_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit signed char channel image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit signed char channel in place image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8s_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit signed char channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8s_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit signed char channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_8s_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, uint nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image right shift by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16u_C1IR(uint nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit signed short channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16s_C1R(CUdeviceptr pSrc1, int nSrc1Step, uint nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit signed short channel in place image right shift by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16s_C1IR(uint nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit signed short channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16s_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit signed short channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16s_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit signed short channel image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit signed short channel in place image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16s_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit signed short channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16s_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit signed short channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_16s_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, uint nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image right shift by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_32s_C1IR(uint nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_32s_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_32s_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_32s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image right shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_32s_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image right shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_32s_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image right shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRShiftC_32s_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel left shift of an image by a constant value.
			/// </summary>
			public static class LeftShiftConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image left shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, uint nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image left shift by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_8u_C1IR(uint nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image left shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image left shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_8u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image left shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image left shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_8u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image left shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image left shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_8u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image left shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, uint nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image left shift by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_16u_C1IR(uint nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image left shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image left shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_16u_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image left shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image left shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_16u_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image left shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image left shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_16u_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel image left shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, uint nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image left shift by constant.
				/// </summary>
				/// <param name="nConstant">Constant.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_32s_C1IR(uint nConstant, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel image left shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_32s_C3R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel in place image left shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_32s_C3IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image left shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_32s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image left shift by constant with unmodified alpha.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_32s_AC4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image left shift by constant.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_32s_C4R(CUdeviceptr pSrc1, int nSrc1Step, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image left shift by constant.
				/// </summary>
				/// <param name="aConstants">fixed size array of constant values, one per channel.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLShiftC_32s_C4IR([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] uint[] aConstants, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel logical and of images.
			/// </summary>
			public static class And
			{
				/// <summary>
				/// One 8-bit unsigned char channel image logical and.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image logical and.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_8u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image logical and.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image logical and.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_8u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical and with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical and with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_8u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical and.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical and.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_8u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image logical and.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image logical and.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_16u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image logical and.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image logical and.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_16u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical and with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical and with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_16u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical and.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical and.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_16u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel image logical and.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image logical and.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_32s_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel image logical and.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_32s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel in place image logical and.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_32s_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical and with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_32s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical and with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_32s_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical and.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_32s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical and.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAnd_32s_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel logical or of images.
			/// </summary>
			public static class Or
			{
				/// <summary>
				/// One 8-bit unsigned char channel image logical or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image logical or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_8u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image logical or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image logical or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_8u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_8u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_8u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image logical or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image logical or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_16u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image logical or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image logical or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_16u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_16u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_16u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel image logical or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image logical or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_32s_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel image logical or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_32s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel in place image logical or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_32s_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_32s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_32s_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_32s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiOr_32s_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel logical exclusive or of images.
			/// </summary>
			public static class Xor
			{
				/// <summary>
				/// One 8-bit unsigned char channel image logical exclusive or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image logical exclusive or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_8u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image logical exclusive or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image logical exclusive or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_8u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical exclusive or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical exclusive or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_8u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical exclusive or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical exclusive or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_8u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image logical exclusive or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image logical exclusive or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_16u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image logical exclusive or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image logical exclusive or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_16u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical exclusive or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical exclusive or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_16u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image logical exclusive or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image logical exclusive or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_16u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel image logical exclusive or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 32-bit signed integer channel in place image logical exclusive or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_32s_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel image logical exclusive or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_32s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 32-bit signed integer channel in place image logical exclusive or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_32s_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical exclusive or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_32s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical exclusive or with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_32s_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel image logical exclusive or.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_32s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 32-bit signed integer channel in place image logical exclusive or.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXor_32s_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Pixel by pixel logical not of image.
			/// </summary>
			public static class Not
			{
				/// <summary>
				/// One 8-bit unsigned char channel image logical not.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNot_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image logical not.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNot_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image logical not.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNot_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image logical not.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNot_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical not with unmodified alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNot_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical not with unmodified alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNot_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image logical not.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNot_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image logical not.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNot_8u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Composite two images using constant alpha values.
			/// </summary>
			public static class AlphaCompConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, byte nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, byte nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Three 8-bit unsigned char channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, byte nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, byte nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Four 8-bit unsigned char channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, byte nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, byte nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Four 8-bit unsigned char channel image composition with alpha using constant source alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, byte nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, byte nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 8-bit signed char channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_8s_C1R(CUdeviceptr pSrc1, int nSrc1Step, sbyte nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, sbyte nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 16-bit unsigned short channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, ushort nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, ushort nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Three 16-bit unsigned short channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, ushort nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, ushort nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Four 16-bit unsigned short channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, ushort nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, ushort nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Four 16-bit unsigned short channel image composition with alpha using constant source alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, ushort nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, ushort nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 16-bit signed short channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_16s_C1R(CUdeviceptr pSrc1, int nSrc1Step, short nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, short nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 32-bit unsigned integer channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_32u_C1R(CUdeviceptr pSrc1, int nSrc1Step, uint nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, uint nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 32-bit signed integer channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_32s_C1R(CUdeviceptr pSrc1, int nSrc1Step, int nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, int nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 32-bit floating point channel image composition using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0.0 - 1.0).</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="nAlpha2">Image alpha opacity (0.0 - 1.0).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompC_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, float nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, float nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);
			}
						
			/// <summary>
			/// Premultiplies pixels of an image using a constant alpha value.
			/// </summary>
			public static class AlphaPremulConst
			{
				/// <summary>
				/// One 8-bit unsigned char channel image premultiplication using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, byte nAlpha1, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 8-bit unsigned char channel in place image premultiplication using constant alpha.
				/// </summary>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_8u_C1IR(byte nAlpha1, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel image premultiplication using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, byte nAlpha1, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 8-bit unsigned char channel in place image premultiplication using constant alpha.
				/// </summary>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_8u_C3IR(byte nAlpha1, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image premultiplication using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, byte nAlpha1, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image premultiplication using constant alpha.
				/// </summary>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_8u_C4IR(byte nAlpha1, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel image premultiplication with alpha using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, byte nAlpha1, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image premultiplication with alpha using constant alpha.
				/// </summary>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_8u_AC4IR(byte nAlpha1, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel image premultiplication using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, ushort nAlpha1, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One 16-bit unsigned short channel in place image premultiplication using constant alpha.
				/// </summary>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_16u_C1IR(ushort nAlpha1, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel image premultiplication using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, ushort nAlpha1, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three 16-bit unsigned short channel in place image premultiplication using constant alpha.
				/// </summary>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_16u_C3IR(ushort nAlpha1, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image premultiplication using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, ushort nAlpha1, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image premultiplication using constant alpha.
				/// </summary>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_16u_C4IR(ushort nAlpha1, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image premultiplication with alpha using constant alpha.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, ushort nAlpha1, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image premultiplication with alpha using constant alpha.
				/// </summary>
				/// <param name="nAlpha1">Image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremulC_16u_AC4IR(ushort nAlpha1, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Composite two images using alpha opacity values contained in each image.
			/// </summary>
			public static class AlphaComp
			{
				/// <summary>
				/// One 8-bit unsigned char channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_8u_AC1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Four 8-bit unsigned char channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 8-bit signed char channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_8s_AC1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 16-bit unsigned short channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_16u_AC1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Four 16-bit unsigned short channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 16-bit signed short channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_16s_AC1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 32-bit unsigned integer channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_32u_AC1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Four 32-bit unsigned integer channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_32u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 32-bit signed integer channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_32s_AC1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Four 32-bit signed integer channel image composition using image alpha values (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_32s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// One 32-bit floating point channel image composition using image alpha values (0.0 - 1.0).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_32f_AC1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);


				/// <summary>
				/// Four 32-bit floating point channel image composition using image alpha values (0.0 - 1.0).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nppAlphaOp">Alpha compositing operation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaComp_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiAlphaOp nppAlphaOp);
			}
						
			/// <summary>
			/// Premultiplies image pixels by image alpha opacity values.
			/// </summary>
			public static class AlphaPremul
			{
				/// <summary>
				/// Four 8-bit unsigned char channel image premultiplication with pixel alpha (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremul_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 8-bit unsigned char channel in place image premultiplication with pixel alpha (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremul_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel image premultiplication with pixel alpha (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremul_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four 16-bit unsigned short channel in place image premultiplication with pixel alpha (0 - max channel pixel value).
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaPremul_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Threshold pixels.
			/// </summary>
			public static class Threshold
			{ 		/// <summary>
				/// 32-bit floating point threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nThreshold, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 8-bit unsigned image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aThresholds, NppCmpOp eComparisonOperation);

				/// <summary>
				/// 1 channel 8-bit unsigned char threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nThreshold, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, byte nThreshold, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit unsigned short threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, ushort nThreshold, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, ushort nThreshold, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit signed short threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, short nThreshold, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit signed short in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, short nThreshold, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 32-bit floating point in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nThreshold, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 8-bit unsigned char threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit unsigned short threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit signed short threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit signed short in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 32-bit floating point threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 32-bit floating point in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 8-bit unsigned char in place image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit unsigned short image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit unsigned short in place image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit signed short image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit signed short in place image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 32-bit floating point image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 32-bit floating point in place image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nThreshold);


				/// <summary>
				/// 1 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, byte nThreshold);


				/// <summary>
				/// 1 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, ushort nThreshold);


				/// <summary>
				/// 1 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, ushort nThreshold);


				/// <summary>
				/// 1 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, short nThreshold);


				/// <summary>
				/// 1 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, short nThreshold);


				/// <summary>
				/// 1 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nThreshold);


				/// <summary>
				/// 1 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nThreshold);


				/// <summary>
				/// 3 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds);


				/// <summary>
				/// 3 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds);


				/// <summary>
				/// 3 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds);


				/// <summary>
				/// 3 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds);


				/// <summary>
				/// 3 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds);


				/// <summary>
				/// 3 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds);


				/// <summary>
				/// 3 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds);


				/// <summary>
				/// 3 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds);


				/// <summary>
				/// 4 channel 8-bit unsigned char image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds);


				/// <summary>
				/// 4 channel 8-bit unsigned char in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds);


				/// <summary>
				/// 4 channel 16-bit unsigned short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds);


				/// <summary>
				/// 4 channel 16-bit unsigned short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds);


				/// <summary>
				/// 4 channel 16-bit signed short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds);


				/// <summary>
				/// 4 channel 16-bit signed short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds);


				/// <summary>
				/// 4 channel 32-bit floating point image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds);


				/// <summary>
				/// 4 channel 32-bit floating point in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GT_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds);


				/// <summary>
				/// 1 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nThreshold);


				/// <summary>
				/// 1 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, byte nThreshold);


				/// <summary>
				/// 1 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, ushort nThreshold);


				/// <summary>
				/// 1 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, ushort nThreshold);


				/// <summary>
				/// 1 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, short nThreshold);


				/// <summary>
				/// 1 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, short nThreshold);


				/// <summary>
				/// 1 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nThreshold);


				/// <summary>
				/// 1 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nThreshold);


				/// <summary>
				/// 3 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds);


				/// <summary>
				/// 3 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds);


				/// <summary>
				/// 3 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds);


				/// <summary>
				/// 3 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds);


				/// <summary>
				/// 3 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds);


				/// <summary>
				/// 3 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds);


				/// <summary>
				/// 3 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds);


				/// <summary>
				/// 3 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds);


				/// <summary>
				/// 4 channel 8-bit unsigned char image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds);


				/// <summary>
				/// 4 channel 8-bit unsigned char in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds);


				/// <summary>
				/// 4 channel 16-bit unsigned short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds);


				/// <summary>
				/// 4 channel 16-bit unsigned short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds);


				/// <summary>
				/// 4 channel 16-bit signed short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds);


				/// <summary>
				/// 4 channel 16-bit signed short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds);


				/// <summary>
				/// 4 channel 32-bit floating point image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds);


				/// <summary>
				/// 4 channel 32-bit floating point in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// value is set to nThreshold, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LT_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds);


				/// <summary>
				/// 1 channel 8-bit unsigned char threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nThreshold, byte nValue, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, byte nThreshold, byte nValue, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit unsigned short threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, ushort nThreshold, ushort nValue, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, ushort nThreshold, ushort nValue, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit signed short threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, short nThreshold, short nValue, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit signed short in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, short nThreshold, short nValue, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 32-bit floating point threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nThreshold, float nValue, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 32-bit floating point in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nThreshold, float nValue, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 8-bit unsigned char threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit unsigned short threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit signed short threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit signed short in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 32-bit floating point threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 32-bit floating point in place threshold.
				/// If for a comparison operations OP the predicate (sourcePixel OP nThreshold) is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 8-bit unsigned char image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 8-bit unsigned char in place image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit unsigned short image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit unsigned short in place image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit signed short image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit signed short in place image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 32-bit floating point image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 32-bit floating point in place image threshold, not affecting Alpha.
				/// If for a comparison operations OP the predicate (sourcePixel.channel OP nThreshold) is true, the channel
				/// value is set to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <param name="eComparisonOperation">The type of comparison operation to be used. The only valid are NPP_CMP_LESS and NPP_CMP_GREATER.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_SUPPORTED_MODE_ERROR if an invalid
				/// comparison operation type is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_Val_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nThreshold, byte nValue);


				/// <summary>
				/// 1 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, byte nThreshold, byte nValue);


				/// <summary>
				/// 1 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, ushort nThreshold, ushort nValue);


				/// <summary>
				/// 1 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, ushort nThreshold, ushort nValue);


				/// <summary>
				/// 1 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, short nThreshold, short nValue);


				/// <summary>
				/// 1 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, short nThreshold, short nValue);


				/// <summary>
				/// 1 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nThreshold, float nValue);


				/// <summary>
				/// 1 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is greater than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nThreshold, float nValue);


				/// <summary>
				/// 3 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues);


				/// <summary>
				/// 3 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues);


				/// <summary>
				/// 3 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues);


				/// <summary>
				/// 3 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues);


				/// <summary>
				/// 3 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues);


				/// <summary>
				/// 3 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues);


				/// <summary>
				/// 3 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues);


				/// <summary>
				/// 3 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues);


				/// <summary>
				/// 4 channel 8-bit unsigned char image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues);


				/// <summary>
				/// 4 channel 8-bit unsigned char in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues);


				/// <summary>
				/// 4 channel 16-bit unsigned short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues);


				/// <summary>
				/// 4 channel 16-bit unsigned short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues);


				/// <summary>
				/// 4 channel 16-bit signed short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues);


				/// <summary>
				/// 4 channel 16-bit signed short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues);


				/// <summary>
				/// 4 channel 32-bit floating point image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues);


				/// <summary>
				/// 4 channel 32-bit floating point in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is greater than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_GTVal_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues);


				/// <summary>
				/// 1 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nThreshold, byte nValue);


				/// <summary>
				/// 1 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, byte nThreshold, byte nValue);


				/// <summary>
				/// 1 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, ushort nThreshold, ushort nValue);


				/// <summary>
				/// 1 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, ushort nThreshold, ushort nValue);


				/// <summary>
				/// 1 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, short nThreshold, short nValue);


				/// <summary>
				/// 1 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, short nThreshold, short nValue);


				/// <summary>
				/// 1 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nThreshold, float nValue);


				/// <summary>
				/// 1 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is less than nThreshold is true, the pixel is set
				/// to nValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThreshold">The threshold value.</param>
				/// <param name="nValue">The threshold replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nThreshold, float nValue);


				/// <summary>
				/// 3 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues);


				/// <summary>
				/// 3 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues);


				/// <summary>
				/// 3 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues);


				/// <summary>
				/// 3 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues);


				/// <summary>
				/// 3 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues);


				/// <summary>
				/// 3 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues);


				/// <summary>
				/// 3 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues);


				/// <summary>
				/// 3 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues);


				/// <summary>
				/// 4 channel 8-bit unsigned char image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues);


				/// <summary>
				/// 4 channel 8-bit unsigned char in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValues);


				/// <summary>
				/// 4 channel 16-bit unsigned short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues);


				/// <summary>
				/// 4 channel 16-bit unsigned short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValues);


				/// <summary>
				/// 4 channel 16-bit signed short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues);


				/// <summary>
				/// 4 channel 16-bit signed short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValues);


				/// <summary>
				/// 4 channel 32-bit floating point image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues);


				/// <summary>
				/// 4 channel 32-bit floating point in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThreshold is true, the pixel is set
				/// value is set to rValue, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholds">The threshold values, one per color channel.</param>
				/// <param name="rValues">The threshold replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTVal_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholds, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValues);


				/// <summary>
				/// 1 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is less than nThresholdLT is true, the pixel is set
				/// to nValueLT, else if sourcePixel is greater than nThresholdGT the pixel is set to nValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThresholdLT">The thresholdLT value.</param>
				/// <param name="nValueLT">The thresholdLT replacement value.</param>
				/// <param name="nThresholdGT">The thresholdGT value.</param>
				/// <param name="nValueGT">The thresholdGT replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nThresholdLT, byte nValueLT, byte nThresholdGT, byte nValueGT);


				/// <summary>
				/// 1 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is less than nThresholdLT is true, the pixel is set
				/// to nValueLT, else if sourcePixel is greater than nThresholdGT the pixel is set to nValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThresholdLT">The thresholdLT value.</param>
				/// <param name="nValueLT">The thresholdLT replacement value.</param>
				/// <param name="nThresholdGT">The thresholdGT value.</param>
				/// <param name="nValueGT">The thresholdGT replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, byte nThresholdLT, byte nValueLT, byte nThresholdGT, byte nValueGT);


				/// <summary>
				/// 1 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is less than nThresholdLT is true, the pixel is set
				/// to nValueLT, else if sourcePixel is greater than nThresholdGT the pixel is set to nValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThresholdLT">The thresholdLT value.</param>
				/// <param name="nValueLT">The thresholdLT replacement value.</param>
				/// <param name="nThresholdGT">The thresholdGT value.</param>
				/// <param name="nValueGT">The thresholdGT replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, ushort nThresholdLT, ushort nValueLT, ushort nThresholdGT, ushort nValueGT);


				/// <summary>
				/// 1 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is less than nThresholdLT is true, the pixel is set
				/// to nValueLT, else if sourcePixel is greater than nThresholdGT the pixel is set to nValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThresholdLT">The thresholdLT value.</param>
				/// <param name="nValueLT">The thresholdLT replacement value.</param>
				/// <param name="nThresholdGT">The thresholdGT value.</param>
				/// <param name="nValueGT">The thresholdGT replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, ushort nThresholdLT, ushort nValueLT, ushort nThresholdGT, ushort nValueGT);


				/// <summary>
				/// 1 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is less than nThresholdLT is true, the pixel is set
				/// to nValueLT, else if sourcePixel is greater than nThresholdGT the pixel is set to nValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThresholdLT">The thresholdLT value.</param>
				/// <param name="nValueLT">The thresholdLT replacement value.</param>
				/// <param name="nThresholdGT">The thresholdGT value.</param>
				/// <param name="nValueGT">The thresholdGT replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, short nThresholdLT, short nValueLT, short nThresholdGT, short nValueGT);


				/// <summary>
				/// 1 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is less than nThresholdLT is true, the pixel is set
				/// to nValueLT, else if sourcePixel is greater than nThresholdGT the pixel is set to nValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThresholdLT">The thresholdLT value.</param>
				/// <param name="nValueLT">The thresholdLT replacement value.</param>
				/// <param name="nThresholdGT">The thresholdGT value.</param>
				/// <param name="nValueGT">The thresholdGT replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, short nThresholdLT, short nValueLT, short nThresholdGT, short nValueGT);


				/// <summary>
				/// 1 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is less than nThresholdLT is true, the pixel is set
				/// to nValueLT, else if sourcePixel is greater than nThresholdGT the pixel is set to nValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThresholdLT">The thresholdLT value.</param>
				/// <param name="nValueLT">The thresholdLT replacement value.</param>
				/// <param name="nThresholdGT">The thresholdGT value.</param>
				/// <param name="nValueGT">The thresholdGT replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nThresholdLT, float nValueLT, float nThresholdGT, float nValueGT);


				/// <summary>
				/// 1 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is less than nThresholdLT is true, the pixel is set
				/// to nValueLT, else if sourcePixel is greater than nThresholdGT the pixel is set to nValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nThresholdLT">The thresholdLT value.</param>
				/// <param name="nValueLT">The thresholdLT replacement value.</param>
				/// <param name="nThresholdGT">The thresholdGT value.</param>
				/// <param name="nValueGT">The thresholdGT replacement value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, float nThresholdLT, float nValueLT, float nThresholdGT, float nValueGT);


				/// <summary>
				/// 3 channel 8-bit unsigned char threshold.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValuesGT);


				/// <summary>
				/// 3 channel 8-bit unsigned char in place threshold.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">Destination-Image Pointer.</param>
				/// <param name="nSrcDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValuesGT);


				/// <summary>
				/// 3 channel 16-bit unsigned short threshold.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValuesGT);


				/// <summary>
				/// 3 channel 16-bit unsigned short in place threshold.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValuesGT);


				/// <summary>
				/// 3 channel 16-bit signed short threshold.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValuesGT);


				/// <summary>
				/// 3 channel 16-bit signed short in place threshold.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValuesGT);


				/// <summary>
				/// 3 channel 32-bit floating point threshold.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValuesGT);


				/// <summary>
				/// 3 channel 32-bit floating point in place threshold.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValuesGT);


				/// <summary>
				/// 4 channel 8-bit unsigned char image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// value is set to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValuesGT);


				/// <summary>
				/// 4 channel 8-bit unsigned char in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// value is set to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] rValuesGT);


				/// <summary>
				/// 4 channel 16-bit unsigned short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// value is set to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValuesGT);


				/// <summary>
				/// 4 channel 16-bit unsigned short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// value is set to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] rValuesGT);


				/// <summary>
				/// 4 channel 16-bit signed short image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// value is set to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValuesGT);


				/// <summary>
				/// 4 channel 16-bit signed short in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// value is set to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] rValuesGT);


				/// <summary>
				/// 4 channel 32-bit floating point image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// value is set to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValuesGT);


				/// <summary>
				/// 4 channel 32-bit floating point in place image threshold, not affecting Alpha.
				/// If for a comparison operations sourcePixel is less than rThresholdLT is true, the pixel is set
				/// value is set to rValueLT, else if sourcePixel is greater than rThresholdGT the pixel is set to rValueGT, otherwise it is set to sourcePixel.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="rThresholdsLT">The thresholdLT values, one per color channel.</param>
				/// <param name="rValuesLT">The thresholdLT replacement values, one per color channel.</param>
				/// <param name="rThresholdsGT">The thresholdGT values, one per channel.</param>
				/// <param name="rValuesGT">The thresholdGT replacement values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiThreshold_LTValGTVal_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholdsLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValuesLT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rThresholdsGT, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] rValuesGT);



			}
						
			/// <summary>
			/// Compare the pixels of two images and create a binary result image. In case of multi-channel
			/// image types, the condition must be fulfilled for all channels, otherwise the comparison
			/// is considered false.<para/>
			/// The "binary" result image is of type 8u_C1. False is represented by 0, true by NPP_MAX_8U.
			/// </summary>
			public static class Compare
			{
				/// <summary>
				/// 4 channel 8-bit unsigned image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 8-bit unsigned image compare, not affecting Alpha.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 32-bit floating point image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);
				/// <summary>
				/// 1 channel 8-bit unsigned char image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 8-bit unsigned char image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit unsigned short image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit unsigned short image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit unsigned short image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit unsigned short image compare, not affecting Alpha.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit signed short image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_16s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit signed short image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_16s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit signed short image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_16s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit signed short image compare, not affecting Alpha.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_16s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 32-bit floating point image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 32-bit floating point image compare.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 32-bit signed floating point compare, not affecting Alpha.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompare_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 8-bit unsigned char image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="nConstant">constant value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_8u_C1R(CUdeviceptr pSrc, int nSrcStep, byte nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 8-bit unsigned char image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constant values, one per color channel..</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_8u_C3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 8-bit unsigned char image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_8u_C4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 8-bit unsigned char image compare, not affecting Alpha.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit unsigned short image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="nConstant">constant value</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_16u_C1R(CUdeviceptr pSrc, int nSrcStep, ushort nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit unsigned short image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_16u_C3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit unsigned short image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_16u_C4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit unsigned short image compare, not affecting Alpha.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] ushort[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 16-bit signed short image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="nConstant">constant value.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_16s_C1R(CUdeviceptr pSrc, int nSrcStep, short nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 16-bit signed short image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_16s_C3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit signed short image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_16s_C4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 16-bit signed short image compare, not affecting Alpha.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] short[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 32-bit floating point image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="nConstant">constant value</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_32f_C1R(CUdeviceptr pSrc, int nSrcStep, float nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 3 channel 32-bit floating point image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_32f_C3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 32-bit floating point image compare with constant value.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_32f_C4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 4 channel 32-bit signed floating point compare, not affecting Alpha.
				/// Compare pSrc's pixels with constant value.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eComparisonOperation">Specifies the comparison operation to be used in the pixel comparison.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareC_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppCmpOp eComparisonOperation);


				/// <summary>
				/// 1 channel 32-bit floating point image compare whether two images are equal within epsilon.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2 to determine whether they are equal with a difference of epsilon.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nEpsilon">epsilon tolerance value to compare to pixel absolute differences</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareEqualEps_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nEpsilon);


				/// <summary>
				/// 3 channel 32-bit floating point image compare whether two images are equal within epsilon.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2 to determine whether they are equal with a difference of epsilon.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nEpsilon">epsilon tolerance value to compare to per color channel pixel absolute differences</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareEqualEps_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nEpsilon);


				/// <summary>
				/// 4 channel 32-bit floating point image compare whether two images are equal within epsilon.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2 to determine whether they are equal with a difference of epsilon.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nEpsilon">epsilon tolerance value to compare to per color channel pixel absolute differences</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareEqualEps_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nEpsilon);


				/// <summary>
				/// 4 channel 32-bit signed floating point compare whether two images are equal within epsilon, not affecting Alpha.
				/// Compare pSrc1's pixels with corresponding pixels in pSrc2 to determine whether they are equal with a difference of epsilon.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nEpsilon">epsilon tolerance value to compare to per color channel pixel absolute differences</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareEqualEps_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nEpsilon);


				/// <summary>
				/// 1 channel 32-bit floating point image compare whether image and constant are equal within epsilon.
				/// Compare pSrc's pixels with constant value to determine whether they are equal within a difference of epsilon.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="nConstant">constant value</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nEpsilon">epsilon tolerance value to compare to pixel absolute differences</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareEqualEpsC_32f_C1R(CUdeviceptr pSrc, int nSrcStep, float nConstant, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nEpsilon);


				/// <summary>
				/// 3 channel 32-bit floating point image compare whether image and constant are equal within epsilon.
				/// Compare pSrc's pixels with constant value to determine whether they are equal within a difference of epsilon.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nEpsilon">epsilon tolerance value to compare to per color channel pixel absolute differences</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareEqualEpsC_32f_C3R(CUdeviceptr pSrc, int nSrcStep, float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nEpsilon);


				/// <summary>
				/// 4 channel 32-bit floating point image compare whether image and constant are equal within epsilon.
				/// Compare pSrc's pixels with constant value to determine whether they are equal within a difference of epsilon.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nEpsilon">epsilon tolerance value to compare to per color channel pixel absolute differences</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareEqualEpsC_32f_C4R(CUdeviceptr pSrc, int nSrcStep, float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nEpsilon);


				/// <summary>
				/// 4 channel 32-bit signed floating point compare whether image and constant are equal within epsilon, not affecting Alpha.
				/// Compare pSrc's pixels with constant value to determine whether they are equal within a difference of epsilon.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="aConstants">pointer to a list of constants, one per color channel.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nEpsilon">epsilon tolerance value to compare to per color channel pixel absolute differences</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompareEqualEpsC_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, float[] aConstants, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nEpsilon);



			}
						
			/// <summary>
			/// Computes the mean and standard deviation of image pixel values.
			/// </summary>
			public static class MeanStdDev
			{
				///// <summary>
				///// Device scratch buffer size (in bytes) for mean and standard deviation of image.
				///// This primitive provides the correct buffer size for nppiMean_StdDev_8u_C1R.
				///// </summary>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a host pointer </param>
				///// <returns>
				///// NPP_SUCCESS
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiMeanStdDev8uC1RGetBufferHostSize(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				///// <summary>
				///// 8-bit unsigned mean standard deviation.
				///// </summary>
				///// <param name="pSrc">Source-Image Pointer.</param>
				///// <param name="nSrcStep">Source-Image Line Step.</param>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				///// <param name="pMean">Contains computed mean.</param>
				///// <param name="pStdDev">Contains computed standard deviation.</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiMean_StdDev_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);
			}
						
			/// <summary>
			/// Norm of pixel differences between two images.
			/// </summary>
			public static class NormDiff
			{
				///// <summary>
				///// 8-bit unsigned L1 norm of pixel differences.
				///// </summary>
				///// <param name="pSrc1">Source-Image Pointer.</param>
				///// <param name="nSrcStep1">Source-Image Line Step.</param>
				///// <param name="pSrc2">Source-Image Pointer.</param>
				///// <param name="nSrcStep2">Source-Image Line Step.</param>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="pRetVal">Contains computed L1-norm of differences. This is a host pointer.</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiNormDiff_L1_8u_C1R(CUdeviceptr pSrc1, int nSrcStep1, CUdeviceptr pSrc2, int nSrcStep2, NppiSize oSizeROI, CUdeviceptr pRetVal);


				///// <summary>
				///// 8-bit unsigned L2 norm of pixel differences.
				///// </summary>
				///// <param name="pSrc1">Source-Image Pointer.</param>
				///// <param name="nSrcStep1">Source-Image Line Step.</param>
				///// <param name="pSrc2">Source-Image Pointer.</param>
				///// <param name="nSrcStep2">Source-Image Line Step.</param>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="pRetVal">Contains computed L1-norm of differences. This is a host pointer.</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiNormDiff_L2_8u_C1R(CUdeviceptr pSrc1, int nSrcStep1, CUdeviceptr pSrc2, int nSrcStep2, NppiSize oSizeROI, CUdeviceptr pRetVal);


				///// <summary>
				///// 8-bit unsigned Infinity Norm of pixel differences.
				///// </summary>
				///// <param name="pSrc1">Source-Image Pointer.</param>
				///// <param name="nSrcStep1">Source-Image Line Step.</param>
				///// <param name="pSrc2">Source-Image Pointer.</param>
				///// <param name="nSrcStep2">Source-Image Line Step.</param>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="pRetVal">Contains computed L1-norm of differences. This is a host pointer.</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiNormDiff_Inf_8u_C1R(CUdeviceptr pSrc1, int nSrcStep1, CUdeviceptr pSrc2, int nSrcStep2, NppiSize oSizeROI, CUdeviceptr pRetVal);

				/// <summary>
				/// One-channel 8-bit unsigned image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit signed image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if an 
				/// invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit signed image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormDiff_Inf ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormDiff_Inf ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormDiff_Inf ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormDiff_Inf ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit unsigned images NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_8u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit signed images NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_8s_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 16-bit unsigned images NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 32-bit floating point images NormDiff_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_32f_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit unsigned image NormDiff_Inf affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_8u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit signed image NormDiff_Inf affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_8s_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 16-bit unsigned image NormDiff_Inf affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_16u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 32-bit floating point image NormDiff_Inf affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel 
				/// of interest is specified, or NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_Inf_32f_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size for \ref nppiNormDiff_Inf_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffInfGetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// One-channel 8-bit unsigned image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit signed image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if an 
				/// invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit signed image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormDiff_L1 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormDiff_L1 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormDiff_L1 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormDiff_L1 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit unsigned image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_8u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit signed image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_8s_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 16-bit unsigned image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 32-bit floating point image NormDiff_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_32f_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit unsigned image NormDiff_L1 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_8u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit signed image NormDiff_L1 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_8s_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 16-bit unsigned image NormDiff_L1 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_16u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 32-bit floating point image NormDiff_L1 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel 
				/// of interest is specified, or NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L1_32f_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L1_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL1GetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// One-channel 8-bit unsigned image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit signed image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if an 
				/// invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit signed image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormDiff_L2 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormDiff_L2 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormDiff_L2 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormDiff_L2 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormDiff">Array that contains computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit unsigned image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_8u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit signed image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_8s_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 16-bit unsigned image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 32-bit floating point image NormDiff_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_32f_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit unsigned image NormDiff_L2 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_8u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit signed image NormDiff_L2 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_8s_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 16-bit unsigned image NormDiff_L2 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_16u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 32-bit floating point image NormDiff_L2 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormDiff">Pointer to the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel 
				/// of interest is specified, or NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiff_L2_32f_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormDiff, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormDiff_L2_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormDiffL2GetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);



			}
						
			/// <summary>
			/// 1D mask Linear Convolution Filter, with rescaling, for 8 bit images.
			/// </summary>
			public static class LinearFilter1D
			{
				/// <summary>
				/// 8-bit unsigned 1D (column) image convolution.
				/// <para/>
				/// Apply convolution filter with user specified 1D column of weights.  
				/// Result pixel is equal to the sum of the products between the kernel
				/// coefficients (pKernel array) and corresponding neighboring column pixel
				/// values in the source image defined by nKernelDim and nAnchorY, divided by
				/// nDivisor. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference w.r.t the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 4 channel 8-bit unsigned 1D (column) image convolution.
				/// <para/>
				/// Apply convolution filter with user specified 1D column of weights.  
				/// Result pixel is equal to the sum of the products between the kernel
				/// coefficients (pKernel array) and corresponding neighboring column pixel
				/// values in the source image defined by nKernelDim and nAnchorY, divided by
				/// nDivisor. 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference w.r.t the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 8-bit unsigned 1D (row) image convolution.
				/// <para/>
				/// Apply general linear Row convolution filter, with rescaling, in a 1D mask
				/// region around each source pixel for 1-channel 8 bit/pixel images.  
				/// Result pixel is equal to the sum of the products between the kernel
				/// coefficients (pKernel array) and corresponding neighboring row pixel values
				/// in the source image defined by iKernelDim and iAnchorX, divided by iDivisor. 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference w.r.t the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 4 channel 8-bit unsigned 1D (row) image convolution.
				/// <para/>
				/// Apply general linear Row convolution filter, with rescaling, in a 1D mask
				/// region around each source pixel for 1-channel 8 bit/pixel images.  
				/// Result pixel is equal to the sum of the products between the kernel
				/// coefficients (pKernel array) and corresponding neighboring row pixel values
				/// in the source image defined by iKernelDim and iAnchorX, divided by iDivisor. 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference w.r.t the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);

				#region new in Cuda 5.5
				/// <summary>
				/// 8-bit unsigned three-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 8-bit unsigned four-channel 1D column convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit unsigned single-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit unsigned three-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit unsigned four-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit unsigned four-channel 1D column convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit single-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit three-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit four-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit four-channel 1D column convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 32-bit float single-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 32-bit float three-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 32-bit float four-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 32-bit float four-channel 1D column convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 64-bit float single-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn_64f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned single-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned three-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned four-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned four-channel 1D column convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit unsigned single-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit unsigned three-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit unsigned four-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit unsigned four-channel 1D column convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit single-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit three-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit four-channel 1D column convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit four-channel 1D column convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterColumn32f_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned three-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 8-bit unsigned four-channel 1D row convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit unsigned single-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit unsigned three-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit unsigned four-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit unsigned four-channel 1D row convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit single-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit three-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit four-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 16-bit four-channel 1D row convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor, int nDivisor);


				/// <summary>
				/// 32-bit float single-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 32-bit float three-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 32-bit float four-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 32-bit float four-channel 1D row convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 64-bit float single-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow_64f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned single-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned three-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned four-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned four-channel 1D row convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit unsigned single-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit unsigned three-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit unsigned four-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit unsigned four-channel 1D row convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit single-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit three-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit four-channel 1D row convolution.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);


				/// <summary>
				/// 16-bit four-channel 1D row convolution ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference relative to the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRow32f_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, CUdeviceptr pKernel, int nMaskSize, int nAnchor);



				#endregion
			}
						
			/// <summary>
			/// 1D mask Window Sum for 8 bit images.
			/// </summary>
			public static class WindowSum1D
			{
				/// <summary>
				/// 8-bit unsigned 1D (column) sum to 32f.
				/// Apply Column Window Summation filter over a 1D mask region around each
				/// source pixel for 1-channel 8 bit/pixel input images with 32-bit floating point
				/// output.  
				/// Result 32-bit floating point pixel is equal to the sum of the corresponding and
				/// neighboring column pixel values in a mask region of the source image defined by
				/// nMaskSize and nAnchor. 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">Y offset of the kernel origin frame of reference w.r.t the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumWindowColumn_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, int nMaskSize, int nAnchor);


				/// <summary>
				/// 8-bit unsigned 1D (row) sum to 32f.
				/// Apply Row Window Summation filter over a 1D mask region around each source
				/// pixel for 1-channel 8-bit pixel input images with 32-bit floating point output.  
				/// Result 32-bit floating point pixel is equal to the sum of the corresponding and
				/// neighboring row pixel values in a mask region of the source image defined
				/// by iKernelDim and iAnchorX. 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="nMaskSize">Length of the linear kernel array.</param>
				/// <param name="nAnchor">X offset of the kernel origin frame of reference w.r.t the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumWindowRow_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, int nMaskSize, int nAnchor);
			}
						
			/// <summary>
			/// Image dilate and erod operations.
			/// </summary>
			public static class MorphologyFilter2D
			{
				/// <summary>
				/// 8-bit unsigned image dilation.
				/// <para/>
				/// Dilation computes the output pixel as the maximum pixel value of the pixels
				/// under the mask. Pixels who's corresponding mask values are zero to not 
				/// participate in the maximum search.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// 4 channel 8-bit unsigned image dilation.
				/// <para/>
				/// Dilation computes the output pixel as the maximum pixel value of the pixels
				/// under the mask. Pixels who's corresponding mask values are zero to not 
				/// participate in the maximum search.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// 8-bit unsigned image erosion.
				/// <para/>
				/// Erosion computes the output pixel as the minimum pixel value of the pixels
				/// under the mask. Pixels who's corresponding mask values are zero to not 
				/// participate in the maximum search.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// 4 channel 8-bit unsigned image erosion.
				/// <para/>
				/// Erosion computes the output pixel as the minimum pixel value of the pixels
				/// under the mask. Pixels who's corresponding mask values are zero to not 
				/// participate in the maximum search.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);

				#region new in Cuda 5.5
				/// <summary>
				/// Three-channel 8-bit unsigned integer dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 8-bit unsigned integer dilation, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single-channel 16-bit unsigned integer dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three-channel 16-bit unsigned integer dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 16-bit unsigned integer dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 16-bit unsigned integer dilation, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single-channel 32-bit floating-point dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three-channel 32-bit floating-point dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 32-bit floating-point dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 32-bit floating-point dilation, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three-channel 8-bit unsigned integer erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 8-bit unsigned integer erosion, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single-channel 16-bit unsigned integer erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three-channel 16-bit unsigned integer erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 16-bit unsigned integer erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 16-bit unsigned integer erosion, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single-channel 32-bit floating-point erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three-channel 32-bit floating-point erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 32-bit floating-point erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four-channel 32-bit floating-point erosion, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMask">Pointer to the start address of the mask array</param>
				/// <param name="oMaskSize">Width and Height mask array.</param>
				/// <param name="oAnchor">X and Y offsets of the mask origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pMask, NppiSize oMaskSize, NppiPoint oAnchor);

				/// <summary>
				/// Single-channel 8-bit unsigned integer 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 8-bit unsigned integer 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned integer 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned integer 3x3 dilation, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel 16-bit unsigned integer 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit unsigned integer 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned integer 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned integer 3x3 dilation, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel 32-bit floating-point 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit floating-point 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating-point 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating-point 3x3 dilation, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel 64-bit floating-point 3x3 dilation.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDilate3x3_64f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel 8-bit unsigned integer 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 8-bit unsigned integer 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned integer 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned integer 3x3 erosion, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel 16-bit unsigned integer 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit unsigned integer 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned integer 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned integer 3x3 erosion, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel 32-bit floating-point 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit floating-point 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating-point 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating-point 3x3 erosion, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single-channel 64-bit floating-point 3x3 erosion.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiErode3x3_64f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);




				#endregion
			}
						
			/// <summary>
			/// General purpose 2D convolution filters.
			/// </summary>
			public static class Convolution
			{
				/// <summary>
				/// 8-bit unsigned convolution filter.
				/// <para/>
				/// Pixels under the mask are multiplied by the respective weights in the mask
				/// and the results are summed. Before writing the result pixel the sum is scaled
				/// back via division by nDivisor.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// 4 channel 8-bit unsigned convolution filter.
				/// <para/> 
				/// Pixels under the mask are multiplied by the respective weights in the mask
				/// and the results are summed. Before writing the result pixel the sum is scaled
				/// back via division by nDivisor.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);

				#region new in Cuda 5.5
				/// <summary>
				/// Three channel 8-bit unsigned convolution filter.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Four channel 8-bit unsigned convolution filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Single channel 16-bit unsigned convolution filter.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Three channel 16-bit unsigned convolution filter.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Four channel channel 16-bit unsigned convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Four channel 16-bit unsigned convolution filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Single channel 16-bit convolution filter.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Three channel 16-bit convolution filter.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Four channel channel 16-bit convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Four channel 16-bit convolution filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <param name="nDivisor">The factor by which the convolved summation from the Filter</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor, int nDivisor);


				/// <summary>
				/// Single channel 32-bit float convolution filter.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 32-bit float convolution filter.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit float convolution filter.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit float convolution filter, ignoring alpha channel.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 64-bit float convolution filter.
				/// 
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter_64f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 8-bit unsigned convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 8-bit unsigned convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit unsigned convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit unsigned convolution filter, ignorint alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 8-bit signed convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 8-bit signed convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit signed convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit signed convolution filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 16-bit unsigned convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 16-bit unsigned convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit unsigned convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit unsigned convolution filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 16-bit convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 16-bit convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit convolution filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 32-bit convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 32-bit convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit convolution filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 8-bit unsigned to 16-bit signed convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8u16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit unsigned to 16-bit signed convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8u16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit unsigned to 16-bit signed convolution filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8u16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 8-bit to 16-bit signed convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 8-bit to 16-bit signed convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8s16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit to 16-bit signed convolution filter.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8s16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit to 16-bit signed convolution filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc"> Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pKernel">Pointer to the start address of the kernel coefficient array.</param>
				/// <param name="oKernelSize">Width and Height of the rectangular kernel.</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilter32f_8s16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pKernel, NppiSize oKernelSize, NppiPoint oAnchor);



				#endregion
			}
						
			/// <summary>
			/// 2D linear fixed filters for 8 bit images.
			/// </summary>
			public static class LinearFixedFilters2D
			{
				/// <summary>
				/// 8-bit unsigned box filter.
				/// Computes the average pixel values of the pixels under a rectangular mask.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference w.r.t</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// 4 channel 8-bit unsigned box filter.
				/// Computes the average pixel values of the pixels under a rectangular mask.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference w.r.t</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);

				#region new in Cuda 5.5
				/// <summary>
				/// Three channel 8-bit unsigned box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit unsigned box filter, ignorting alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 16-bit unsigned box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 16-bit unsigned box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit unsigned box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit unsigned box filter, ignorting alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 16-bit box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 16-bit box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit box filter, ignorting alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 32-bit floating-point box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 32-bit floating-point box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit floating-point box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit floating-point box filter, ignorting alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 64-bit floating-point box filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference relative to</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterBox_64f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);



				#endregion
			}
						
			/// <summary>
			/// Min, Median, and Max image filters.
			/// </summary>
			public static class RankFilters
			{
				/// <summary>
				/// 8-bit unsigned maximum filter.
				/// Result pixel value is the maximum of pixel values under the rectangular
				/// mask region.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// 4 channel 8-bit unsigned maximum filter.
				/// Result pixel value is the maximum of pixel values under the rectangular
				/// mask region.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// 8-bit unsigned minimum filter.
				/// Result pixel value is the minimum of pixel values under the rectangular
				/// mask region.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// 4 channel 8-bit unsigned minimum filter.
				/// Result pixel value is the minimum of pixel values under the rectangular
				/// mask region.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);

				#region new in Cuda 5.5
				/// <summary>
				/// Three channel 8-bit unsigned maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit unsigned maximum filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 16-bit unsigned maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 16-bit unsigned maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit unsigned maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit unsigned maximum filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 16-bit signed maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 16-bit signed maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit signed maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit signed maximum filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 32-bit floating-point maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 32-bit floating-point maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit floating-point maximum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit floating-point maximum filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMax_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 8-bit unsigned minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 8-bit unsigned minimum filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 16-bit unsigned minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 16-bit unsigned minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit unsigned minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit unsigned minimum filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 16-bit signed minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 16-bit signed minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit signed minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 16-bit signed minimum filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Single channel 32-bit floating-point minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Three channel 32-bit floating-point minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit floating-point minimum filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);


				/// <summary>
				/// Four channel 32-bit floating-point minimum filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oMaskSize">Width and Height of the neighborhood region for the local</param>
				/// <param name="oAnchor">X and Y offsets of the kernel origin frame of reference</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterMin_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiSize oMaskSize, NppiPoint oAnchor);



				#endregion
			}

						


			/// <summary>
			/// Linear image transforms, like Fourier and DCT transformations.
			/// </summary>
			public static class LinearTransforms
			{
				/// <summary>
				/// 32-bit floating point complex to 32-bit floating point magnitude.
				/// <para/>
				/// Converts complex-number pixel image to single channel image computing
				/// the result pixels as the magnitude of the complex values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMagnitude_32fc32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 32-bit floating point complex to 32-bit floating point squared magnitude.
				/// <para/>
				/// Converts complex-number pixel image to single channel image computing
				/// the result pixels as the squared magnitude of the complex values.
				/// <para/>
				/// The squared magnitude is an itermediate result of magnitude computation and
				/// can thus be computed faster than actual magnitude. If magnitudes are required
				/// for sorting/comparing only, using this function instead of nppiMagnitude_32fc32f_C1R
				/// can be a worthwhile performance optimization.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMagnitudeSqr_32fc32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// Histogram.
			/// </summary>
			public static class Histogram
			{ 
				/// <summary>
				/// Compute levels with even distribution.
				/// </summary>
				/// <param name="hpLevels">A host pointer to array which receives the levels being</param>
				/// <param name="nLevels">The number of levels being computed. nLevels must be at least</param>
				/// <param name="nLowerLevel">Lower boundary value of the lowest level.</param>
				/// <param name="nUpperLevel">Upper boundary value of the greatest level.</param>
				/// <returns>
				/// Error code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiEvenLevelsHost_32s(int[] hpLevels, int nLevels, int nLowerLevel, int nUpperLevel);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nLevels">Number of levels in the histogram.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_8u_C1R(NppiSize oSizeROI, int nLevels , ref int hpBufferSize);


				/// <summary>
				/// 8-bit unsigned histogram with evenly distributed bins.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Pointer to array that receives the computed histogram. </param>
				/// <param name="nLevels">Number of levels.</param>
				/// <param name="nLowerLevel">Lower boundary of lowest level bin.</param>
				/// <param name="nUpperLevel">Upper boundary of highest level bin.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_8u_C1R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pHist, int nLevels, int nLowerLevel, int nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_8u_C3R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 3 channel 8-bit unsigned histogram with evenly distributed bins.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="nLowerLevel">Array containing lower-level of lowest bin per color channel.</param>
				/// <param name="nUpperLevel">Array containing upper-level of highest bin per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_8u_C3R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLowerLevel, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_8u_C4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel 8-bit unsigned histogram with evenly distributed bins.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="nLowerLevel">Array containing lower-level of lowest bin per color channel.</param>
				/// <param name="nUpperLevel">Array containing upper-level of highest bin per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_8u_C4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLowerLevel, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_8u_AC4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel (alpha as the last channel) 8-bit unsigned histogram with evenly distributed bins.
				/// Alpha channel is ignored during histogram computation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="nLowerLevel">Array containing lower-level of lowest bin per color channel.</param>
				/// <param name="nUpperLevel">Array containing upper-level of highest bin per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_8u_AC4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLowerLevel, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nLevels">Number of levels in the histogram.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_16u_C1R(NppiSize oSizeROI, int nLevels , ref int hpBufferSize);


				/// <summary>
				/// 16-bit unsigned histogram with evenly distributed bins.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Pointer to array that receives the computed histogram. </param>
				/// <param name="nLevels">Number of levels.</param>
				/// <param name="nLowerLevel">Lower boundary of lowest level bin.</param>
				/// <param name="nUpperLevel">Upper boundary of highest level bin.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_16u_C1R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pHist, int nLevels, int nLowerLevel, int nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_16u_C3R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 3 channel 16-bit unsigned histogram with evenly distributed bins.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="nLowerLevel">Array containing lower-level of lowest bin per color channel.</param>
				/// <param name="nUpperLevel">Array containing upper-level of highest bin per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_16u_C3R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLowerLevel, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_16u_C4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel 16-bit unsigned histogram with evenly distributed bins.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="nLowerLevel">Array containing lower-level of lowest bin per color channel.</param>
				/// <param name="nUpperLevel">Array containing upper-level of highest bin per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_16u_C4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLowerLevel, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_16u_AC4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel (alpha as the last channel) 16-bit unsigned histogram with evenly distributed bins.
				/// Alpha channel is ignored during histogram computation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="nLowerLevel">Array containing lower-level of lowest bin per color channel.</param>
				/// <param name="nUpperLevel">Array containing upper-level of highest bin per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_16u_AC4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLowerLevel, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nLevels">Number of levels in the histogram.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_16s_C1R(NppiSize oSizeROI, int nLevels , ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed histogram with evenly distributed bins.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Pointer to array that receives the computed histogram. </param>
				/// <param name="nLevels">Number of levels.</param>
				/// <param name="nLowerLevel">Lower boundary of lowest level bin.</param>
				/// <param name="nUpperLevel">Upper boundary of highest level bin.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_16s_C1R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pHist, int nLevels, int nLowerLevel, int nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_16s_C3R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels,ref  int hpBufferSize);


				/// <summary>
				/// 3 channel 16-bit signed histogram with evenly distributed bins.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="nLowerLevel">Array containing lower-level of lowest bin per color channel.</param>
				/// <param name="nUpperLevel">Array containing upper-level of highest bin per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_16s_C3R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLowerLevel, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_16s_C4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels,ref  int hpBufferSize);


				/// <summary>
				/// 4 channel 16-bit signed histogram with evenly distributed bins.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="nLowerLevel">Array containing lower-level of lowest bin per color channel.</param>
				/// <param name="nUpperLevel">Array containing upper-level of highest bin per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_16s_C4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLowerLevel, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramEven_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEvenGetBufferSize_16s_AC4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel (alpha as the last channel) 16-bit signed histogram with evenly distributed bins.
				/// Alpha channel is ignored during histogram computation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="nLowerLevel">Array containing lower-level of lowest bin per color channel.</param>
				/// <param name="nUpperLevel">Array containing upper-level of highest bin per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramEvenGetBufferSize_16s_AC4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramEven_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLowerLevel, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nUpperLevel, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nLevels">Number of levels in the histogram.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_8u_C1R(NppiSize oSizeROI, int nLevels , ref int hpBufferSize);


				/// <summary>
				/// 8-bit unsigned histogram with bins determined by pLevels array.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Pointer to array that receives the computed histogram. </param>
				/// <param name="pLevels">Pointer to array containing the level sizes of the bins.</param>
				/// <param name="nLevels">Number of levels in histogram.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_8u_C1R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pHist, CUdeviceptr pLevels, int nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_8u_C3R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel 8-bit unsigned histogram with bins determined by pLevels.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_8u_C3R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_8u_C4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel 8-bit unsigned histogram with bins determined by pLevels.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_8u_C4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_8u_AC4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel (alpha as a last channel) 8-bit unsigned histogram with bins determined by pLevels.
				/// Alpha channel is ignored during the histograms computations.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_8u_AC4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nLevels">Number of levels in the histogram.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_16u_C1R(NppiSize oSizeROI, int nLevels , ref int hpBufferSize);


				/// <summary>
				/// 16-bit unsigned histogram with bins determined by pLevels array.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Pointer to array that receives the computed histogram. </param>
				/// <param name="pLevels">Pointer to array containing the level sizes of the bins.</param>
				/// <param name="nLevels">Number of levels in histogram.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_16u_C1R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pHist, CUdeviceptr pLevels, int nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_16u_C3R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, ref int hpBufferSize);


				/// <summary>
				/// 4 channel 16-bit unsigned histogram with bins determined by pLevels.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_16u_C3R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_16u_C4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels, ref int hpBufferSize);


				/// <summary>
				/// 4 channel 16-bit unsigned histogram with bins determined by pLevels.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_16u_C4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_16u_AC4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, ref int hpBufferSize);


				/// <summary>
				/// 4 channel (alpha as a last channel) 16-bit unsigned histogram with bins determined by pLevels.
				/// Alpha channel is ignored during the histograms computations.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_16u_AC4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nLevels">Number of levels in the histogram.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_16s_C1R(NppiSize oSizeROI, int nLevels, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed histogram with bins determined by pLevels array.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Pointer to array that receives the computed histogram. </param>
				/// <param name="pLevels">Pointer to array containing the level sizes of the bins.</param>
				/// <param name="nLevels">Number of levels in histogram.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_16_C1R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pHist, CUdeviceptr pLevels, int nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_16s_C3R(NppiSize oSizeROI,  [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel 16-bit signed histogram with bins determined by pLevels.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_16s_C3R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_16s_C4R(NppiSize oSizeROI,  [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels , ref int hpBufferSize);


				/// <summary>
				/// 4 channel 16-bit signed histogram with bins determined by pLevels.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_16s_C4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_16s_AC4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, ref int hpBufferSize);


				/// <summary>
				/// 4 channel (alpha as a last channel) 16-bit signed histogram with bins determined by pLevels.
				/// Alpha channel is ignored during the histograms computations.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_16_AC4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nLevels">Number of levels in the histogram.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_32f_C1R(NppiSize oSizeROI, int nLevels, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float histogram with bins determined by pLevels array.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Pointer to array that receives the computed histogram. </param>
				/// <param name="pLevels">Pointer to array containing the level sizes of the bins.</param>
				/// <param name="nLevels">Number of levels in histogram.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_32f_C1R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pHist, CUdeviceptr pLevels, int nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_32f_C3R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, ref int hpBufferSize);


				/// <summary>
				/// 4 channel 32-bit float histogram with bins determined by pLevels.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_32f_C3R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_32f_C4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels, ref int hpBufferSize);


				/// <summary>
				/// 4 channel 32-bit float histogram with bins determined by pLevels.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_32f_C4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=4)] int[] nLevels, CUdeviceptr pBuffer);


				/// <summary>
				/// Scratch-buffer size for nppiHistogramRange_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">ROI size.</param>
				/// <param name="nLevels">Array containing number of levels per color channel.</param>
				/// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				/// <returns>
				/// Error Code.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRangeGetBufferSize_32f_AC4R(NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, ref int hpBufferSize);


				/// <summary>
				/// 4 channel (alpha as a last channel) 32-bit float histogram with bins determined by pLevels.
				/// Alpha channel is ignored during the histograms computations.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pHist">Array of pointers which are receiving the computed histograms per color channel. </param>
				/// <param name="nLevels">Array containing number of levels per color channel. </param>
				/// <param name="pLevels">Array containing pointers to level-arrays per color channel.</param>
				/// <param name="pBuffer">Pointer to appropriately sized (nppiHistogramRangeGetBufferSize_32f_AC4R) </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHistogramRange_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pHist, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst=3)] int[] nLevels, CUdeviceptr pBuffer);



			}
						
			/// <summary>
			/// Image compression primitives.<para/>
			/// The JPEG standard defines a flow of level shift, DCT and quantization for
			/// forward JPEG transform and inverse level shift, IDCT and de-quantization
			/// for inverse JPEG transform. This group has the functions for both forward
			/// and inverse functions. 
			/// </summary>
			public static class ImageCompression
			{
				/// <summary>
				/// Apply quality factor to raw 8-bit quantization table.
				/// This is effectively and in-place method that modifies a given raw
				/// quantization table based on a quality factor.
				/// Note that this method is a host method and that the pointer to the
				/// raw quantization table is a host pointer.
				/// </summary>
				/// <param name="hpQuantRawTable">Raw quantization table.</param>
				/// <param name="nQualityFactor">Quality factor for the table. Range is [1:100].</param>
				/// <returns>
				/// Error code:
				///      #NPP_NULL_POINTER_ERROR is returned if hpQuantRawTable is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQuantFwdRawTableInit_JPEG_8u(byte[] hpQuantRawTable, int nQualityFactor);


				/// <summary>
				/// Initializes a quantization table for nppiDCTQuantFwd8x8LS_JPEG_8u16s_C1R().
				///    The method creates a 16-bit version of the raw table and converts the 
				/// data order from zigzag layout to original row-order layout since raw
				/// quantization tables are typically stored in zigzag format.
				/// This method is a host method. It consumes and produces host data. I.e. the pointers
				/// passed to this function must be host pointers. The resulting table needs to be
				/// transferred to device memory in order to be used with nppiDCTQuantFwd8x8LS_JPEG_8u16s_C1R()
				/// function.
				/// </summary>
				/// <param name="hpQuantRawTable">Host pointer to raw quantization table as returned by </param>
				/// <param name="hpQuantFwdRawTable">Forward quantization table for use with nppiDCTQuantFwd8x8LS_JPEG_8u16s_C1R().</param>
				/// <returns>
				/// Error code:
				///      #NPP_NULL_POINTER_ERROR pQuantRawTable is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQuantFwdTableInit_JPEG_8u16u(byte[] hpQuantRawTable, ushort[] hpQuantFwdRawTable);


				/// <summary>
				/// Initializes a quantization table for nppiDCTQuantInv8x8LS_JPEG_16s8u_C1R().
				///      The nppiDCTQuantFwd8x8LS_JPEG_8u16s_C1R() method uses a quantization table
				/// in a 16-bit format allowing for faster processing. In addition it converts the 
				/// data order from zigzag layout to original row-order layout. Typically raw
				/// quantization tables are stored in zigzag format.
				/// This method is a host method and consumes and produces host data. I.e. the pointers
				/// passed to this function must be host pointers. The resulting table needs to be
				/// transferred to device memory in order to be used with nppiDCTQuantFwd8x8LS_JPEG_8u16s_C1R()
				/// function.
				/// </summary>
				/// <param name="hpQuantRawTable">Raw quantization table.</param>
				/// <param name="hpQuantFwdRawTable">Inverse quantization table.</param>
				/// <returns>
				/// #NPP_NULL_POINTER_ERROR pQuantRawTable or pQuantFwdRawTable is0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQuantInvTableInit_JPEG_8u16u(byte[] hpQuantRawTable, ushort[] hpQuantFwdRawTable);


				/// <summary>
				/// Forward DCT, quantization and level shift part of the JPEG encoding.
				/// Input is expected in 8x8 macro blocks and output is expected to be in 64x1
				/// macro blocks.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pQuantFwdTable">Forward quantization tables for JPEG encoding created</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// Error codes:
				///         - #NPP_SIZE_ERROR For negative input height/width or not a multiple of
				///           8 width/height.
				///         - #NPP_STEP_ERROR If input image width is not multiple of 8 or does not
				///           match ROI.
				///         - #NPP_NULL_POINTER_ERROR If the destination pointer is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDCTQuantFwd8x8LS_JPEG_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, CUdeviceptr pQuantFwdTable, NppiSize oSizeROI);


				/// <summary>
				/// Inverse DCT, de-quantization and level shift part of the JPEG decoding.
				/// Input is expected in 64x1 macro blocks and output is expected to be in 8x8
				/// macro blocks.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pQuantInvTable">Inverse quantization tables for JPEG decoding created</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// Error codes:
				///         - #NPP_SIZE_ERROR For negative input height/width or not a multiple of
				///           8 width/height.
				///         - #NPP_STEP_ERROR If input image width is not multiple of 8 or does not
				///           match ROI.
				///         - #NPP_NULL_POINTER_ERROR If the destination pointer is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDCTQuantInv8x8LS_JPEG_16s8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, CUdeviceptr pQuantInvTable, NppiSize oSizeROI);
			}
						
			/// <summary>
			/// </summary>
			public static class unsorted
			{
				///// <summary>
				///// SqrIntegral Transforms an image to integral and integral of pixel squares
				///// representation. This function assumes that the integral and integral of squares 
				///// images.
				///// </summary>
				///// <param name="pSrc">Source-Image Pointer.</param>
				///// <param name="nSrcStep">Source-Image Line Step.</param>
				///// <param name="pDst">Destination-Image Pointer.</param>
				///// <param name="nDstStep">Destination-Image Line Step.</param>
				///// <param name="pSqr">Destination-Image Pointer.</param>
				///// <param name="nSqrStep">Destination-Image Line Step.</param>
				///// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				///// <param name="val">The value to add to pDst image pixels</param>
				///// <param name="valSqr">The value to add to pSqr image pixels</param>
				///// <param name="integralImageNewHeight">Extended height of output surfaces (needed by</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiSqrIntegral_8u32s32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, CUdeviceptr pSqr, int nSqrStep, NppiSize oSrcROI, int val, float valSqr, int integralImageNewHeight);


				///// <summary>
				///// RectStdDev Computes the standard deviation of integral images
				///// </summary>
				///// <param name="pSrc">Source-Image Pointer.</param>
				///// <param name="nSrcStep">Source-Image Line Step.</param>
				///// <param name="pSqr">Destination-Image Pointer.</param>
				///// <param name="nSqrStep">Destination-Image Line Step.</param>
				///// <param name="pDst">Destination-Image Pointer.</param>
				///// <param name="nDstStep">Destination-Image Line Step.</param>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="rect">rectangular window</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiRectStdDev_32s32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSqr, int nSqrStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiRect rect);
			}
						
			/// <summary>
			/// Sum of 8 bit images.
			/// </summary>
			public static class Sum
			{
				///// <summary>
				///// Scratch-buffer size for nppiSum_8u_C1R.
				///// </summary>
				///// <param name="oSizeROI">ROI size.</param>
				///// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				///// <returns>
				///// Error Code.
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiReductionGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				///// <summary>
				///// Scratch-buffer size for nppiSum_8u_C4R.
				///// </summary>
				///// <param name="oSizeROI">ROI size.</param>
				///// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				///// <returns>
				///// Error Code.
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiReductionGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 8-bit unsigned image sum with 64-bit long long result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pSum">Contains computed sum.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_8u64s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pSum);


				/// <summary>
				/// 8-bit unsigned image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pSum">Contains computed sum.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pSum);


				/// <summary>
				/// 4 channel 8-bit unsigned image sum with 64-bit long long result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aSum">Array contains computed sum for each channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_8u64s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 4 channel 8-bit unsigned image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aSum">Array contains computed sum for each channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);

				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_8u64s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_8u64s_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_8u64s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_8u64s_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiSum_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSumGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 1-channel 16-bit unsigned short image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="pSum">Pointer to the computed sum.</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pSum);


				/// <summary>
				/// 1-channel 16-bit signed short image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="pSum">Pointer to the computed sum.</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pSum);


				/// <summary>
				/// 1-channel 32-bit floating point image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pSum">Pointer to the computed sum.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pSum);


				/// <summary>
				/// 3-channel 8-bit unsigned char image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="aSum">Array that contains computed sum for each channel.</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 3-channel 16-bit unsigned short image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="aSum">Array that contains computed sum for each channel.</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 3-channel 16-bit signed short image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="aSum">Array that contains computed sum for each channel.</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 3-channel 32-bit floating point image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="aSum">Array that contains computed sum for each channel.</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 4-channel 8-bit unsigned char image sum with 64-bit double precision result.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="aSum">Array that contains computed sum for each channel (alpha channel is not computed).</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 4-channel 16-bit unsigned short image sum with 64-bit double precision result.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="aSum">Array that contains computed sum for each channel (alpha channel is not computed).</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 4-channel 16-bit signed short image sum with 64-bit double precision result.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="aSum">Array that contains computed sum for each channel (alpha channel is not computed).</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 4-channel 32-bit floating point image sum with 64-bit double precision result.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aSum">Array that contains computed sum for each channel (alpha channel is not computed).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 4-channel 16-bit unsigned short image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aSum">Array that contains computed sum for each channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 4-channel 16-bit signed short image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aSum">Array that contains computed sum for each channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);


				/// <summary>
				/// 4-channel 32-bit floating point image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aSum">Array that contains computed sum for each channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSum_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aSum);



			}
						
			/// <summary>
			/// Minimum and maximum of 8-bit images.
			/// </summary>
			public static class MinMax
			{
				///// <summary>
				///// Scratch-buffer size for nppiMinManx_8u_C1R.
				///// </summary>
				///// <param name="oSizeROI">ROI size.</param>
				///// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				///// <returns>
				///// Error Code.
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiMinMaxGetBufferSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				///// <summary>
				///// 8-bit unsigned pixel minimum and maximum.
				///// </summary>
				///// <param name="pSrc">Source-Image Pointer.</param>
				///// <param name="nSrcStep">Source-Image Line Step.</param>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				///// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				///// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferSize_8u_C1R to determine</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiMinMax_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				///// <summary>
				///// Scratch-buffer size for nppiMinManx_8u_C4R.
				///// </summary>
				///// <param name="oSizeROI">ROI size.</param>
				///// <param name="hpBufferSize">Host pointer where required buffer size is returned.</param>
				///// <returns>
				///// Error Code.
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiMinMaxGetBufferSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				///// <summary>
				///// 4 channel 8-bit unsigned pixel minimum and maximum.
				///// </summary>
				///// <param name="pSrc">Source-Image Pointer.</param>
				///// <param name="nSrcStep">Source-Image Line Step.</param>
				///// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				///// <param name="aMin">Device-pointer (array) receiving the minimum result.</param>
				///// <param name="aMax">Device-pointer (array) receiving the maximum result.</param>
				///// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferSize_8u_C4R to determine</param>
				///// <returns>
				///// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///// \note Unlike nppiMinMax_8u_C1R, this primitive returns its results as device
				/////       pointers.
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiMinMax_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);
			}
						
			/// <summary>
			/// Routines manipulating an image's geometry.
			/// </summary>
			public static class GeometricTransforms
			{
				/// <summary>
				/// 8-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image</param>
				/// <param name="xFactor">Factors by which x dimension is changed </param>
				/// <param name="yFactor">Factors by which y dimension is changed </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR indicates an error condition if
				///           srcROIRect has no intersection with the source image.
				///         - #NPP_RESIZE_NO_OPERATION_ERROR if either destination ROI width or
				///           height is less than 1 pixel.
				///         - #NPP_RESIZE_FACTOR_ERROR Indicates an error condition if either xFactor or
				///           yFactor is less than or equal to zero.
				///         - #NPP_INTERPOLATION_ERROR if eInterpolation has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double xFactor, double yFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image</param>
				/// <param name="xFactor">Factors by which x dimension is changed </param>
				/// <param name="yFactor">Factors by which y dimension is changed </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR indicates an error condition if
				///           srcROIRect has no intersection with the source image.
				///         - #NPP_RESIZE_NO_OPERATION_ERROR if either destination ROI width or
				///           height is less than 1 pixel.
				///         - #NPP_RESIZE_FACTOR_ERROR Indicates an error condition if either xFactor or
				///           yFactor is less than or equal to zero.
				///         - #NPP_INTERPOLATION_ERROR if eInterpolation has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double xFactor, double yFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// Compute shape of rotated image.
				/// </summary>
				/// <param name="oSrcROI">Region-of-interest of the source image.</param>
				/// <param name="aQuad">Array of 2D points. These points are the locations of the corners</param>
				/// <param name="nAngle">The rotation nAngle.</param>
				/// <param name="nShiftX">Post-rotation shift in x-direction</param>
				/// <param name="nShiftY">Post-rotation shift in y-direction</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGetRotateQuad(NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] aQuad, double nAngle, double nShiftX, double nShiftY);


				/// <summary>
				/// Compute bounding-box of rotated image.
				/// </summary>
				/// <param name="oSrcROI">Region-of-interest of the source image.</param>
				/// <param name="aBoundingBox">Two 2D points representing the bounding-box of the rotated image. All four points</param>
				/// <param name="nAngle">The rotation angle.</param>
				/// <param name="nShiftX">Post-rotation shift in x-direction.</param>
				/// <param name="nShiftY">Post-rotation shift in y-direction.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGetRotateBound(NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] double[,] aBoundingBox, double nAngle, double nShiftX, double nShiftY);


				/// <summary>
				/// 8-bit unsigned image rotate.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 8-bit unsigned image rotate.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned image rotate.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned image rotate ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 16-bit unsigned image rotate.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit unsigned image rotate.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned image rotate.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned image rotate ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 32-bit float image rotate.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 32-bit float image rotate.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit float image rotate.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit float image rotate ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nAngle">The angle of rotation in degrees.</param>
				/// <param name="nShiftX">Shift along horizontal axis </param>
				/// <param name="nShiftY">Shift along vertical axis </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref rotate_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRotate_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nAngle, double nShiftX, double nShiftY, InterpolationMode eInterpolation);


				/// <summary>
				/// 8-bit unsigned image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 8-bit unsigned image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 8-bit unsigned image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 8-bit unsigned image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 16-bit unsigned image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 16-bit unsigned image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 16-bit unsigned image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 16-bit unsigned image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 32-bit image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 32-bit image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 32-bit image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 32-bit image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 32-bit float image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 32-bit float image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 32-bit float image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 32-bit float image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_MIRROR_FLIP_ERR if flip has an illegal value.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				#region new in Cuda 5.5
				/// <summary>
				/// 3 channel 8-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned image resize not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 8-bit unsigned planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_8u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 16-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed.</param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned image resize not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit unsigned planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_16u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 32-bit floating point image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 32-bit floating point image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit floating point image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit floating point image resize not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 32-bit floating point planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit floating point planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstROISize">Size in pixels of the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResize_32f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize dstROISize, double nXFactor, double nYFactor, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 8-bit unsigned in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 8-bit unsigned in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 8-bit unsigned in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_8u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 8-bit unsigned in place image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 1 channel 16-bit unsigned in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 16-bit unsigned in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 16-bit unsigned in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 16-bit unsigned in place image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 1 channel 16-bit signed image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 1 channel 16-bit signed in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 16-bit signed image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 16-bit signed in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 16-bit signed image mirror.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 16-bit signed in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16s_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 16-bit signed image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Distance in bytes between starts of consecutive lines of the</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 16-bit signed in place image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 1 channel 32-bit signed in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 32-bit signed in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 32-bit signed in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32s_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 32-bit signed in place image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 1 channel 32-bit float in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 3 channel 32-bit float in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 32-bit float in place image mirror.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32f_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);


				/// <summary>
				/// 4 channel 32-bit float in place image mirror not affecting alpha.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="flip">Specifies the axis about which the image is to be mirrored.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref mirror_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMirror_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oROI, NppiAxis flip);



				#endregion
			}
						
			/// <summary>
			/// Affine warping, affine transform calculation
			/// Affine warping of an image is the transform of image pixel positions,
			/// defined by the following formulas:
			/// \f[
			/// X_{new} = C_{00} * x + C_{01} * y + C_{02} \qquad
			/// Y_{new} = C_{10} * x + C_{11} * y + C_{12} \qquad
			/// C = \left[ \matrix{C_{00} <![CDATA[&]]> C_{01} <![CDATA[&]]> C_{02} \cr C_{10} <![CDATA[&]]> C_{11} <![CDATA[&]]> C_{12} } \right]
			/// \f]
			/// That is, any pixel with coordinates 
			/// \f$(X_{new},Y_{new})\f$ in the transformed image is sourced from 
			/// coordinates \f$(x,y)\f$ in the original image. The mapping \f$C\f$ is completely
			/// specified by 6 values
			/// \f$C_{ij}, i=\overline{0,1}, j=\overline{0,2}\f$.
			/// The transform maps parallel lines to parallel
			/// lines and preserves ratios of distances of points to lines.
			/// Implementation specific properties are
			/// discussed in each function's documentation.
			/// </summary>
			public static class AffinTransforms
			{
				/// <summary>
				/// Calculates affine transform coefficients given source rectangular ROI and
				/// its destination quadrangle projection
				/// </summary>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="quad">Destination quadrangle [4,2]</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <returns>
				/// Error codes:
				///         - #NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///           has zero or negative value
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_AFFINE_QUAD_INCORRECT_WARNING Indicates a warning when quad
				///           does not conform to the transform properties. Fourth vertex is
				///           ignored, internally computed coordinates are used instead
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGetAffineTransform(NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] quad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs);


				/// <summary>
				/// Calculates affine transform projection of given source rectangular ROI
				/// </summary>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="quad">Destination quadrangle [4,2]</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <returns>
				/// Error codes:
				///         - #NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///           has zero or negative value
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGetAffineQuad(NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] quad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs);


				/// <summary>
				/// Calculates bounding box of the affine transform projection of the given
				/// source rectangular ROI
				/// </summary>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="bound">Bounding box of the transformed source ROI [2,2]</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <returns>
				/// Error codes:
				///         - #NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///           has zero or negative value
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGetAffineBound(NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] double[,] bound, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, single channel). This
				/// function operates using given transform coefficients that can be obtained
				/// by using nppiGetAffineTransform function or set explicitly. The function
				/// operates on source and destination regions of interest. The affine warp
				/// function transforms the source image pixel coordinates \f$(x,y)\f$ according
				/// to the following formulas:
				/// \f[
				/// X_{new} = C_{00}		/// x + C_{01}		/// y + C_{02} \qquad
				/// Y_{new} = C_{10}		/// x + C_{11}		/// y + C_{12} \qquad
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetAffineQuad and nppiGetAffineBound can help with 
				/// destination ROI specification.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant,
				/// but does not perform memory access checks and requires the destination ROI
				/// to be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes
				/// aligned. This is always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI in
				/// such way that left and right sides of the projected 
				/// image are separated from the ROI by at least 63 bytes from each side.
				/// However, this requires the whole ROI to be part of allocated memory. In case
				/// when the conditions above are not satisfied, the function may decrease in
				/// speed slightly and will return NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           interpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);
								
				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, three channels).
				/// <see cref="nppiWarpAffine_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, four channels).
				/// <see cref="nppiWarpAffine_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, four channels RGBA).
				/// <see cref="nppiWarpAffine_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, three planes).
				/// <see cref="nppiWarpAffine_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, four planes).
				/// <see cref="nppiWarpAffine_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_8u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (8bit unsigned integer, single channel).
				/// This function operates using given transform coefficients that can be
				/// obtained by using nppiGetAffineTransform function or set explicitly. Thus
				/// there is no need to invert coefficients in your application before calling
				/// WarpAffineBack. The function operates on source and destination regions of
				/// interest.
				/// The affine warp function transforms the source image pixel coordinates
				/// \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// C_{00}		/// X_{new} + C_{01}		/// Y_{new} + C_{02} = x  \qquad
				/// C_{10}		/// X_{new} + C_{11}		/// Y_{new} + C_{12} = y
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetAffineQuad and nppiGetAffineBound can help with
				/// destination ROI specification.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant,
				/// but doesn't perform memory access checks and requires the destination ROI
				/// to be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes aligned. This is
				/// always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI in
				/// such way that left and right sides of the projected image are separated from
				/// the ROI by at least 63 bytes from each side. However, this requires the
				/// whole ROI to be part of allocated memory. In case when the conditions above
				/// are not satisfied, the function may decrease in speed slightly and will
				/// return NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           interpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (8bit unsigned integer, three channels).
				/// <see cref="nppiWarpAffineBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (8bit unsigned integer, four channels).
				/// <see cref="nppiWarpAffineBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (8bit unsigned integer, four channels RGBA).
				/// <see cref="nppiWarpAffineBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (8bit unsigned integer, three planes).
				/// <see cref="nppiWarpAffineBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (8bit unsigned integer, four planes).
				/// <see cref="nppiWarpAffineBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_8u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, single channel). This
				/// function performs affine warping of a the specified quadrangle in the
				/// source image to the specified quadrangle in the destination image. The
				/// function nppiWarpAffineQuad uses the same formulas for pixel mapping as in
				/// nppiWarpAffine function. The transform coefficients are computed internally.
				/// The transformed part of the source image is resampled using the specified
				/// eInterpolation method and written to the destination ROI.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant, but does not
				/// perform memory access checks and requires the destination ROI to be 64 bytes
				/// aligned. Hence any destination ROI is chunked into 3 vertical stripes: the
				/// first and the third are processed by accurate kernels and the central one is
				/// processed by the fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes aligned. This is
				/// always true if the values <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI in
				/// such way that left and right sides of the projected image are separated from
				/// the ROI by at least 63 bytes from each side. However, this requires the
				/// whole ROI to be part of allocated memory. In case when the conditions above
				/// are not satisfied, the function may decrease in speed slightly and will
				/// return NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="srcQuad">Source quadrangle [4,2]</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="dstQuad">Destination quadrangle [4,2]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				///         ignored, internally computed coordinates are used instead
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, three channels).
				/// <see cref="nppiWarpAffineQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, four channels).
				/// <see cref="nppiWarpAffineQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, four channels RGBA).
				/// <see cref="nppiWarpAffineQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, three planes).
				/// <see cref="nppiWarpAffineQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (8bit unsigned integer, four planes).
				/// <see cref="nppiWarpAffineQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_8u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, single channel). This
				/// function operates using given transform coefficients that can be obtained by
				/// using nppiGetAffineTransform function or set explicitly. The function
				/// operates on source and destination regions of interest. The affine warp
				/// function transforms the source image pixel coordinates \f$(x,y)\f$
				/// according to the following formulas:
				/// \f[
				/// X_{new} = C_{00}		/// x + C_{01}		/// y + C_{02} \qquad
				/// Y_{new} = C_{10}		/// x + C_{11}		/// y + C_{12} \qquad
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetAffineQuad and nppiGetAffineBound can help with
				/// destination ROI specification.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant, but does not
				/// perform memory access checks and requires the destination ROI to be 64 bytes
				/// aligned. Hence any destination ROI is chunked into 3 vertical stripes: the
				/// first and the third are processed by accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes aligned. This is
				/// always true if the values <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt>  are multiples of
				/// 64. Another rule of thumb is to specify destination ROI in such way that
				/// left and right sides of the projected image are separated from the ROI by at
				/// least 63 bytes from each side. However, this requires the whole ROI to be
				/// part of allocated memory. In case when the conditions above are not
				/// satisfied, the function may decrease in speed slightly and will return 
				/// NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, three channels)
				/// <see cref="nppiWarpAffine_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, four channels)
				/// <see cref="nppiWarpAffine_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, four channels RGBA)
				/// <see cref="nppiWarpAffine_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, three planes)
				/// <see cref="nppiWarpAffine_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, four planes)
				/// <see cref="nppiWarpAffine_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_16u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (16bit unsigned integer, single 
				/// channel). This function operates using given transform coefficients that 
				/// can be obtained by using nppiGetAffineTransform function or set explicitly.
				/// Thus there is no need to invert coefficients in your application 
				/// before calling WarpAffineBack. The function operates on source and
				/// destination regions of interest.
				/// The affine warp function transforms the source image pixel coordinates
				/// \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// C_{00}		/// X_{new} + C_{01}		/// Y_{new} + C_{02} = x  \qquad
				/// C_{10}		/// X_{new} + C_{11}		/// Y_{new} + C_{12} = y
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetAffineQuad and nppiGetAffineBound can help with
				/// destination ROI specification.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant,
				/// but doesn't perform memory access checks and requires the destination ROI
				/// to be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes
				/// aligned. This is always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI in
				/// such way that left and right sides of the projected 
				/// image are separated from the ROI by at least 63 bytes from each side.
				/// However, this requires the whole ROI to be part of allocated 
				/// memory. In case when the conditions above are not satisfied, the function may
				/// decrease in speed slightly and will return 
				/// NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (16bit unsigned integer, three channels)
				/// <see cref="nppiWarpAffineBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (16bit unsigned integer, four channels)
				/// <see cref="nppiWarpAffineBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (16bit unsigned integer, four channels
				/// RGBA)
				/// <see cref="nppiWarpAffineBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (16bit unsigned integer, three planes)
				/// <see cref="nppiWarpAffineBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (16bit unsigned integer, four planes)
				/// <see cref="nppiWarpAffineBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_16u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, single channel). This
				/// function performs affine warping of a the specified
				/// quadrangle in the source image to the specified quadrangle in the
				/// destination image. The function nppiWarpAffineQuad uses the same
				/// formulas for pixel mapping as in nppiWarpAffine function. The transform
				/// coefficients are computed internally.
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant,
				/// but doesn't perform memory access checks and requires the destination ROI
				/// to be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes
				/// aligned. This is always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI in 
				/// such way that left and right sides of the projected 
				/// image are separated from the ROI by at least 63 bytes from each side.
				/// However, this requires the whole ROI to be part of allocated 
				/// memory. In case when the conditions above are not satisfied, the function
				/// may decrease in speed slightly and will return 
				/// NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="srcQuad">Source quadrangle [4,2]</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="dstQuad">Destination quadrangle [4,2]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, three channels). 
				/// <see cref="nppiWarpAffineQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, four channels). 
				/// <see cref="nppiWarpAffineQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, four channels RGBA). 
				/// <see cref="nppiWarpAffineQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, three planes). 
				/// <see cref="nppiWarpAffineQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (16bit unsigned integer, four planes). 
				/// <see cref="nppiWarpAffineQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_16u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, single channel). This function
				/// operates using given transform coefficients that 
				/// can be obtained by using nppiGetAffineTransform function or set explicitly.
				/// The function operates on source and destination regions 
				/// of interest. The affine warp function transforms the source image pixel
				/// coordinates \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// X_{new} = C_{00}		/// x + C_{01}		/// y + C_{02} \qquad
				/// Y_{new} = C_{10}		/// x + C_{11}		/// y + C_{12} \qquad
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetAffineQuad and nppiGetAffineBound can help with
				/// destination ROI specification.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, three channels).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, four channels).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, four channels RGBA).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, three planes).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, four planes).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit float, single channel). This
				/// function operates using given transform coefficients that 
				/// can be obtained by using nppiGetAffineTransform function or set explicitly.
				/// Thus there is no need to invert coefficients in your application 
				/// before calling WarpAffineBack. The function operates on source and
				/// destination regions of interest.
				/// The affine warp function transforms the source image pixel coordinates
				/// \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// C_{00}		/// X_{new} + C_{01}		/// Y_{new} + C_{02} = x  \qquad
				/// C_{10}		/// X_{new} + C_{11}		/// Y_{new} + C_{12} = y
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetAffineQuad and nppiGetAffineBound can help with
				/// destination ROI specification.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit float, three channels).
				/// <see cref="nppiWarpAffineBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit float, four channels).
				/// <see cref="nppiWarpAffineBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit float, four channels RGBA).
				/// <see cref="nppiWarpAffineBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit float, three planes).
				/// <see cref="nppiWarpAffineBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit float, four planes).
				/// <see cref="nppiWarpAffineBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, single channel). This function
				/// performs affine warping of a the specified
				/// quadrangle in the source image to the specified quadrangle in the
				/// destination image. The function nppiWarpAffineQuad uses the same
				/// formulas for pixel mapping as in nppiWarpAffine function. The transform
				/// coefficients are computed internally.
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="srcQuad">Source quadrangle [4,2]</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="dstQuad">Destination quadrangle [4,2]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, three channels).
				/// <see cref="nppiWarpAffineQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, four channels).
				/// <see cref="nppiWarpAffineQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, four channels RGBA).
				/// <see cref="nppiWarpAffineQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, three planes).
				/// <see cref="nppiWarpAffineQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, four planes).
				/// <see cref="nppiWarpAffineQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, single channel). This
				/// function operates using given transform coefficients that 
				/// can be obtained by using nppiGetAffineTransform function or set explicitly.
				/// The function operates on source and destination regions 
				/// of interest. The affine warp function transforms the source image pixel
				/// coordinates \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// X_{new} = C_{00}		/// x + C_{01}		/// y + C_{02} \qquad
				/// Y_{new} = C_{10}		/// x + C_{11}		/// y + C_{12} \qquad
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetAffineQuad and nppiGetAffineBound can help with
				/// destination ROI specification.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32s_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, three channels).
				/// <see cref="nppiWarpAffine_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32s_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode interpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, four channels).
				/// <see cref="nppiWarpAffine_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32s_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, four channels RGBA).
				/// <see cref="nppiWarpAffine_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32s_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, three planes).
				/// <see cref="nppiWarpAffine_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, four planes).
				/// <see cref="nppiWarpAffine_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_32s_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit signed integer, single
				/// channel). This function operates using given transform coefficients that 
				/// can be obtained by using nppiGetAffineTransform function or set explicitly.
				/// Thus there is no need to invert coefficients in your application 
				/// before calling WarpAffineBack. The function operates on source and
				/// destination regions of interest.
				/// The affine warp function transforms the source image pixel coordinates
				/// \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// C_{00}		/// X_{new} + C_{01}		/// Y_{new} + C_{02} = x  \qquad
				/// C_{10}		/// X_{new} + C_{11}		/// Y_{new} + C_{12} = y
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetAffineQuad and nppiGetAffineBound can help with
				/// destination ROI specification.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Affine transform coefficients [2,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32s_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit signed integer, three channels).
				/// <see cref="nppiWarpAffineBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32s_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit signed integer, four channels).
				/// <see cref="nppiWarpAffineBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32s_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit signed integer, four channels RGBA).
				/// <see cref="nppiWarpAffineBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32s_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit signed integer, three planes).
				/// <see cref="nppiWarpAffineBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse affine transform of an image (32bit signed integer, four planes).
				/// <see cref="nppiWarpAffineBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineBack_32s_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, single channel). This
				/// function performs affine warping of a the specified
				/// quadrangle in the source image to the specified quadrangle in the
				/// destination image. The function nppiWarpAffineQuad uses the same
				/// formulas for pixel mapping as in nppiWarpAffine function. The transform
				/// coefficients are computed internally.
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="srcQuad">Source quadrangle [4,2]</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="dstQuad">Destination quadrangle [4,2]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				///           misalignment
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32s_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, three channels).
				/// <see cref="nppiWarpAffineQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32s_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, four channels).
				/// <see cref="nppiWarpAffineQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32s_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, four channels RGBA).
				/// <see cref="nppiWarpAffineQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32s_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, three planes).
				/// <see cref="nppiWarpAffineQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit signed integer, four planes).
				/// <see cref="nppiWarpAffineQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffineQuad_32s_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);




				/// <summary>
				/// Affine transform of an image (32bit float, three channels).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_64f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, four channels).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_64f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (32bit float, four channels RGBA).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_64f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);

				/// <summary>
				/// Affine transform of an image (64bit float, one channel).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_64f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (64bit float, three channels RGB).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_64f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] aSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] aDst, int nDstStep, NppiRect oDestROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] aCoeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Affine transform of an image (64bit float, four channels RGBA).
				/// <see cref="nppiWarpAffine_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpAffine_64f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] aSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] aDst, int nDstStep, NppiRect oDestROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 6)] double[,] aCoeffs, InterpolationMode eInterpolation);



			}

			/// <summary>
			/// Perspective warping, perspective transform calculation
			/// Perspective warping of an image is the transform of image pixel positions,
			/// defined by the following formulas:
			/// \f[
			/// X_{new} = \frac{C_{00} * x + C_{01} * y + C_{02}}{C_{20} * x + C_{21} * y + C_{22}} \qquad
			/// Y_{new} = \frac{C_{10} * x + C_{11} * y + C_{12}}{C_{20} * x + C_{21} * y + C_{22}} \qquad
			/// C = \left[ \matrix{C_{00} <![CDATA[&]]> C_{01} <![CDATA[&]]> C_{02} \cr C_{10} <![CDATA[&]]> C_{11} <![CDATA[&]]> C_{12} \cr C_{20} <![CDATA[&]]> C_{21} <![CDATA[&]]> C_{22} } \right]
			/// \f]
			/// That is, any pixel of the transformed image with coordinates 
			/// \f$(X_{new},Y_{new})\f$ has a preimage with 
			/// coordinates \f$(x,y)\f$. The mapping \f$C\f$ is fully defined by 8 values
			/// \f$C_{ij}, (i,j)=\overline{0,2}\f$, except of \f$C_{22}\f$, which is a
			/// normalizer.
			/// The transform has a property of mapping any convex quadrangle to a convex
			/// quadrangle, which is used in a group of functions nppiWarpPerspectiveQuad.
			/// The NPPI implementation of perspective transform has some issues which are
			/// discussed in each function's documentation.
			/// </summary>
			public static class PerspectiveTransforms
			{
				/// <summary>
				/// Calculates perspective transform coefficients given source rectangular ROI
				/// and its destination quadrangle projection
				/// </summary>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="quad">Destination quadrangle [4,2]</param>
				/// <param name="coeffs">Perspective transform coefficients [3,3]</param>
				/// <returns>
				/// Error codes:
				///         - #NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///           has zero or negative value
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGetPerspectiveTransform(NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] quad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs);


				/// <summary>
				/// Calculates perspective transform projection of given source rectangular
				/// ROI
				/// </summary>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="quad">Destination quadrangle [4,2]</param>
				/// <param name="coeffs">Perspective transform coefficients [3,3]</param>
				/// <returns>
				/// Error codes:
				///         - #NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///           has zero or negative value
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGetPerspectiveQuad(NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] quad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs);


				/// <summary>
				/// Calculates bounding box of the perspective transform projection of the
				/// given source rectangular ROI
				/// </summary>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="bound">Bounding box of the transformed source ROI</param>
				/// <param name="coeffs">Perspective transform coefficients [3,3]</param>
				/// <returns>
				/// Error codes:
				///         - #NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///           has zero or negative value
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGetPerspectiveBound(NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] double[,] bound, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, single channel).
				/// This function operates using given transform coefficients that 
				/// can be obtained by using nppiGetPerspectiveTransform function or set
				/// explicitly. The function operates on source and destination regions 
				/// of interest. The perspective warp function transforms the source image
				/// pixel coordinates \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// X_{new} = \frac{C_{00}		/// x + C_{01}		/// y + C_{02}}{C_{20}		/// x + C_{21}		/// y + C_{22}} \qquad
				/// Y_{new} = \frac{C_{10}		/// x + C_{11}		/// y + C_{12}}{C_{20}		/// x + C_{21}		/// y + C_{22}}
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetPerspectiveQuad and nppiGetPerspectiveBound can help
				/// with destination ROI specification.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant,
				/// but does not perform memory access checks and requires the destination ROI
				/// to be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes
				/// aligned. This is always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI in
				/// such way that left and right sides of the projected 
				/// image are separated from the ROI by at least 63 bytes from each side.
				/// However, this requires the whole ROI to be part of allocated 
				/// memory. In case when the conditions above are not satisfied, the function
				/// may decrease in speed slightly and will return 
				/// NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Perspective transform coefficients [3,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, three channels).
				/// <see cref="nppiWarpPerspective_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, four channels).
				/// <see cref="nppiWarpPerspective_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, four channels RGBA).
				/// <see cref="nppiWarpPerspective_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, three planes).
				/// <see cref="nppiWarpPerspective_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, four planes).
				/// <see cref="nppiWarpPerspective_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_8u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (8bit unsigned integer, single
				/// channel). This function operates using given transform coefficients that 
				/// can be obtained by using nppiGetPerspectiveTransform function or set
				/// explicitly. Thus there is no need to invert coefficients in your application 
				/// before calling WarpPerspectiveBack. The function operates on source and
				/// destination regions of interest.
				/// The perspective warp function transforms the source image pixel
				/// coordinates \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// \frac{C_{00}		/// X_{new} + C_{01}		/// Y_{new} + C_{02}}{C_{20}		/// X_{new} + C_{21}		/// Y_{new} + C_{22}} = x  \qquad
				/// \frac{C_{10}		/// X_{new} + C_{11}		/// Y_{new} + C_{12}}{C_{20}		/// X_{new} + C_{21}		/// Y_{new} + C_{22}} = y
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetPerspectiveQuad and nppiGetPerspectiveBound can help
				/// with destination ROI specification.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant,
				/// but does not perform memory access checks and requires the destination ROI 
				/// to be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes
				/// aligned. This is always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI
				/// in such way that left and right sides of the projected 
				/// image are separated from the ROI by at least 63 bytes from each side.
				/// However, this requires the whole ROI to be part of allocated 
				/// memory. In case when the conditions above are not satisfied, the function
				/// may decrease in speed slightly and will return 
				/// NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Perspective transform coefficients [3,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (8bit unsigned integer, three channels).
				/// <see cref="nppiWarpPerspectiveBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (8bit unsigned integer, four channels).
				/// <see cref="nppiWarpPerspectiveBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (8bit unsigned integer, four channels RGBA).
				/// <see cref="nppiWarpPerspectiveBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (8bit unsigned integer, three planes).
				/// <see cref="nppiWarpPerspectiveBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (8bit unsigned integer, four planes).
				/// <see cref="nppiWarpPerspectiveBack_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_8u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, single channel).
				/// This function performs perspective warping of a the specified
				/// quadrangle in the source image to the specified quadrangle in the
				/// destination image. The function nppiWarpPerspectiveQuad uses the same
				/// formulas for pixel mapping as in nppiWarpPerspective function. The
				/// transform coefficients are computed internally.
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant,
				/// but does not perform memory access checks and requires the destination ROI
				/// to be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection
				/// of destination ROI onto image addresses must be 64 bytes
				/// aligned. This is always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI
				/// in such way that left and right sides of the projected 
				/// image are separated from the ROI by at least 63 bytes from each side.
				/// However, this requires the whole ROI to be part of allocated 
				/// memory. In case when the conditions above are not satisfied, the function
				/// may decrease in speed slightly and will return 
				/// NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="srcQuad">Source quadrangle [4,2]</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="dstQuad">Destination quadrangle [4,2]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, three channels).
				/// <see cref="nppiWarpPerspectiveQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, four channels).
				/// <see cref="nppiWarpPerspectiveQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, four channels RGBA).
				/// <see cref="nppiWarpPerspectiveQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, three planes).
				/// <see cref="nppiWarpPerspectiveQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (8bit unsigned integer, four planes).
				/// <see cref="nppiWarpPerspectiveQuad_8u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_8u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, single channel).
				/// This function operates using given transform coefficients that 
				/// can be obtained by using nppiGetPerspectiveTransform function or set
				/// explicitly. The function operates on source and destination regions 
				/// of interest. The perspective warp function transforms the source image pixel
				/// coordinates \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// X_{new} = \frac{C_{00}		/// x + C_{01}		/// y + C_{02}}{C_{20}		/// x + C_{21}		/// y + C_{22}} \qquad
				/// Y_{new} = \frac{C_{10}		/// x + C_{11}		/// y + C_{12}}{C_{20}		/// x + C_{21}		/// y + C_{22}}
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetPerspectiveQuad and nppiGetPerspectiveBound can help
				/// with destination ROI specification.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The
				/// fast method is about 4 times faster than its accurate variant,
				/// but does not perform memory access checks and requires the destination ROI
				/// to be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes
				/// aligned. This is always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI in
				/// such way that left and right sides of the projected 
				/// image are separated from the ROI by at least 63 bytes from each side.
				/// However, this requires the whole ROI to be part of allocated 
				/// memory. In case when the conditions above are not satisfied, the function
				/// may decrease in speed slightly and will return 
				/// NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs"> Perspective transform coefficients [3,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, three channels)
				/// <see cref="nppiWarpPerspective_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, four channels)
				/// <see cref="nppiWarpPerspective_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, four channels RGBA)
				/// <see cref="nppiWarpPerspective_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, three planes)
				/// <see cref="nppiWarpPerspective_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, four planes)
				/// <see cref="nppiWarpPerspective_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_16u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (16bit unsigned integer, single
				/// channel). This function operates using given transform coefficients that 
				/// can be obtained by using nppiGetPerspectiveTransform function or set
				/// explicitly. Thus there is no need to invert coefficients in your application 
				/// before calling WarpPerspectiveBack. The function operates on source and
				/// destination regions of interest.
				/// The perspective warp function transforms the source image pixel coordinates
				/// \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// \frac{C_{00}		/// X_{new} + C_{01}		/// Y_{new} + C_{02}}{C_{20}		/// X_{new} + C_{21}		/// Y_{new} + C_{22}} = x  \qquad
				/// \frac{C_{10}		/// X_{new} + C_{11}		/// Y_{new} + C_{12}}{C_{20}		/// X_{new} + C_{21}		/// Y_{new} + C_{22}} = y
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetPerspectiveQuad and nppiGetPerspectiveBound can help
				/// with destination ROI specification.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant,
				/// but does not perform memory access checks and requires the destination ROI
				/// to be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes
				/// aligned. This is always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI in
				/// such way that left and right sides of the projected 
				/// image are separated from the ROI by at least 63 bytes from each side.
				/// However, this requires the whole ROI to be part of allocated 
				/// memory. In case when the conditions above are not satisfied, the function
				/// may decrease in speed slightly and will return 
				/// NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Perspective transform coefficients [3,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (16bit unsigned integer, three channels)
				/// <see cref="nppiWarpPerspectiveBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (16bit unsigned integer, four channels)
				/// <see cref="nppiWarpPerspectiveBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (16bit unsigned integer, four channels RGBA)
				/// <see cref="nppiWarpPerspectiveBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (16bit unsigned integer, three planes)
				/// <see cref="nppiWarpPerspectiveBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (16bit unsigned integer, four planes)
				/// <see cref="nppiWarpPerspectiveBack_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_16u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, single channel).
				/// This function performs perspective warping of a the specified
				/// quadrangle in the source image to the specified quadrangle in the
				/// destination image. The function nppiWarpPerspectiveQuad uses the same
				/// formulas for pixel mapping as in nppiWarpPerspective function. The transform
				/// coefficients are computed internally.
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// 
				/// NPPI specific recommendation: 
				/// The function operates using 2 types of kernels: fast and accurate. The fast
				/// method is about 4 times faster than its accurate variant,
				/// but does not perform memory access checks and requires the destination ROI to
				/// be 64 bytes aligned. Hence any destination ROI is 
				/// chunked into 3 vertical stripes: the first and the third are processed by
				/// accurate kernels and the central one is processed by the
				/// fast one.
				/// In order to get the maximum available speed of execution, the projection of
				/// destination ROI onto image addresses must be 64 bytes
				/// aligned. This is always true if the values 
				/// <tt>(int)((void *)(pDst + dstRoi.x))</tt> and 
				/// <tt>(int)((void *)(pDst + dstRoi.x + dstRoi.width))</tt> 
				/// are multiples of 64. Another rule of thumb is to specify destination ROI in
				/// such way that left and right sides of the projected 
				/// image are separated from the ROI by at least 63 bytes from each side.
				/// However, this requires the whole ROI to be part of allocated 
				/// memory. In case when the conditions above are not satisfied, the function
				/// may decrease in speed slightly and will return 
				/// NPP_MISALIGNED_DST_ROI_WARNING warning.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="srcQuad">Source quadrangle [4,2]</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="dstQuad">Destination quadrangle [4,2]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, three channels). 
				/// <see cref="nppiWarpPerspectiveQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, four channels). 
				/// <see cref="nppiWarpPerspectiveQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, four channels RGBA). 
				/// <see cref="nppiWarpPerspectiveQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, three planes). 
				/// <see cref="nppiWarpPerspectiveQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (16bit unsigned integer, four planes). 
				/// <see cref="nppiWarpPerspectiveQuad_16u_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_16u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, single channel). This
				/// function operates using given transform coefficients that 
				/// can be obtained by using nppiGetPerspectiveTransform function or set
				/// explicitly. The function operates on source and destination regions 
				/// of interest. The perspective warp function transforms the source image pixel
				/// coordinates \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// X_{new} = \frac{C_{00}		/// x + C_{01}		/// y + C_{02}}{C_{20}		/// x + C_{21}		/// y + C_{22}} \qquad
				/// Y_{new} = \frac{C_{10}		/// x + C_{11}		/// y + C_{12}}{C_{20}		/// x + C_{21}		/// y + C_{22}}
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetPerspectiveQuad and nppiGetPerspectiveBound can help
				/// with destination ROI specification.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Perspective transform coefficients [3,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, three channels).
				/// <see cref="nppiWarpPerspective_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, four channels).
				/// <see cref="nppiWarpPerspective_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, four channels RGBA).
				/// <see cref="nppiWarpPerspective_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, three planes).
				/// <see cref="nppiWarpPerspective_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, four planes).
				/// <see cref="nppiWarpPerspective_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit float, single channel).
				/// This function operates using given transform coefficients that 
				/// can be obtained by using nppiGetPerspectiveTransform function or set
				/// explicitly. Thus there is no need to invert coefficients in your application 
				/// before calling WarpPerspectiveBack. The function operates on source and
				/// destination regions of interest.
				/// The perspective warp function transforms the source image pixel coordinates
				/// \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// \frac{C_{00}		/// X_{new} + C_{01}		/// Y_{new} + C_{02}}{C_{20}		/// X_{new} + C_{21}		/// Y_{new} + C_{22}} = x  \qquad
				/// \frac{C_{10}		/// X_{new} + C_{11}		/// Y_{new} + C_{12}}{C_{20}		/// X_{new} + C_{21}		/// Y_{new} + C_{22}} = y
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetPerspectiveQuad and nppiGetPerspectiveBound can help
				/// with destination ROI specification.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Perspective transform coefficients [3,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit float, three channels).
				/// <see cref="nppiWarpPerspectiveBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit float, four channels).
				/// <see cref="nppiWarpPerspectiveBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit float, four channels RGBA).
				/// <see cref="nppiWarpPerspectiveBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit float, three planes).
				/// <see cref="nppiWarpPerspectiveBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit float, four planes).
				/// <see cref="nppiWarpPerspectiveBack_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, single channel).
				/// This function performs perspective warping of a the specified
				/// quadrangle in the source image to the specified quadrangle in the
				/// destination image. The function nppiWarpPerspectiveQuad uses the same
				/// formulas for pixel mapping as in nppiWarpPerspective function. The transform
				/// coefficients are computed internally.
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="srcQuad">Source quadrangle [4,2]</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="dstQuad">Destination quadrangle [4,2]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, three channels).
				/// <see cref="nppiWarpPerspectiveQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, four channels).
				/// <see cref="nppiWarpPerspectiveQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, four channels RGBA).
				/// <see cref="nppiWarpPerspectiveQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, three planes).
				/// <see cref="nppiWarpPerspectiveQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit float, four planes).
				/// <see cref="nppiWarpPerspectiveQuad_32f_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, single channel).
				/// This function operates using given transform coefficients that 
				/// can be obtained by using nppiGetPerspectiveTransform function or set
				/// explicitly. The function operates on source and destination regions 
				/// of interest. The perspective warp function transforms the source image pixel
				/// coordinates \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// X_{new} = \frac{C_{00}		/// x + C_{01}		/// y + C_{02}}{C_{20}		/// x + C_{21}		/// y + C_{22}} \qquad
				/// Y_{new} = \frac{C_{10}		/// x + C_{11}		/// y + C_{12}}{C_{20}		/// x + C_{21}		/// y + C_{22}}
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetPerspectiveQuad and nppiGetPerspectiveBound can help
				/// with destination ROI specification.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs"> Perspective transform coefficients [3,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32s_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, three channels).
				/// <see cref="nppiWarpPerspective_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32s_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, four channels).
				/// <see cref="nppiWarpPerspective_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32s_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, four channels RGBA).
				/// <see cref="nppiWarpPerspective_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32s_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, three planes).
				/// <see cref="nppiWarpPerspective_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, four planes).
				/// <see cref="nppiWarpPerspective_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspective_32s_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit signed integer, single
				/// channel). This function operates using given transform coefficients that 
				/// can be obtained by using nppiGetPerspectiveTransform function or set
				/// explicitly. Thus there is no need to invert coefficients in your
				/// application 
				/// before calling WarpPerspectiveBack. The function operates on source
				/// and destination regions of interest.
				/// The perspective warp function transforms the source image pixel coordinates
				/// \f$(x,y)\f$ according to the following formulas:
				/// \f[
				/// \frac{C_{00}		/// X_{new} + C_{01}		/// Y_{new} + C_{02}}{C_{20}		/// X_{new} + C_{21}		/// Y_{new} + C_{22}} = x  \qquad
				/// \frac{C_{10}		/// X_{new} + C_{11}		/// Y_{new} + C_{12}}{C_{20}		/// X_{new} + C_{21}		/// Y_{new} + C_{22}} = y
				/// \f]
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// The functions nppiGetPerspectiveQuad and nppiGetPerspectiveBound can help
				/// with destination ROI specification.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="coeffs">Perspective transform coefficients [3,3]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32s_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit signed integer, three channels).
				/// <see cref="nppiWarpPerspectiveBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32s_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit signed integer, four channels).
				/// <see cref="nppiWarpPerspectiveBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32s_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit signed integer, four channels RGBA).
				/// <see cref="nppiWarpPerspectiveBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32s_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit signed integer, three planes).
				/// <see cref="nppiWarpPerspectiveBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Inverse perspective transform of an image (32bit signed integer, four planes).
				/// <see cref="nppiWarpPerspectiveBack_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveBack_32s_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 9)] double[,] coeffs, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, single channel).
				/// This function performs perspective warping of a the specified
				/// quadrangle in the source image to the specified quadrangle in the
				/// destination image. The function nppiWarpPerspectiveQuad uses the same
				/// formulas for pixel mapping as in nppiWarpPerspective function. The transform
				/// coefficients are computed internally.
				/// The transformed part of the source image is resampled using the specified
				/// interpolation method and written to the destination ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="oSrcSize">Size of source image in pixels</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="srcRoi">Source ROI</param>
				/// <param name="srcQuad">Source quadrangle [4,2]</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="dstRoi">Destination ROI</param>
				/// <param name="dstQuad">Destination quadrangle [4,2]</param>
				/// <param name="eInterpolation">Interpolation mode: can be <see cref="InterpolationMode.NearestNeighbor"/>, <see cref="InterpolationMode.Linear"/> or <see cref="InterpolationMode.Cubic"/></param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///         - #NPP_RECT_ERROR Indicates an error condition if width or height of
				///           the intersection of the srcRoi and source image is less than or
				///           equal to 1
				///         - #NPP_WRONG_INTERSECTION_ROI_ERROR Indicates an error condition if
				///           srcRoi has no intersection with the source image
				///         - #NPP_INTERPOLATION_ERROR Indicates an error condition if
				///           eInterpolation has an illegal value
				///         - #NPP_COEFF_ERROR Indicates an error condition if coefficient values
				///           are invalid
				///         - #NPP_WRONG_INTERSECTION_QUAD_WARNING Indicates a warning that no
				///           operation is performed if the transformed source ROI has no
				///           intersection with the destination ROI
				///         - #NPP_MISALIGNED_DST_ROI_WARNING Indicates a warning that the speed
				///           of primitive execution was reduced due to destination ROI
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32s_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, three channels).
				/// <see cref="nppiWarpPerspectiveQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32s_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, four channels).
				/// <see cref="nppiWarpPerspectiveQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32s_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, four channels RGBA).
				/// <see cref="nppiWarpPerspectiveQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32s_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, CUdeviceptr pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, three planes).
				/// <see cref="nppiWarpPerspectiveQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);


				/// <summary>
				/// Perspective transform of an image (32bit signed integer, four planes).
				/// <see cref="nppiWarpPerspectiveQuad_32s_C1R"/>
				/// </summary>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiWarpPerspectiveQuad_32s_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect srcRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] srcQuad, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect dstRoi, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 8)] double[,] dstQuad, InterpolationMode eInterpolation);



			}

			/// <summary>
			/// RGB to YCbCr color conversion.
			/// </summary>
			public static class RGBToYCbCr
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to packed YCbCr color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCbCr_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned RGB to 2 channel chroma packed YCbCr422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCbCr422_8u_C3C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to planar YCbCr420 color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCbCr420_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel planar 8-bit unsigned RGB to YCbCr color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCbCr_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned RGB to YCbCr color conversion, ignoring Alpha.
				/// 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCbCr_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel unsigned 8-bit packed YCbCr color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCbCr_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 3 channel 8-bit unsigned planar YCbCr color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCbCr_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);

				
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned planar YCbCr422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCbCr422_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar RGB to 2 channel 8-bit unsigned packed YCbCr422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCbCr422_8u_P3C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// YCbCr to RGB color conversion.
			/// </summary>
			public static class YCbCrToRGB
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed YCbCr to RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCrToRGB_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr to RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCrToRGB_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed YCbCr to RGB color conversion, not affecting Alpha.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCrToRGB_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned YCbCr422 to 3 channel packed RGB color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToRGB_8u_C2C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr420 to packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToRGB_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCrToRGB_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr to 4 channel 8-bit unsigned packed RGB color conversion with constant alpha.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.                                         </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCrToRGB_8u_P3C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCbCr422 to 3 channel 8-bit unsigned planar RGB color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToRGB_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr422 to 3 channel 8-bit unsigned packed RGB color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToRGB_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 2 channel 8-bit unsigned packed YCrCb422 to 3 channel 8-bit unsigned packed RGB color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb422ToRGB_8u_C2C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCrCb422 to 3 channel 8-bit unsigned planar RGB color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb422ToRGB_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// Sample Pattern Conversion.
			/// </summary>
			public static class SamplePatternConversion
			{
				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr:422 to YCbCr:420 resampling.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCbCr420_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr:422 to YCbCr:411 resampling.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCbCr411_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr:420 to YCbCr:422 resampling.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToYCbCr422_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr:420 to YCbCr:411 resampling.
				/// </summary>
				/// <param name="pSrc">Array of pointers to the source image planes.</param>
				/// <param name="aSrcStep">Array with distances in bytes between starts of consecutive</param>
				/// <param name="pDstY">Destination-Image Pointer. Y-channel.</param>
				/// <param name="nDstYStep">Destination-Image Line Step. Y-channel.</param>
				/// <param name="pDstCbCr">Destination-Image Pointer. CbCr image.</param>
				/// <param name="nDstCbCrStep">Destination-Image Line Step. CbCr image.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToYCbCr411_8u_P3P2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);



			}

			/// <summary>
			/// Color manipuliation functions.
			/// </summary>
			public static class ColorProcessing
			{
				/// <summary>
				/// 3 channel 8-bit unsigned color twist.
				/// 
				/// An input color twist matrix with floating-point pixel values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="twist">The color twist matrix with floating-point pixel values [3,4].</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] twist);


				/// <summary>
				/// 3 channel planar 8-bit unsigned color twist.
				/// An input color twist matrix with floating-point pixel values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="twist">The color twist matrix with floating-point pixel values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] twist);


				/// <summary>
				/// 4 channel 8-bit unsigned color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point pixel values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="twist">The color twist matrix with floating-point pixel values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] twist);


				/// <summary>
				/// 8-bit unsigned look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points through linear
				/// interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values</param>
				/// <param name="nLevels">Number of user defined input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - #NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than
				///         2.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 8-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through linear
				/// interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Double pointer to an [3] of arrays of user defined OUTPUT</param>
				/// <param name="pLevels">Double pointer to an [3] of arrays of user defined INPUT</param>
				/// <param name="nLevels">A [3] array of user defined input/output mapping points</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - #NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than
				///         2.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points through linear
				/// interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Double pointer to an [3] of arrays of user defined OUTPUT</param>
				/// <param name="pLevels">Double pointer to an [3] of arrays of user defined INPUT</param>
				/// <param name="nLevels">A [3] array of user defined input/output mapping points</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - #NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than
				///         2.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);

				/// <summary>
				/// 4 channel 8-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through linear
				/// interpolation. 
				/// 
				/// <remarks>As of the 5.0 release of NPP, the pValues and pLevels pointers need to be host memory pointers to arrays of device memory pointers.</remarks>
				/// 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - #NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);

			}

			/// <summary>
			/// Image Labeling Techniques.
			/// </summary>
			public static class ImageLabeling
			{
				/// <summary>
				/// Calculates the size of the temporary buffer for graph-cut with 4 neighborhood labeling.
				/// <see cref="nppiGraphcut_32s8u"/>
				/// </summary>
				/// <param name="oSize">Graph size.</param>
				/// <param name="pBufSize">Pointer to variable that returns the size of the</param>
				/// <returns>
				/// NPP_SUCCESS Indicates no error. Any other value indicates an error
				///         or a warning
				/// NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///         has zero or negative value
				/// NPP_NULL_POINTER_ERROR Indicates an error condition if pBufSize
				///         pointer is NULL
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGraphcutGetSize(NppiSize oSize, ref int pBufSize);


				/// <summary>
				/// Calculates the size of the temporary buffer for graph-cut with 8 neighborhood labeling.
				/// <see cref="nppiGraphcut8_32s8u"/>
				/// </summary>
				/// <param name="oSize">Graph size.</param>
				/// <param name="pBufSize">Pointer to variable that returns the size of the</param>
				/// <returns>
				/// NPP_SUCCESS Indicates no error. Any other value indicates an error
				///         or a warning
				/// NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///         has zero or negative value
				/// NPP_NULL_POINTER_ERROR Indicates an error condition if pBufSize
				///         pointer is NULL
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGraphcut8GetSize(NppiSize oSize, ref int pBufSize);


				/// <summary>
				/// Initializes graph-cut state structure and allocates additional resources for graph-cut with 8 neighborhood labeling.
				/// <see cref="nppiGraphcut_32s8u"/>, 
				/// <see cref="nppiGraphcutGetSize"/> 
				/// </summary>
				/// <param name="oSize">Graph size</param>
				/// <param name="ppState">Pointer to pointer to graph-cut state structure. </param>
				/// <param name="pDeviceMem">to the device memory used </param>
				/// <returns>
				/// NPP_SUCCESS Indicates no error. Any other value indicates an error
				///         or a warning
				/// NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///         has zero or negative value
				/// NPP_NULL_POINTER_ERROR Indicates an error condition if pBufSize
				///         pointer is NULL
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGraphcutInitAlloc(NppiSize oSize, ref NppiGraphcutState ppState, CUdeviceptr pDeviceMem);


				/// <summary>
				/// Allocates and initializes the graph-cut state structure and additional resources for graph-cut with 8 neighborhood labeling.
				/// <see cref="nppiGraphcut8_32s8u"/>, 
				/// <see cref="nppiGraphcut8GetSize"/>
				/// </summary>
				/// <param name="oSize">Graph size</param>
				/// <param name="ppState">Pointer to pointer to graph-cut state structure. </param>
				/// <param name="pDeviceMem">to the device memory used </param>
				/// <returns>
				/// NPP_SUCCESS Indicates no error. Any other value indicates an error
				///         or a warning
				/// NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///         has zero or negative value
				/// NPP_NULL_POINTER_ERROR Indicates an error condition if pBufSize
				///         pointer is NULL
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGraphcut8InitAlloc(NppiSize oSize, ref NppiGraphcutState ppState, CUdeviceptr pDeviceMem);


				/// <summary>
				/// Frees the additional resources of the graph-cut state structure.
				/// <see cref="nppiGraphcutInitAlloc"/>, 
				/// <see cref="nppiGraphcut8InitAlloc"/>
				/// </summary>
				/// <param name="pState">Pointer to graph-cut state structure. </param>
				/// <returns>
				/// NPP_SUCCESS Indicates no error. Any other value indicates an error
				///         or a warning
				/// NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///         has zero or negative value
				/// NPP_NULL_POINTER_ERROR Indicates an error condition if pState
				///         pointer is NULL
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGraphcutFree(NppiGraphcutState pState);


				/// <summary>
				/// Graphcut of a flow network (32bit signed integer edge capacities). The
				/// function computes the minimal cut (graphcut) of a 2D regular 4-connected
				/// graph. 
				/// The inputs are the capacities of the horizontal (in transposed form),
				/// vertical and terminal (source and sink) edges. The capacities to source and
				/// sink 
				/// are stored as capacity differences in the terminals array 
				/// ( terminals(x) = source(x) - sink(x) ). The implementation assumes that the
				/// edge capacities 
				/// for boundary edges that would connect to nodes outside the specified domain
				/// are set to 0 (for example left(0,*) == 0). If this is not fulfilled the
				/// computed labeling may be wrong!
				/// The computed binary labeling is encoded as unsigned 8bit values (0 and >0).
				/// <see cref="nppiGraphcutGetSize"/>, 
				/// <see cref="nppiGraphcutInitAlloc"/>
				/// </summary>
				/// <param name="pTerminals">Pointer to differences of terminal edge capacities</param>
				/// <param name="pLeftTransposed">Pointer to transposed left edge capacities</param>
				/// <param name="pRightTransposed">Pointer to transposed right edge capacities</param>
				/// <param name="pTop">Pointer to top edge capacities (top(*,0) must be 0)</param>
				/// <param name="pBottom">Pointer to bottom edge capacities (bottom(*,height-1)</param>
				/// <param name="nStep">Step in bytes between any pair of sequential rows of edge</param>
				/// <param name="nTransposedStep">Step in bytes between any pair of sequential</param>
				/// <param name="size">Graph size</param>
				/// <param name="pLabel">Pointer to destination label image </param>
				/// <param name="nLabelStep">Step in bytes between any pair of sequential rows of</param>
				/// <param name="pState">Pointer to graph-cut state structure</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGraphcut_32s8u(CUdeviceptr pTerminals, CUdeviceptr pLeftTransposed, CUdeviceptr pRightTransposed, CUdeviceptr pTop, CUdeviceptr pBottom, int nStep, int nTransposedStep, NppiSize size, CUdeviceptr pLabel, int nLabelStep, NppiGraphcutState pState);


				/// <summary>
				/// Graphcut of a flow network (32bit signed integer edge capacities). The
				/// function computes the minimal cut (graphcut) of a 2D regular 8-connected
				/// graph. 
				/// The inputs are the capacities of the horizontal (in transposed form),
				/// vertical and terminal (source and sink) edges. The capacities to source and
				/// sink 
				/// are stored as capacity differences in the terminals array 
				/// ( terminals(x) = source(x) - sink(x) ). The implementation assumes that the
				/// edge capacities 
				/// for boundary edges that would connect to nodes outside the specified domain
				/// are set to 0 (for example left(0,*) == 0). If this is not fulfilled the
				/// computed labeling may be wrong!
				/// The computed binary labeling is encoded as unsigned 8bit values (0 and >0).
				/// <see cref="nppiGraphcut8GetSize"/>, 
				/// <see cref="nppiGraphcut8InitAlloc"/>
				/// </summary>
				/// <param name="pTerminals">Pointer to differences of terminal edge capacities</param>
				/// <param name="pLeftTransposed">Pointer to transposed left edge capacities</param>
				/// <param name="pRightTransposed">Pointer to transposed right edge capacities</param>
				/// <param name="pTop">Pointer to top edge capacities (top(*,0) must be 0)</param>
				/// <param name="pTopLeft">Pointer to top left edge capacities (topleft(*,0) </param>
				/// <param name="pTopRight">Pointer to top right edge capacities (topright(*,0)</param>
				/// <param name="pBottom">Pointer to bottom edge capacities (bottom(*,height-1)</param>
				/// <param name="pBottomLeft">Pointer to bottom left edge capacities </param>
				/// <param name="pBottomRight">Pointer to bottom right edge capacities </param>
				/// <param name="nStep">Step in bytes between any pair of sequential rows of edge</param>
				/// <param name="nTransposedStep">Step in bytes between any pair of sequential</param>
				/// <param name="size">Graph size</param>
				/// <param name="pLabel">Pointer to destination label image </param>
				/// <param name="nLabelStep">Step in bytes between any pair of sequential rows of</param>
				/// <param name="pState">Pointer to graph-cut state structure </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGraphcut8_32s8u(CUdeviceptr pTerminals, CUdeviceptr pLeftTransposed, CUdeviceptr pRightTransposed, CUdeviceptr pTop, CUdeviceptr pTopLeft, CUdeviceptr pTopRight, CUdeviceptr pBottom, CUdeviceptr pBottomLeft, CUdeviceptr pBottomRight, int nStep, int nTransposedStep, NppiSize size, CUdeviceptr pLabel, int nLabelStep, NppiGraphcutState pState);



				///// <summary>
				///// Calculates the size of the temporary buffer for graph-cut with 4 neighborhood labeling.
				///// \see nppiGraphcutInitAlloc(), nppiGraphcut_32s8u().
				///// </summary>
				///// <param name="oSize">Graph size.</param>
				///// <param name="pBufSize">Pointer to variable that returns the size of the</param>
				///// <returns>
				///// NPP_SUCCESS Indicates no error. Any other value indicates an error
				/////         or a warning
				///// \return NPP_SIZE_ERROR Indicates an error condition if any image dimension
				/////         has zero or negative value
				///// \return NPP_NULL_POINTER_ERROR Indicates an error condition if pBufSize
				/////         pointer is NULL
				///// </returns>
				//[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppiGraphcutGetSize(NppiSize oSize, ref int pBufSize);


				/// <summary>
				/// Graphcut of a flow network (32bit float edge capacities). The
				/// function computes the minimal cut (graphcut) of a 2D regular 4-connected
				/// graph. 
				/// The inputs are the capacities of the horizontal (in transposed form),
				/// vertical and terminal (source and sink) edges. The capacities to source and
				/// sink 
				/// are stored as capacity differences in the terminals array 
				/// ( terminals(x) = source(x) - sink(x) ). The implementation assumes that the
				/// edge capacities 
				/// for boundary edges that would connect to nodes outside the specified domain
				/// are set to 0 (for example left(0,*) == 0). If this is not fulfilled the
				/// computed labeling may be wrong!
				/// The computed binary labeling is encoded as unsigned 8bit values (0 and >0).
				/// \see nppiGraphcutInitAlloc(), nppiGraphcutFree(), nppiGraphcutGetSize().
				/// </summary>
				/// <param name="pTerminals">Pointer to differences of terminal edge capacities</param>
				/// <param name="pLeftTransposed">Pointer to transposed left edge capacities</param>
				/// <param name="pRightTransposed">Pointer to transposed right edge capacities</param>
				/// <param name="pTop">Pointer to top edge capacities (top(*,0) must be 0)</param>
				/// <param name="pBottom">Pointer to bottom edge capacities (bottom(*,height-1)</param>
				/// <param name="nStep">Step in bytes between any pair of sequential rows of edge</param>
				/// <param name="nTransposedStep">Step in bytes between any pair of sequential</param>
				/// <param name="size">Graph size</param>
				/// <param name="pLabel">Pointer to destination label image </param>
				/// <param name="nLabelStep">Step in bytes between any pair of sequential rows of</param>
				/// <param name="pState">Pointer to graph-cut state structure. This structure must be</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGraphcut_32f8u(CUdeviceptr pTerminals, CUdeviceptr pLeftTransposed, CUdeviceptr pRightTransposed, CUdeviceptr pTop, CUdeviceptr pBottom, int nStep, int nTransposedStep, NppiSize size, CUdeviceptr pLabel, int nLabelStep, NppiGraphcutState pState);


				/// <summary>
				/// Graphcut of a flow network (32bit float edge capacities). The
				/// function computes the minimal cut (graphcut) of a 2D regular 8-connected
				/// graph. 
				/// The inputs are the capacities of the horizontal (in transposed form),
				/// vertical and terminal (source and sink) edges. The capacities to source and
				/// sink 
				/// are stored as capacity differences in the terminals array 
				/// ( terminals(x) = source(x) - sink(x) ). The implementation assumes that the
				/// edge capacities 
				/// for boundary edges that would connect to nodes outside the specified domain
				/// are set to 0 (for example left(0,*) == 0). If this is not fulfilled the
				/// computed labeling may be wrong!
				/// The computed binary labeling is encoded as unsigned 8bit values (0 and >0).
				/// \see nppiGraphcut8InitAlloc(), nppiGraphcutFree(), nppiGraphcut8GetSize().
				/// </summary>
				/// <param name="pTerminals">Pointer to differences of terminal edge capacities</param>
				/// <param name="pLeftTransposed">Pointer to transposed left edge capacities</param>
				/// <param name="pRightTransposed">Pointer to transposed right edge capacities</param>
				/// <param name="pTop">Pointer to top edge capacities (top(*,0) must be 0)</param>
				/// <param name="pTopLeft">Pointer to top left edge capacities (topleft(*,0) </param>
				/// <param name="pTopRight">Pointer to top right edge capacities (topright(*,0)</param>
				/// <param name="pBottom">Pointer to bottom edge capacities (bottom(*,height-1)</param>
				/// <param name="pBottomLeft">Pointer to bottom left edge capacities </param>
				/// <param name="pBottomRight">Pointer to bottom right edge capacities </param>
				/// <param name="nStep">Step in bytes between any pair of sequential rows of edge</param>
				/// <param name="nTransposedStep">Step in bytes between any pair of sequential</param>
				/// <param name="size">Graph size</param>
				/// <param name="pLabel">Pointer to destination label image </param>
				/// <param name="nLabelStep">Step in bytes between any pair of sequential rows of</param>
				/// <param name="pState">Pointer to graph-cut state structure. This structure must be</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGraphcut8_32f8u(CUdeviceptr pTerminals, CUdeviceptr pLeftTransposed, CUdeviceptr pRightTransposed, CUdeviceptr pTop, CUdeviceptr pTopLeft, CUdeviceptr pTopRight, CUdeviceptr pBottom, CUdeviceptr pBottomLeft, CUdeviceptr pBottomRight, int nStep, int nTransposedStep, NppiSize size, CUdeviceptr pLabel, int nLabelStep, NppiGraphcutState pState);

			}

			/// <summary>
			/// RGB to YUV Conversion
			/// </summary>
			public static class RGBToYUV
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned packed YUV color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 4 channel 8-bit unsigned packed YUV color conversion with alpha, not affecting alpha.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar RGB to 3 channel 8-bit unsigned planar YUV color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned planar YUV color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 4 channel 8-bit unsigned planar YUV color conversion with alpha.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV_8u_AC4P4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// YUV to RGB Conversion
			/// </summary>
			public static class YUVToRGB
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed YUV to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUVToRGB_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit packed YUV with alpha to 4 channel 8-bit unsigned packed RGB color conversion with alpha, not affecting alpha.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUVToRGB_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YUV to 3 channel 8-bit unsigned planar RGB color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUVToRGB_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YUV to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUVToRGB_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// RGB To YUV422 Conversion
			/// </summary>
			public static class RGBToYUV422
			{

				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 2 channel 8-bit unsigned packed YUV422 color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV422_8u_C3C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar RGB to 3 channel 8-bit unsigned planar YUV422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV422_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned planar YUV422 color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV422_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// YUV422 To RGB Conversion
			/// </summary>
			public static class YUV422ToRGB
			{
				/// <summary>
				/// 2 channel 8-bit unsigned packed YUV422 to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUV422ToRGB_8u_C2C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YUV422 to 3 channel 8-bit unsigned planar RGB color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUV422ToRGB_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YUV422 to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUV422ToRGB_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YUV422 to 4 channel 8-bit unsigned packed RGB color conversion with alpha.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUV422ToRGB_8u_P3AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// RGB to YUV420 Conversion
			/// </summary>
			public static class RGBToYUV420
			{
				/// <summary>
				/// 3 channel 8-bit unsigned planar RGB to 3 channel 8-bit unsigned planar YUV420 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV420_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned planar YUV420 color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYUV420_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// YUV420 to RGB Conversion
			/// </summary>
			public static class YUV420ToRGB
			{
				/// <summary>
				/// 3 channel 8-bit unsigned planar YUV420 to 3 channel 8-bit unsigned planar RGB color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUV420ToRGB_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YUV420 to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUV420ToRGB_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YUV420 to 4 channel 8-bit unsigned packed RGB color conversion with alpha.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUV420ToRGB_8u_P3AC4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);
 
			}

			/// <summary>
			/// BGR To YUV420 Conversion
			/// </summary>
			public static class BGRToYUV420
			{
				/// <summary>
				/// 4 channel 8-bit unsigned pacmed BGR with alpha to 3 channel 8-bit unsigned planar YUV420 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYUV420_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);
			}

			/// <summary>
			/// YUV420 to BGR Conversion
			/// </summary>
			public static class YUV420ToBGR
			{
				/// <summary>
				/// 3 channel 8-bit unsigned planar YUV420 to 3 channel 8-bit unsigned packed BGR color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYUV420ToBGR_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// YCbCr To BGR Conversion
			/// </summary>
			public static class YCbCrToBGR
			{
				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr to 3 channel 8-bit unsigned packed BGR color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCrToBGR_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr to 4 channel 8-bit unsigned packed BGR color conversion with constant alpha.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCrToBGR_8u_P3C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr to 3 channel 8-bit unsigned packed BGR_709CSC color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCrToBGR_709CSC_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr to 4 channel 8-bit unsigned packed BGR_709CSC color conversion with constant alpha.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCrToBGR_709CSC_8u_P3C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);

				/// <summary>
				/// 2 channel 8-bit unsigned packed YCrCb422 to 3 channel 8-bit unsigned packed BGR color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToBGR_8u_C2C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCrCb422 to 4 channel 8-bit unsigned packed BGR color conversion with constant alpha.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToBGR_8u_C2C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr422 to 3 channel 8-bit unsigned packed BGR color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToBGR_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr420 to 3 channel 8-bit unsigned packed BGR color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToBGR_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr420 to 4 channel 8-bit unsigned packed BGR color conversion with constant alpha.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToBGR_8u_P3C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr420 to 3 channel 8-bit unsigned packed BGR_709CSC color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToBGR_709CSC_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr420 to 4 channel 8-bit unsigned packed BGR_709HDTV color conversion with constant alpha.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToBGR_709HDTV_8u_P3C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);

				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr411 to 3 channel 8-bit unsigned packed BGR color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToBGR_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr411 to 4 channel 8-bit unsigned packed BGR color conversion with constant alpha.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToBGR_8u_P3C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);

			}

			/// <summary>
			/// BGR to YCbCr Conversion
			/// </summary>
			public static class BGRToYCbCr
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 2 channel 8-bit unsigned packed YCrCb422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr422_8u_C3C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 2 channel 8-bit unsigned packed YCrCb422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr422_8u_AC4C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 3 channel 8-bit unsigned planar YCbCr422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr422_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 3 channel 8-bit unsigned planar YCbCr422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr422_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 3 channel 8-bit unsigned planar YCbCr420 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr420_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 3 channel 8-bit unsigned planar YCbCr420 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr420_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 3 channel 8-bit unsigned planar YCbCr420_709CSC color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr420_709CSC_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 3 channel 8-bit unsigned planar YCbCr420_709CSC color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr420_709CSC_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 3 channel 8-bit unsigned planar YCbCr420_709HDTV color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr420_709HDTV_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 3 channel 8-bit unsigned planar YCbCr411 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr411_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 3 channel 8-bit unsigned planar YCbCr411 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCbCr411_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// RGB To CbYCr Conversion
			/// </summary>
			public static class RGBToCbYCr
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 2 channel 8-bit unsigned packed CbYCr422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToCbYCr422_8u_C3C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB first gets forward gamma corrected then converted to 2 channel 8-bit unsigned packed CbYCr422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToCbYCr422Gamma_8u_C3C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// CbYCr to RGB Conversion
			/// </summary>
			public static class CbYCrToRGB
			{
				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCrC22 to 3 channel 8-bit unsigned packed RGB color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToRGB_8u_C2C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);
 
			}

			/// <summary>
			/// BGR To CbYCr Conversion
			/// </summary>
			public static class BGRToCbYCr
			{
				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 2 channel 8-bit unsigned packed CbYCr422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToCbYCr422_8u_AC4C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 2 channel 8-bit unsigned packed CbYCr422_709HDTV color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToCbYCr422_709HDTV_8u_C3C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 2 channel 8-bit unsigned packed CbYCr422_709HDTV color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToCbYCr422_709HDTV_8u_AC4C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// CbYCrToBGR Conversion
			/// </summary>
			public static class CbYCrToBGR
			{
				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCr422 to 4 channel 8-bit unsigned packed BGR color conversion with alpha.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToBGR_8u_C2C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);


				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCr422 to 3 channel 8-bit unsigned packed BGR_709HDTV color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToBGR_709HDTV_8u_C2C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCr422 to 4 channel 8-bit unsigned packed BGR_709HDTV color conversion with constant alpha.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToBGR_709HDTV_8u_C2C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);

			}

			/// <summary>
			/// RGB To YCrCB Conversion
			/// </summary>
			public static class RGBToYCrCb
			{
				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 3 channel 8-bit unsigned planar YCrCb420 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCrCb420_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 2 channel 8-bit unsigned packed YCrCb422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCrCb422_8u_C3C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar RGB to 2 channel 8-bit unsigned packed YCrCb422 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCrCb422_8u_P3C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);
			}

			/// <summary>
			/// YCrCB to RGB Conversion
			/// </summary>
			public static class YCrCbToRGB
			{
				/// <summary>
				/// 3 channel 8-bit unsigned planar YCrCb420 to 4 channel 8-bit unsigned packed RGB color conversion with constant alpha.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nAval">8-bit unsigned alpha constant.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb420ToRGB_8u_P3C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nAval);

			}
			
			/// <summary>
			/// BGR to YCrCb Conversion
			/// </summary>
			public static class BGRToYCrCb
			{

				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 3 channel 8-bit unsigned planar YCrCb420_709CSC color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCrCb420_709CSC_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 3 channel 8-bit unsigned planar YCrCb420_709CSC color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCrCb420_709CSC_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 3 channel 8-bit unsigned planar YCrCb420 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCrCb420_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 3 channel 8-bit unsigned planar YCrCb420 color conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToYCrCb420_8u_AC4P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// RGB to XYZ
			/// </summary>
			public static class RGBToXYZ
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned packed XYZ color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToXYZ_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 4 channel 8-bit unsigned packed XYZ with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToXYZ_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// XYZ to RGB
			/// </summary>
			public static class XYZToRGB
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed XYZ to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXYZToRGB_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed XYZ with alpha to 4 channel 8-bit unsigned packed RGB with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiXYZToRGB_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// RGB to LUV
			/// </summary>
			public static class RGBToLUV
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned packed LUV color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToLUV_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 4 channel 8-bit unsigned packed LUV with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToLUV_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// LUV to RGB
			/// </summary>
			public static class LUVToRGB
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed LUV to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUVToRGB_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed LUV with alpha to 4 channel 8-bit unsigned packed RGB with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUVToRGB_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// BGR to LAB
			/// </summary>
			public static class BGRToLab
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 3 channel 8-bit unsigned packed Lab color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToLab_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// LAB to BGR
			/// </summary>
			public static class LabToBGR
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed Lab to 3 channel 8-bit unsigned packed BGR color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLabToBGR_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);
			}

			/// <summary>
			/// RGB to YCC
			/// </summary>
			public static class RGBToYCC
			{

				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned packed YCC color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCC_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 4 channel 8-bit unsigned packed YCC with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToYCC_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// YCC to RGB
			/// </summary>
			public static class YCCToRGB
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed YCC to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCCToRGB_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed YCC with alpha to 4 channel 8-bit unsigned packed RGB with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCCToRGB_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// RGB to HLS
			/// </summary>
			public static class RGBToHLS
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned packed HLS color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToHLS_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 4 channel 8-bit unsigned packed HLS with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToHLS_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// HLS to RGB
			/// </summary>
			public static class HLSToRGB
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed HLS to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHLSToRGB_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed HLS with alpha to 4 channel 8-bit unsigned packed RGB with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHLSToRGB_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// BGR to HLS
			/// </summary>
			public static class BGRToHLS
			{
				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 4 channel 8-bit unsigned packed HLS with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToHLS_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed BGR to 3 channel 8-bit unsigned planar HLS color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToHLS_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed BGR with alpha to 4 channel 8-bit unsigned planar HLS with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToHLS_8u_AC4P4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar BGR to 3 channel 8-bit unsigned packed HLS color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToHLS_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned planar BGR with alpha to 4 channel 8-bit unsigned packed HLS with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToHLS_8u_AP4C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar BGR to 3 channel 8-bit unsigned planar HLS color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToHLS_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned planar BGR with alpha to 4 channel 8-bit unsigned planar HLS with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiBGRToHLS_8u_AP4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// HLS to BGR
			/// </summary>
			public static class HLSToBGR
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed HLS to 3 channel 8-bit unsigned planar BGR color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHLSToBGR_8u_C3P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed HLS with alpha to 4 channel 8-bit unsigned planar BGR with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHLSToBGR_8u_AC4P4R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar HLS to 3 channel 8-bit unsigned planar BGR color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHLSToBGR_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned planar HLS with alpha to 4 channel 8-bit unsigned planar BGR with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHLSToBGR_8u_AP4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed HLS with alpha to 4 channel 8-bit unsigned packed BGR with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHLSToBGR_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar HLS to 3 channel 8-bit unsigned packed BGR color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHLSToBGR_8u_P3C3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned planar HLS with alpha to 4 channel 8-bit unsigned packed BGR with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHLSToBGR_8u_AP4C4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// RGB to HSV
			/// </summary>
			public static class RGBToHSV
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 3 channel 8-bit unsigned packed HSV color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToHSV_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 4 channel 8-bit unsigned packed HSV with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToHSV_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// HSV to RGB
			/// </summary>
			public static class HSVToRGB
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed HSV to 3 channel 8-bit unsigned packed RGB color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHSVToRGB_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed HSV with alpha to 4 channel 8-bit unsigned packed RGB with alpha color conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiHSVToRGB_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

			}

			/// <summary>
			/// YCbCr ACrCb ...
			/// </summary>
			public static class YCbCrAndACrCbAndOther
			{
				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr420 to 3 channel 8-bit unsigned planar YCbCr411 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToYCbCr411_8u_P2P3R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCbCr422 to 3 channel 8-bit unsigned planar YCbCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr422 to 2 channel 8-bit unsigned packed YCbCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422_8u_P3C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCbCr422 to 2 channel 8-bit unsigned packed YCrCb422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCrCb422_8u_C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr422 to 2 channel 8-bit unsigned packed YCrCb422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCrCb422_8u_P3C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCbCr422 to 2 channel 8-bit unsigned packed CbYCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToCbYCr422_8u_C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCr422 to 3 channel 8-bit unsigned planar YCbCr411 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToYCbCr411_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr422 to 2 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCbCr420_8u_P3P2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCbCr422 to 3 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCbCr420_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCbCr422 to 2 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCbCr420_8u_C2P2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCrCb420 to 3 channel 8-bit unsigned planar YCbCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb420ToYCbCr422_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCrCb420 to 2 channel 8-bit unsigned packed YCbCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb420ToYCbCr422_8u_P3C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCbCr422 to 3 channel 8-bit unsigned planar YCrCb420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCrCb420_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr422 to 2 channel 8-bit unsigned planar YCbCr411 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCbCr411_8u_P3P2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCbCr422 to 3 channel 8-bit unsigned planar YCbCr411 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCbCr411_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCbCr422 to 2 channel 8-bit unsigned planar YCbCr411 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr422ToYCbCr411_8u_C2P2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCrCb422 to 3 channel 8-bit unsigned planar YCbCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb422ToYCbCr422_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCrCb422 to 3 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb422ToYCbCr420_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed YCrCb422 to 3 channel 8-bit unsigned planar YCbCr411 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb422ToYCbCr411_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCr422 to 2 channel 8-bit unsigned packed YCbCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToYCbCr422_8u_C2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCr422 to 3 channel 8-bit unsigned planar YCbCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToYCbCr422_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCr422 to 3 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToYCbCr420_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCr422 to 2 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToYCbCr420_8u_C2P2R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned packed CbYCr422 to 3 channel 8-bit unsigned planar YCrCb420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCbYCr422ToYCrCb420_8u_C2P3R(CUdeviceptr pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr420 to 2 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420_8u_P3P2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr420 to 3 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420_8u_P2P3R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr420 to 3 channel 8-bit unsigned planar YCbCr422 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToYCbCr422_8u_P2P3R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr420 to 2 channel 8-bit unsigned packed YCbCr422 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToYCbCr422_8u_P2C2R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr420 to 2 channel 8-bit unsigned packed CbYCr422 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToCbYCr422_8u_P2C2R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr420 to 3 channel 8-bit unsigned planar YCrCb420 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr420ToYCrCb420_8u_P2P3R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCrCb420 to 2 channel 8-bit unsigned packed CbYCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb420ToCbYCr422_8u_P3C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCrCb420 to 2 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb420ToYCbCr420_8u_P3P2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCrCb420 to 2 channel 8-bit unsigned planar YCbCr411 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCrCb420ToYCbCr411_8u_P3P2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr411 to 2 channel 8-bit unsigned planar YCbCr411 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411_8u_P3P2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr411 to 3 channel 8-bit unsigned planar YCbCr411 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411_8u_P2P3R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr411 to 3 channel 8-bit unsigned planar YCbCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCbCr422_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr411 to 2 channel 8-bit unsigned packed YCbCr422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCbCr422_8u_P3C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr411 to 3 channel 8-bit unsigned planar YCbCr422 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCbCr422_8u_P2P3R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr411 to 2 channel 8-bit unsigned packed YCbCr422 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCbCr422_8u_P2C2R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr411 to 3 channel 8-bit unsigned planar YCrCb422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCrCb422_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr411 to 2 channel 8-bit unsigned packed YCrCb422 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCrCb422_8u_P3C2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr411 to 3 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCbCr420_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar YCbCr411 to 2 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// images.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="rSrcStep">\ref source_planar_image_line_step_array.</param>
				/// <param name="pDstY">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstYStep">\ref destination_planar_image_line_step.</param>
				/// <param name="pDstCbCr">\ref destination_planar_image_pointer.</param>
				/// <param name="nDstCbCrStep">\ref destination_planar_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCbCr420_8u_P3P2R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rSrcStep, CUdeviceptr pDstY, int nDstYStep, CUdeviceptr pDstCbCr, int nDstCbCrStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr411 to 3 channel 8-bit unsigned planar YCbCr420 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCbCr420_8u_P2P3R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 2 channel 8-bit unsigned planar YCbCr411 to 3 channel 8-bit unsigned planar YCrCb420 sampling format conversion.
				/// </summary>
				/// <param name="pSrcY">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcYStep">\ref source_planar_image_line_step.</param>
				/// <param name="pSrcCbCr">\ref source_planar_image_pointer.</param>
				/// <param name="nSrcCbCrStep">\ref source_planar_image_line_step.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="rDstStep">\ref destination_planar_image_line_step_array.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiYCbCr411ToYCrCb420_8u_P2P3R(CUdeviceptr pSrcY, int nSrcYStep, CUdeviceptr pSrcCbCr, int nSrcCbCrStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] rDstStep, NppiSize oSizeROI);

			}
			
			/// <summary>
			/// Gamma correction
			/// </summary>
			public static class Gamma
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed color not in place forward gamma correction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaFwd_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed color in place forward gamma correction.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaFwd_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed color with alpha not in place forward gamma correction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaFwd_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed color with alpha in place forward gamma correction.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaFwd_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar color not in place forward gamma correction.
				/// </summary>
				/// <param name="pSrc">source planar pixel format image pointer array.</param>
				/// <param name="nSrcStep">source planar pixel format image line step.</param>
				/// <param name="pDst">destination planar pixel format image pointer array.</param>
				/// <param name="nDstStep">destination planar pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaFwd_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar color in place forward gamma correction.
				/// </summary>
				/// <param name="pSrcDst">in place planar pixel format image pointer array.</param>
				/// <param name="nSrcDstStep">in place planar pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaFwd_8u_IP3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed color not in place inverse gamma correction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaInv_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned packed color in place inverse gamma correction.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaInv_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed color with alpha not in place inverse gamma correction.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaInv_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed color with alpha in place inverse gamma correction.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaInv_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar color not in place inverse gamma correction.
				/// </summary>
				/// <param name="pSrc">source planar pixel format image pointer array.</param>
				/// <param name="nSrcStep">source planar pixel format image line step.</param>
				/// <param name="pDst">destination planar pixel format image pointer array.</param>
				/// <param name="nDstStep">destination planar pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaInv_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 8-bit unsigned planar color in place inverse gamma correction.
				/// </summary>
				/// <param name="pSrcDst">in place planar pixel format image pointer array.</param>
				/// <param name="nSrcDstStep">in place planar pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGammaInv_8u_IP3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrcDst, int nSrcDstStep, NppiSize oSizeROI);
			}

			/// <summary>
			/// Composition color key
			/// </summary>
			public static class CompColorKey
			{
				/// <summary>
				/// 1 channel 8-bit unsigned packed color complement color key replacement of source image 1 by source image 2.
				/// </summary>
				/// <param name="pSrc1">source1 packed pixel format image pointer.</param>
				/// <param name="nSrc1Step">source1 packed pixel format image line step.</param>
				/// <param name="pSrc2">source2 packed pixel format image pointer.</param>
				/// <param name="nSrc2Step">source2 packed pixel format image line step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nColorKeyConst">color key constant</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompColorKey_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, byte nColorKeyConst);


				/// <summary>
				/// 3 channel 8-bit unsigned packed color complement color key replacement of source image 1 by source image 2.
				/// </summary>
				/// <param name="pSrc1">source1 packed pixel format image pointer.</param>
				/// <param name="nSrc1Step">source1 packed pixel format image line step.</param>
				/// <param name="pSrc2">source2 packed pixel format image pointer.</param>
				/// <param name="nSrc2Step">source2 packed pixel format image line step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nColorKeyConst">color key constant array</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompColorKey_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] nColorKeyConst);


				/// <summary>
				/// 4 channel 8-bit unsigned packed color complement color key replacement of source image 1 by source image 2.
				/// </summary>
				/// <param name="pSrc1">source1 packed pixel format image pointer.</param>
				/// <param name="nSrc1Step">source1 packed pixel format image line step.</param>
				/// <param name="pSrc2">source2 packed pixel format image pointer.</param>
				/// <param name="nSrc2Step">source2 packed pixel format image line step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nColorKeyConst">color key constant array</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCompColorKey_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] nColorKeyConst);


				/// <summary>
				/// 4 channel 8-bit unsigned packed color complement color key replacement of source image 1 by source image 2 with alpha blending.
				/// </summary>
				/// <param name="pSrc1">source1 packed pixel format image pointer.</param>
				/// <param name="nSrc1Step">source1 packed pixel format image line step.</param>
				/// <param name="nAlpha1">source1 image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pSrc2">source2 packed pixel format image pointer.</param>
				/// <param name="nSrc2Step">source2 packed pixel format image line step.</param>
				/// <param name="nAlpha2">source2 image alpha opacity (0 - max channel pixel value).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nColorKeyConst">color key constant array</param>
				/// <param name="nppAlphaOp">NppiAlphaOp alpha compositing operation selector  (excluding premul ops).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiAlphaCompColorKey_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, byte nAlpha1, CUdeviceptr pSrc2, int nSrc2Step, byte nAlpha2, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] byte[] nColorKeyConst, NppiAlphaOp nppAlphaOp);

			}

			/// <summary>
			/// Minimum
			/// </summary>
			public static class Min
			{
				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMin_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image min. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMin);


				/// <summary>
				/// 1-channel 16-bit unsigned short integer image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMin);


				/// <summary>
				/// 1-channel 16-bit signed short integer image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMin);


				/// <summary>
				/// 1-channel 32-bit floating point image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMin);


				/// <summary>
				/// 3-channel 8-bit unsigned char image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum results, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 3-channel 16-bit unsigned short integer image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum results, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 3-channel 16-bit signed short integer image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum results, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 3-channel 32-bit floating point image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum results, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 4-channel 8-bit unsigned char image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 4-channel 16-bit unsigned short integer image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 4-channel 16-bit signed short integer image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 4-channel 32-bit floating point image min.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 4-channel 8-bit unsigned char image min (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 4-channel 16-bit unsigned short integer image min (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 4-channel 16-bit signed short integer image min (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);


				/// <summary>
				/// 4-channel 32-bit floating point image min (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMin_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin);



			}

			/// <summary>
			/// Minimum index
			/// </summary>
			public static class MinIdx
			{ 		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinIndx_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndxGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image min with its X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <param name="pIndexX">Device-memory pointer to the X coordinate of the image min value.</param>
				/// <param name="pIndexY">Device-memory pointer to the Y coordinate of the image min value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMin, CUdeviceptr pIndexX, CUdeviceptr pIndexY);


				/// <summary>
				/// 1-channel 16-bit unsigned short integer image min with its X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <param name="pIndexX">Device-memory pointer to the X coordinate of the image min value.</param>
				/// <param name="pIndexY">Device-memory pointer to the Y coordinate of the image min value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMin, CUdeviceptr pIndexX, CUdeviceptr pIndexY);


				/// <summary>
				/// 1-channel 16-bit signed short integer image min with its X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <param name="pIndexX">Device-memory pointer to the X coordinate of the image min value.</param>
				/// <param name="pIndexY">Device-memory pointer to the Y coordinate of the image min value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMin, CUdeviceptr pIndexX, CUdeviceptr pIndexY);


				/// <summary>
				/// 1-channel 32-bit floating point image min with its X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <param name="pIndexX">Device-memory pointer to the X coordinate of the image min value.</param>
				/// <param name="pIndexY">Device-memory pointer to the Y coordinate of the image min value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMin, CUdeviceptr pIndexX, CUdeviceptr pIndexY);


				/// <summary>
				/// 3-channel 8-bit unsigned char image min values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, three elements for three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 3-channel 16-bit unsigned short integer image min values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, three elements for three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 3-channel 16-bit signed short integer image min values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, three elements for three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 3-channel 32-bit floating point image min values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, three elements for three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 8-bit unsigned char image min values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, four elements for four channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 16-bit unsigned short integer image min values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, four elements for four channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 16-bit signed short integer image min values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, four elements for four channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 32-bit floating point image min values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, four elements for four channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 8-bit unsigned char image min values with their X and Y coordinates (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, three elements for the three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, three elements for the three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 16-bit unsigned short integer image min values with their X and Y coordinates (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, three elements for the three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, three elements for the three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 16-bit signed short integer image min values with their X and Y coordinates (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, three elements for the three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, three elements for the three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 32-bit floating point image min values with their X and Y coordinates (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image min values, three elements for the three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image min values, three elements for the three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinIndx_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMin, CUdeviceptr aIndexX, CUdeviceptr aIndexY);



			}

			/// <summary>
			/// Maximum
			/// </summary>
			public static class Max
			{ 		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMax_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMax);


				/// <summary>
				/// 1-channel 16-bit unsigned short integer image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMax);


				/// <summary>
				/// 1-channel 16-bit signed short integer image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMax);


				/// <summary>
				/// 1-channel 32-bit floating point image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMax);


				/// <summary>
				/// 3-channel 8-bit unsigned char image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum results, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 3-channel 16-bit unsigned short integer image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum results, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 3-channel 16-bit signed short integer image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum results, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 3-channel 32-bit floating point image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum results, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 4-channel 8-bit unsigned char image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum results, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 4-channel 16-bit unsigned short integer image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum results, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 4-channel 16-bit signed short integer image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum results, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 4-channel 32-bit floating point image max.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum results, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 4-channel 8-bit unsigned char image max (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 4-channel 16-bit unsigned short integer image max (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 4-channel 16-bit signed short integer image max (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


				/// <summary>
				/// 4-channel 32-bit floating point image max (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMax_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax);


			}

			/// <summary>
			/// Maximum Index
			/// </summary>
			public static class MaxIdx
			{ 		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMaxIndx_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndxGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image max value with its X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <param name="pIndexX">Device-memory pointer to the X coordinate of the image max value.</param>
				/// <param name="pIndexY">Device-memory pointer to the Y coordinate of the image max value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMax, CUdeviceptr pIndexX, CUdeviceptr pIndexY);


				/// <summary>
				/// 1-channel 16-bit unsigned short integer image max value with its X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <param name="pIndexX">Device-memory pointer to the X coordinate of the image max value.</param>
				/// <param name="pIndexY">Device-memory pointer to the Y coordinate of the image max value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMax, CUdeviceptr pIndexX, CUdeviceptr pIndexY);


				/// <summary>
				/// 1-channel 16-bit signed short integer image max value with its X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <param name="pIndexX">Device-memory pointer to the X coordinate of the image max value.</param>
				/// <param name="pIndexY">Device-memory pointer to the Y coordinate of the image max value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMax, CUdeviceptr pIndexX, CUdeviceptr pIndexY);


				/// <summary>
				/// 1-channel 32-bit floating point image max value with its X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <param name="pIndexX">Device-memory pointer to the X coordinate of the image max value.</param>
				/// <param name="pIndexY">Device-memory pointer to the Y coordinate of the image max value.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMax, CUdeviceptr pIndexX, CUdeviceptr pIndexY);


				/// <summary>
				/// 3-channel 8-bit unsigned char image max values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, three elements for three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 3-channel 16-bit unsigned short integer image max values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, three elements for three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 3-channel 16-bit signed short integer image max values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, three elements for three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 3-channel 32-bit floating point image max values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, three elements for three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 8-bit unsigned char image max valueswith their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, four elements for four channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, four elements for four channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 16-bit unsigned short integer image max values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, four elements for four channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, four elements for four channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 16-bit signed short integer image max values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, four elements for four channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, four elements for four channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 32-bit floating point image max values with their X and Y coordinates.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, four elements for four channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, four elements for four channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 8-bit unsigned char image max values with their X and Y coordinates (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, three elements for the first three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 16-bit unsigned short integer image max values with their X and Y coordinates (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, three elements for the first three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 16-bit signed short integer image max values with their X and Y coordinates (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, three elements for the first three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


				/// <summary>
				/// 4-channel 32-bit floating point image max values with their X and Y coordinates (alpha channel is not processed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <param name="aIndexX">Device-memory array to the X coordinates of the image max values, three elements for the first three channels.</param>
				/// <param name="aIndexY">Device-memory array to the Y coordinates of the image max values, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxIndx_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMax, CUdeviceptr aIndexX, CUdeviceptr aIndexY);


			}

			/// <summary>
			/// Min / Max (new in CUDA 5)
			/// </summary>
			public static class MinMaxNew
			{ 		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinManx_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit unsigned short image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_16u_C1R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit signed short image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_16s_C1R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit floating point image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMin">Device-memory pointer receiving the minimum result.</param>
				/// <param name="pMax">Device-memory pointer receiving the maximum result.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_32f_C1R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit unsigned image minimum and maximum  values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for three channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for three channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_8u_C3R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit unsigned short image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for three channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for three channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_16u_C3R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit signed short image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for three channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for three channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_16s_C3R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 32-bit floating point image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for three channels..</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for three channels..</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_32f_C3R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 8-bit unsigned image minimum and maximum values (alpha channel is not calculated).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the first three channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_8u_AC4R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit unsigned short image minimum and maximum values (alpha channel is not calculated).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the first three channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_16u_AC4R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit signed short image minimum and maximum values (alpha channel is not calculated).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the first three channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_16s_AC4R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 32-bit floating point image minimum and maximum values (alpha channel is not calculated).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, three elements for the first three channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, three elements for the first three channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_32f_AC4R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 8-bit unsigned image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, four elements for four channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_8u_C4R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit unsigned short image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, four elements for four channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_16u_C4R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit signed short image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, four elements for four channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_16s_C4R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 32-bit floating point image minimum and maximum values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aMin">Device-memory array receiving the minimum result, four elements for four channels.</param>
				/// <param name="aMax">Device-memory array receiving the maximum result, four elements for four channels.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxGetBufferHostSize_32f_C4R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMax_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aMin, CUdeviceptr aMax, CUdeviceptr pDeviceBuffer);



			}

			/// <summary>
			/// Min / Max Index (new in CUDA 5)
			/// </summary>
			public static class MinMaxIndxNew
			{ 		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_8s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_8s_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_8u_C3CR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_8u_C3CR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_8s_C3CR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_8s_C3CR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_16u_C3CR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_16u_C3CR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_32f_C3CR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_32f_C3CR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMinMaxIndx_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndxGetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image minimum and maximum values with their indices.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_8u_C1R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>. 
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 8-bit signed char image minimum and maximum values with their indices.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_8s_C1R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>. 
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_8s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit unsigned short image minimum and maximum values with their indices.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_16u_C1R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>. 
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit floating point image minimum and maximum values with their indices.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_32f_C1R to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point image is specified.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 8-bit unsigned char image minimum and maximum values with their indices, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_8u_C1MR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>. 
				/// If the mask is filled with zeros, then all the returned values are zeros, i.e., pMinIndex = {0, 0}, pMaxIndex = {0, 0},
				/// pMinValue = 0, pMaxValue = 0.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_8u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 8-bit signed char image minimum and maximum values with their indices, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_8s_C1MR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>. 
				/// If the mask is filled with zeros, then all the returned values are zeros, i.e., pMinIndex = {0, 0}, pMaxIndex = {0, 0},
				/// pMinValue = 0, pMaxValue = 0.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_8s_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit unsigned short image minimum and maximum values with their indices, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_16u_C1MR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>. 
				/// If the mask is filled with zeros, then all the returned values are zeros, i.e., pMinIndex = {0, 0}, pMaxIndex = {0, 0},
				/// pMinValue = 0, pMaxValue = 0.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_16u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit floating point image minimum and maximum values with their indices, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_32f_C1MR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point image is specified.
				/// If the mask is filled with zeros, then all the returned values are zeros, i.e., pMinIndex = {0, 0}, pMaxIndex = {0, 0},
				/// pMinValue = 0, pMaxValue = 0.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_32f_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit unsigned char image minimum and maximum values with their indices, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_8u_C3CR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_8u_C3CR(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit signed char image minimum and maximum values with their indices, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_8s_C3CR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_8s_C3CR(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit unsigned short image minimum and maximum values with their indices, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_16u_C3CR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_16u_C3CR(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit floating point image minimum and maximum values with their indices, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_32f_C3CR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point image is specified, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_32f_C3CR(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit unsigned char image minimum and maximum values with their indices, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_8u_C3CMR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// If the mask is filled with zeros, then all the returned values are zeros, i.e., pMinIndex = {0, 0}, pMaxIndex = {0, 0},
				/// pMinValue = 0, pMaxValue = 0.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_8u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit signed char image minimum and maximum values with their indices, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_8s_C3CMR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// If the mask is filled with zeros, then all the returned values are zeros, i.e., pMinIndex = {0, 0}, pMaxIndex = {0, 0},
				/// pMinValue = 0, pMaxValue = 0.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_8s_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit unsigned short image minimum and maximum values with their indices, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_16u_C3CMR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// If the mask is filled with zeros, then all the returned values are zeros, i.e., pMinIndex = {0, 0}, pMaxIndex = {0, 0},
				/// pMinValue = 0, pMaxValue = 0.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_16u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 32-bit floating point image minimum and maximum values with their indices, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pMinValue">Device-memory pointer receiving the minimum value.</param>
				/// <param name="pMaxValue">Device-memory pointer receiving the maximum value.</param>
				/// <param name="pMinIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the minimum value.</param>
				/// <param name="pMaxIndex">Device-memory pointer receiving the indicies (X and Y coordinates) of the maximum value.</param>
				/// <param name="pDeviceBuffer">Buffer to a scratch memory. Use \ref nppiMinMaxIndxGetBufferHostSize_32f_C3CMR to determine</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point image is specified, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// If the mask is filled with zeros, then all the returned values are zeros, i.e., pMinIndex = {0, 0}, pMaxIndex = {0, 0},
				/// pMinValue = 0, pMaxValue = 0.
				/// If any of pMinValue, pMaxValue, pMinIndex, or pMaxIndex is not needed, zero pointer must be passed correspondingly.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinMaxIndx_32f_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pMinValue, CUdeviceptr pMaxValue, CUdeviceptr pMinIndex, CUdeviceptr pMaxIndex, CUdeviceptr pDeviceBuffer);



			}

			/// <summary>
			/// Mean (new in CUDA 5)
			/// </summary>
			public static class MeanNew
			{		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanGetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image sum with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 1-channel 16-bit unsigned short integer image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 1-channel 16-bit signed short integer image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 1-channel 32-bit floating point image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 3-channel 8-bit unsigned char image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 3-channel 16-bit unsigned short image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 3-channel 16-bit signed short image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, three elements for three channels.</param>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 3-channel 32-bit floating point image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, three elements for three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 4-channel 8-bit unsigned char image mean with 64-bit double precision result.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 4-channel 16-bit unsigned short image mean with 64-bit double precision result.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 4-channel 16-bit signed short image mean with 64-bit double precision result.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 4-channel 32-bit floating point image mean with 64-bit double precision result.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, three elements for the first three channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 4-channel 8-bit unsigned char image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 4-channel 16-bit unsigned short image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 4-channel 16-bit signed short image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 4-channel 32-bit floating point image mean with 64-bit double precision result.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="aMean">Array that contains computed mean, four elements for four channels.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr aMean);


				/// <summary>
				/// 1-channel 8-bit unsigned char image mean with 64-bit double precision result, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_8u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 1-channel 8-bit signed char image mean with 64-bit double precision result, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_8s_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 1-channel 16-bit unsigned short integer image mean with 64-bit double precision result, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 1-channel 32-bit floating point image mean with 64-bit double precision result, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_32f_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 3-channel 8-bit unsigned char image mean with 64-bit double precision result, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_8u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 3-channel 8-bit signed char image mean with 64-bit double precision result, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_8s_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 3-channel 16-bit unsigned short integer image mean with 64-bit double precision result, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_16u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


				/// <summary>
				/// 3-channel 32-bit floating point image mean with 64-bit double precision result, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Device-memory pointer receiving the mean result.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_32f_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean);


 
			}

			/// <summary>
			/// Mean + Std deviation (new in CUDA 5)
			/// </summary>
			public static class MeanStdDevNew
			{		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_8s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_8s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_8u_C3CR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_8u_C3CR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_8s_C3CR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_8s_C3CR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_16u_C3CR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_16u_C3CR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_32f_C3CR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_32f_C3CR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiMean_StdDev_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMeanStdDevGetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image mean and standard deviation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 1-channel 8-bit signed char image mean and standard deviation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_8s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 1-channel 16-bit unsigned short int image mean and standard deviation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 1-channel 32-bit floating-point image mean and standard deviation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 1-channel 8-bit unsigned char image mean and standard deviation, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_8u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 1-channel 8-bit signed char image mean and standard deviation, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_8s_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 1-channel 16-bit unsigned short int image mean and standard deviation, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_16u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 1-channel 32-bit floating-point image mean and standard deviation, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_32f_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 3-channel 8-bit unsigned char image mean and standard deviation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_8u_C3CR(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 3-channel 8-bit signed char image mean and standard deviation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_8s_C3CR(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 3-channel 16-bit unsigned short int image mean and standard deviation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_16u_C3CR(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 3-channel 32-bit floating-point image mean and standard deviation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_32f_C3CR(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 3-channel 8-bit unsigned char image mean and standard deviation, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_8u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 3-channel 8-bit signed char image mean and standard deviation, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_8s_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 3-channel 16-bit unsigned short int image mean and standard deviation, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_16u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


				/// <summary>
				/// 3-channel 32-bit floating-point image mean and standard deviation, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer</param>
				/// <param name="pMean">Contains computed mean.</param>
				/// <param name="pStdDev">Contains computed standard deviation.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMean_StdDev_32f_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pDeviceBuffer, CUdeviceptr pMean, CUdeviceptr pStdDev);


 
			}

			/// <summary>
			/// Infinite Norm
			/// </summary>
			public static class NormInf
			{		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_32s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_32s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormInf_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormInfGetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit unsigned short image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit signed short image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit signed int image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_32s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit floating-point image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 8-bit unsigned char image infinity norm, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_8u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 8-bit signed char image infinity norm, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_8s_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit unsigned short image infinity norm, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit floating-point image infinity norm, \ref masked_operation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_32f_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit unsigned char image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit unsigned short image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit signed short image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 32-bit floating-point image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 8-bit unsigned char image infinity norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit unsigned short image infinity norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit signed short image infinity norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 32-bit floating-point image infinity norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 8-bit unsigned char image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit unsigned short image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit signed short image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 32-bit floating-point image infinity norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit unsigned char image infinity norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_8u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit signed char image infinity norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_8s_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit unsigned short image infinity norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_16u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 32-bit floating-point image infinity norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an invalid 
				/// floating-point image is specified, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_Inf_32f_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


 
			}

			/// <summary>
			/// L1 Norm
			/// </summary>
			public static class NormL1
			{ 		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL1_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL1GetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit unsigned short image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit signed short image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit floating-point image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 8-bit unsigned char image L1 norm, \ref masked_operation
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_8u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 8-bit signed char image L1 norm, \ref masked_operation
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_8s_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit unsigned short image L1 norm, \ref masked_operation
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit floating-point image L1 norm, \ref masked_operation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_32f_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit unsigned char image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit unsigned short image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit signed short image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 32-bit floating-point image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 8-bit unsigned char image L1 norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit unsigned short image L1 norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit signed short image L1 norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 32-bit floating-point image L1 norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 8-bit unsigned char image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit unsigned short image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit signed short image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 32-bit floating-point image L1 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit unsigned char image L1 norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_8u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit signed char image L1 norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_8s_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit unsigned short image L1 norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_16u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 32-bit floating-point image L1 norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if the step of 
				/// the source image cannot be divided by 4, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L1_32f_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);



			}

			/// <summary>
			/// L2 Norm
			/// </summary>
			public static class NormL2
			{ 		/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiNormL2_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormL2GetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// 1-channel 8-bit unsigned char image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit unsigned short image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 16-bit signed short image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 1-channel 32-bit floating-point image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// \ref masked_operation 1-channel 8-bit unsigned char image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_8u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// \ref masked_operation 1-channel 8-bit signed char image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_8s_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// \ref masked_operation 1-channel 16-bit unsigned short image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16u_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// \ref masked_operation 1-channel 32-bit floating-point image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if the step 
				/// of the source image cannot be divided by 4. 
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_32f_C1MR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit unsigned char image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit unsigned short image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit signed short image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 32-bit floating-point image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 8-bit unsigned char image L2 norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit unsigned short image L2 norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit signed short image L2 norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 32-bit floating-point image L2 norm (alpha channel is not computed).
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of three channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 8-bit unsigned char image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit unsigned short image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 16-bit signed short image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 4-channel 32-bit floating-point image L2 norm.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNorm">Array that contains the norm values of four channels.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit unsigned char image L2 norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_8u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 8-bit signed char image L2 norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_8s_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 16-bit unsigned short image L2 norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_16u_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 3-channel 32-bit floating-point image L2 norm, \ref masked_operation, \ref channel_of_interest.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNorm">Pointer to the norm value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if the step 
				/// of the source image cannot be divided by 4, or NPP_COI_ERROR if an invalid channel of interest is specified.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNorm_L2_32f_C3CMR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);



			}


			//new in Cuda 5.5
			/// <summary>
			/// RGB to CCIR601 Gray conversion.<para/>
			/// Here is how NPP converts gamma corrected RGB to CCIR601 Gray.<para/>
			/// <code>nGray =  0.299F * R + 0.587F * G + 0.114F * B;</code>
			/// </summary>
			public static class RGBToGray
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 1 channel 8-bit unsigned packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToGray_8u_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 1 channel 8-bit unsigned packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToGray_8u_AC4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 16-bit unsigned packed RGB to 1 channel 16-bit unsigned packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToGray_16u_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit unsigned packed RGB with alpha to 1 channel 16-bit unsigned packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToGray_16u_AC4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 16-bit signed packed RGB to 1 channel 16-bit signed packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToGray_16s_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 16-bit signed packed RGB with alpha to 1 channel 16-bit signed packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToGray_16s_AC4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 3 channel 32-bit floating point packed RGB to 1 channel 32-bit floating point packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToGray_32f_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// 4 channel 32-bit floating point packed RGB with alpha to 1 channel 32-bit floating point packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRGBToGray_32f_AC4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);
			}

			/// <summary>
			/// RGB Color to Gray conversion using user supplied conversion coefficients.<para/>
			/// Here is how NPP converts gamma corrected RGB Color to Gray using user supplied conversion coefficients.<para/>
			/// <code>nGray =  aCoeffs[0] * R + aCoeffs[1] * G + aCoeffs[2] * B; </code>
			/// </summary>
			public static class ColorToGray
			{
				/// <summary>
				/// 3 channel 8-bit unsigned packed RGB to 1 channel 8-bit unsigned packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCoeffs">fixed size array of constant floating point conversion coefficient values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorToGray_8u_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aCoeffs);


				/// <summary>
				/// 4 channel 8-bit unsigned packed RGB with alpha to 1 channel 8-bit unsigned packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCoeffs">fixed size array of constant floating point conversion coefficient values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorToGray_8u_AC4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aCoeffs);


				/// <summary>
				/// 3 channel 16-bit unsigned packed RGB to 1 channel 16-bit unsigned packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCoeffs">fixed size array of constant floating point conversion coefficient values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorToGray_16u_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aCoeffs);


				/// <summary>
				/// 4 channel 16-bit unsigned packed RGB with alpha to 1 channel 16-bit unsigned packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCoeffs">fixed size array of constant floating point conversion coefficient values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorToGray_16u_AC4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aCoeffs);


				/// <summary>
				/// 3 channel 16-bit signed packed RGB to 1 channel 16-bit signed packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCoeffs">fixed size array of constant floating point conversion coefficient values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorToGray_16s_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aCoeffs);


				/// <summary>
				/// 4 channel 16-bit signed packed RGB with alpha to 1 channel 16-bit signed packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCoeffs">fixed size array of constant floating point conversion coefficient values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorToGray_16s_AC4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aCoeffs);


				/// <summary>
				/// 3 channel 32-bit floating point packed RGB to 1 channel 32-bit floating point packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCoeffs">fixed size array of constant floating point conversion coefficient values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorToGray_32f_C3C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aCoeffs);


				/// <summary>
				/// 4 channel 32-bit floating point packed RGB with alpha to 1 channel 32-bit floating point packed Gray conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCoeffs">fixed size array of constant floating point conversion coefficient values, one per color channel.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorToGray_32f_AC4C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aCoeffs);



			}

			/// <summary>
			/// Perform color twist pixel processing.  Color twist consists of applying the following formula to each
			/// image pixel using coefficients from the user supplied color twist host matrix array as follows where 
			/// dst[x] and src[x] represent destination pixel and source pixel channel or plane x.<para/>
			/// dst[0] = aTwist[0][0] * src[0] + aTwist[0][1] * src[1] + aTwist[0][2] * src[2] + aTwist[0][3]<para/>
			/// dst[1] = aTwist[1][0] * src[0] + aTwist[1][1] * src[1] + aTwist[1][2] * src[2] + aTwist[1][3]<para/>
			/// dst[2] = aTwist[2][0] * src[0] + aTwist[2][1] * src[1] + aTwist[2][2] * src[2] + aTwist[2][3]<para/>
			/// </summary>
			public static class ColorTwist
			{
				/// <summary>
				/// 3 channel 8-bit unsigned in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 4 channel 8-bit unsigned in place color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point coefficient values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 8-bit unsigned planar in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place planar pixel format image pointer array, one pointer per plane.</param>
				/// <param name="nSrcDstStep">in place planar pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8u_IP3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 8-bit signed color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 8-bit signed in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 4 channel 8-bit signed color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point coefficient values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 4 channel 8-bit signed in place color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point coefficient values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 8-bit signed planar color twist.
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 8-bit signed planar in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place planar pixel format image pointer array, one pointer per plane.</param>
				/// <param name="nSrcDstStep">in place planar pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_8s_IP3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 16-bit unsigned color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 16-bit unsigned in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 4 channel 16-bit unsigned color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point coefficient values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 4 channel 16-bit unsigned in place color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point coefficient values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 16-bit unsigned planar color twist.
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 16-bit unsigned planar in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place planar pixel format image pointer array, one pointer per plane.</param>
				/// <param name="nSrcDstStep">in place planar pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16u_IP3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 16-bit signed color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 16-bit signed in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 4 channel 16-bit signed color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point coefficient values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 4 channel 16-bit signed in place color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point coefficient values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 16-bit signed planar color twist.
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 16-bit signed planar in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place planar pixel format image pointer array, one pointer per plane.</param>
				/// <param name="nSrcDstStep">in place planar pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist32f_16s_IP3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 32-bit floating point color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 32-bit floating point in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 4 channel 32-bit floating point color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point coefficient values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 4 channel 32-bit floating point in place color twist, not affecting Alpha.
				/// An input color twist matrix with floating-point coefficient values is applied with
				/// in ROI.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">in place packed pixel format image pointer.</param>
				/// <param name="nSrcDstStep">in place packed pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 32-bit floating point planar color twist.
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, int nSrcStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);


				/// <summary>
				/// 3 channel 32-bit floating point planar in place color twist.
				/// 
				/// An input color twist matrix with floating-point coefficient values is applied
				/// within ROI.
				/// </summary>
				/// <param name="pSrcDst">in place planar pixel format image pointer array, one pointer per plane.</param>
				/// <param name="nSrcDstStep">in place planar pixel format image line step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aTwist">The color twist matrix with floating-point coefficient values.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiColorTwist_32f_IP3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 12)] float[,] aTwist);



			}

			/// <summary>
			/// Perform image color processing using members of various types of color look up tables.
			/// </summary>
			public static class ColorLUT
			{
				/// <summary>
				/// 8-bit unsigned look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 8-bit unsigned look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 8-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 8-bit unsigned look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_8u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 16-bit unsigned look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 16-bit unsigned look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 16-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 16-bit unsigned look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 16-bit signed look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 16-bit signed look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 16-bit signed look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 16-bit signed look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16s_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 32-bit floating point look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 32-bit floating point look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 32-bit floating point look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 32-bit floating point look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points with no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_32f_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);



			}

			/// <summary>
			/// Perform image color processing using linear interpolation between members of various types of color look up tables.
			/// </summary>
			public static class ColorLUTLinear
			{
				/// <summary>
				/// 8-bit unsigned linear interpolated look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points through linear interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 8-bit unsigned linear interpolated look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through linear interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned linear interpolated look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through linear interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_8u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned linear interpolated look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points through linear interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 16-bit unsigned look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 16-bit unsigned look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 16-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 16-bit unsigned look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 16-bit signed look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 16-bit signed look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 16-bit signed look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 16-bit signed look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16s_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 32-bit floating point look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 32-bit floating point look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 32-bit floating point look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 32-bit floating point look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using linear interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_32f_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Linear_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);



			}

			/// <summary>
			/// Perform image color processing using linear interpolation between members of various types of color look up tables.
			/// </summary>
			public static class ColorLUTCubic
			{
				/// <summary>
				/// 8-bit unsigned cubic interpolated look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 8-bit unsigned cubic interpolated look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_8u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 8-bit unsigned cubic interpolated look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 8-bit unsigned cubic interpolated look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_8u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned cubic interpolated look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned cubic interpolated look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_8u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned cubic interpolated look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 8-bit unsigned cubic interpolated look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 256.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_8u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 16-bit unsigned look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 16-bit unsigned look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16u_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 16-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 16-bit unsigned look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16u_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16u_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit unsigned look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16u_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 16-bit signed look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 16-bit signed look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16s_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 16-bit signed look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 16-bit signed look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16s_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16s_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 16-bit signed look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_16s_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 32-bit floating point look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 32-bit floating point look-up-table in place color conversion.
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Pointer to an array of user defined OUTPUT values (this is a device memory pointer)</param>
				/// <param name="pLevels">Pointer to an array of user defined INPUT values  (this is a device memory pointer)</param>
				/// <param name="nLevels">Number of user defined number of input/output mapping points (levels)</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_32f_C1IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, CUdeviceptr pValues, CUdeviceptr pLevels, int nLevels);


				/// <summary>
				/// 3 channel 32-bit floating point look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 3 channel 32-bit floating point look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_32f_C3IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points through cubic interpolation. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table in place color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 4 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_32f_C4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);


				/// <summary>
				/// 4 channel 32-bit floating point look-up-table in place color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points using no interpolation. 
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">In-Place Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pValues">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT values.</param>
				/// <param name="pLevels">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined INPUT values.</param>
				/// <param name="nLevels">Host pointer to an array of 3 user defined number of input/output mapping points, one per color CHANNEL.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_NUMBER_OF_LEVELS_ERROR if the number of levels is less than 2 or greater than 1024 (the current size limit).
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUT_Cubic_32f_AC4IR(CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pValues, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pLevels, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] nLevels);



			}

			/// <summary>
			/// Perform image color processing using various types of bit range restricted palette color look up tables.
			/// </summary>
			public static class ColorLUTPalette
			{
				/// <summary>
				/// One channel 8-bit unsigned bit range restricted palette look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTable">Pointer to an array of user defined OUTPUT palette values (this is a device memory pointer)</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 8) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 8.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pTable, int nBitSize);


				/// <summary>
				/// One channel 8-bit unsigned bit range restricted 24-bit palette look-up-table color conversion with 24-bit destination output per pixel.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step (3 bytes per pixel).</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTable">Pointer to an array of user defined OUTPUT palette values (this is a device memory pointer)</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 8) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 8.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_8u24u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pTable, int nBitSize);


				/// <summary>
				/// One channel 8-bit unsigned bit range restricted 32-bit palette look-up-table color conversion with 32-bit destination output per pixel.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step (4 bytes per pixel).</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTable">Pointer to an array of user defined OUTPUT palette values (this is a device memory pointer)</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 8) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 8.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_8u32u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pTable, int nBitSize);


				/// <summary>
				/// Three channel 8-bit unsigned bit range restricted palette look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTables">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT palette values.</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 8) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 8.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pTables, int nBitSize);


				/// <summary>
				/// Four channel 8-bit unsigned bit range restricted palette look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTables">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT palette values.</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 8) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 8.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pTables, int nBitSize);


				/// <summary>
				/// Four channel 8-bit unsigned bit range restricted palette look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTables">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT palette values.</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 8) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 8.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pTables, int nBitSize);


				/// <summary>
				/// One channel 16-bit unsigned bit range restricted palette look-up-table color conversion.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTable">Pointer to an array of user defined OUTPUT palette values (this is a device memory pointer)</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 16) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 16.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pTable, int nBitSize);


				/// <summary>
				/// One channel 16-bit unsigned bit range restricted 8-bit unsigned palette look-up-table color conversion with 8-bit unsigned destination output per pixel.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step (1 unsigned byte per pixel).</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTable">Pointer to an array of user defined OUTPUT palette values (this is a device memory pointer)</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 16) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 16.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_16u8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pTable, int nBitSize);


				/// <summary>
				/// One channel 16-bit unsigned bit range restricted 24-bit unsigned palette look-up-table color conversion with 24-bit unsigned destination output per pixel.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step (3 unsigned bytes per pixel).</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTable">Pointer to an array of user defined OUTPUT palette values (this is a device memory pointer)</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 16) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 16.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_16u24u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pTable, int nBitSize);


				/// <summary>
				/// One channel 16-bit unsigned bit range restricted 32-bit palette look-up-table color conversion with 32-bit unsigned destination output per pixel.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step (4 bytes per pixel).</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTable">Pointer to an array of user defined OUTPUT palette values (this is a device memory pointer)</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 16) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 16.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_16u32u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, CUdeviceptr pTable, int nBitSize);


				/// <summary>
				/// Three channel 16-bit unsigned bit range restricted palette look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTables">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT palette values.</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 16) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 16.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pTables, int nBitSize);


				/// <summary>
				/// Four channel 16-bit unsigned bit range restricted palette look-up-table color conversion.
				/// 
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTables">Host pointer to an array of 4 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT palette values.</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 16) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 16.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pTables, int nBitSize);


				/// <summary>
				/// Four channel 16-bit unsigned bit range restricted palette look-up-table color conversion, not affecting Alpha.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// Alpha channel is the last channel and is not processed.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTables">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT palette values.</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 16) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 16.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPalette_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pTables, int nBitSize);


				/// <summary>
				/// Three channel 8-bit unsigned source bit range restricted palette look-up-table color conversion to four channel 8-bit unsigned destination output with alpha.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// This function also reverses the source pixel channel order in the destination so the Alpha channel is the first channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step (3 bytes per pixel).</param>
				/// <param name="nAlphaValue">Signed alpha value that will be used to initialize the pixel alpha channel position in all modified destination pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step (4 bytes per pixel with alpha).</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTables">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT palette values.</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 8) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 8.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPaletteSwap_8u_C3A0C4R(CUdeviceptr pSrc, int nSrcStep, int nAlphaValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pTables, int nBitSize);


				/// <summary>
				/// Three channel 16-bit unsigned source bit range restricted palette look-up-table color conversion to four channel 16-bit unsigned destination output with alpha.
				/// The LUT is derived from a set of user defined mapping points in a palette and 
				/// source pixels are then processed using a restricted bit range when looking up palette values.
				/// This function also reverses the source pixel channel order in the destination so the Alpha channel is the first channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step (3 unsigned short integers per pixel).</param>
				/// <param name="nAlphaValue">Signed alpha value that will be used to initialize the pixel alpha channel position in all modified destination pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step (4 unsigned short integers per pixel with alpha).</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTables">Host pointer to an array of 3 device memory pointers, one per color CHANNEL, pointing to user defined OUTPUT palette values.</param>
				/// <param name="nBitSize">Number of least significant bits (must be &gt; 0 and &lt;= 16) of each source pixel value to use as index into palette table during conversion.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				///        - ::NPP_LUT_PALETTE_BITSIZE_ERROR if nBitSize is &lt; 1 or &gt; 16.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiLUTPaletteSwap_16u_C3A0C4R(CUdeviceptr pSrc, int nSrcStep, int nAlphaValue, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pTables, int nBitSize);



			}

			/// <summary>
			/// Image compression primitives.
			/// </summary>
			public static class CompressionDCT
			{
				/// <summary>
				/// Initializes DCT state structure and allocates additional resources.
				/// \see nppiDCTQuantFwd8x8LS_JPEG_8u16s_C1R_NEW(), nppiDCTQuantInv8x8LS_JPEG_16s8u_C1R_NEW.
				/// </summary>
				/// <param name="ppState">Pointer to pointer to DCT state structure. </param>
				/// <returns>
				/// NPP_SUCCESS Indicates no error. Any other value indicates an error
				///         or a warning
				/// \return NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///         has zero or negative value
				/// \return NPP_NULL_POINTER_ERROR Indicates an error condition if pBufSize
				///         pointer is NULL
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDCTInitAlloc(ref NppiDCTState ppState);


				/// <summary>
				/// Frees the additional resources of the DCT state structure.
				/// \see nppiDCTInitAlloc
				/// </summary>
				/// <param name="pState">Pointer to DCT state structure. </param>
				/// <returns>
				/// NPP_SUCCESS Indicates no error. Any other value indicates an error
				///         or a warning
				/// \return NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///         has zero or negative value
				/// \return NPP_NULL_POINTER_ERROR Indicates an error condition if pState
				///         pointer is NULL
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDCTFree(NppiDCTState pState);


				/// <summary>
				/// Forward DCT, quantization and level shift part of the JPEG encoding.
				/// Input is expected in 8x8 macro blocks and output is expected to be in 64x1
				/// macro blocks. The new version of the primitive takes the ROI in image pixel size and
				/// works with DCT coefficients that are in zig-zag order.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Image width in pixels x 8 x sizeof(Npp16s).</param>
				/// <param name="pQuantizationTable">Quantization Table in zig-zag order.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pState">Pointer to DCT state structure. This structure must be</param>
				/// <returns>
				/// Error codes:
				///         - ::NPP_SIZE_ERROR For negative input height/width or not a multiple of
				///           8 width/height.
				///         - ::NPP_STEP_ERROR If input image width is not multiple of 8 or does not
				///           match ROI.
				///         - ::NPP_NULL_POINTER_ERROR If the destination pointer is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDCTQuantFwd8x8LS_JPEG_8u16s_C1R_NEW(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, CUdeviceptr pQuantizationTable, NppiSize oSizeROI, NppiDCTState pState);


				/// <summary>
				/// Inverse DCT, de-quantization and level shift part of the JPEG decoding.
				/// Input is expected in 64x1 macro blocks and output is expected to be in 8x8
				/// macro blocks. The new version of the primitive takes the ROI in image pixel size and
				/// works with DCT coefficients that are in zig-zag order.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Image width in pixels x 8 x sizeof(Npp16s).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pQuantizationTable">Quantization Table in zig-zag order.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pState">Pointer to DCT state structure. This structure must be</param>
				/// <returns>
				/// Error codes:
				///         - ::NPP_SIZE_ERROR For negative input height/width or not a multiple of
				///           8 width/height.
				///         - ::NPP_STEP_ERROR If input image width is not multiple of 8 or does not
				///           match ROI.
				///         - ::NPP_NULL_POINTER_ERROR If the destination pointer is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDCTQuantInv8x8LS_JPEG_16s8u_C1R_NEW(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, CUdeviceptr pQuantizationTable, NppiSize oSizeROI, NppiDCTState pState);

				/// <summary>
				/// Returns the length of the NppiDecodeHuffmanSpec structure.
				/// </summary>
				/// <param name="pSize">Pointer to a variable that will receive the length of the NppiDecodeHuffmanSpec structure.</param>
				/// <returns></returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDecodeHuffmanSpecGetBufSize_JPEG(ref int pSize);

				
				/// <summary>
				/// Creates a Huffman table in a format that is suitable for the decoder on the host.
				/// </summary>
				/// <param name="pRawHuffmanTable">Huffman table formated as specified in the JPEG standard.</param>
				/// <param name="eTableType">Enum specifying type of table (nppiDCTable or nppiACTable)</param>
				/// <param name="pHuffmanSpec">Pointer to the Huffman table for the decoder</param>
				/// <returns>NPP_NULL_POINTER_ERROR If one of the pointers is 0.</returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDecodeHuffmanSpecInitHost_JPEG(byte[] pRawHuffmanTable, NppiHuffmanTableType eTableType, NppiDecodeHuffmanSpec  pHuffmanSpec);

				/// <summary>
				/// Allocates memory and creates a Huffman table in a format that is suitable for the decoder on the host.
				/// </summary>
				/// <param name="pRawHuffmanTable">Huffman table formated as specified in the JPEG standard.</param>
				/// <param name="eTableType">Enum specifying type of table (nppiDCTable or nppiACTable).</param>
				/// <param name="ppHuffmanSpec">Pointer to returned pointer to the Huffman table for the decoder</param>
				/// <returns>NPP_NULL_POINTER_ERROR If one of the pointers is 0.</returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDecodeHuffmanSpecInitAllocHost_JPEG(byte[] pRawHuffmanTable, NppiHuffmanTableType eTableType, ref NppiDecodeHuffmanSpec ppHuffmanSpec);

				/// <summary>
				/// Frees the host memory allocated by nppiDecodeHuffmanSpecInitAllocHost_JPEG.
				/// </summary>
				/// <param name="pHuffmanSpec">Pointer to the Huffman table for the decoder</param>
				/// <returns></returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDecodeHuffmanSpecFreeHost_JPEG(NppiDecodeHuffmanSpec pHuffmanSpec);


				/// <summary>
				/// Huffman Decoding of the JPEG decoding on the host.
				/// Input is expected in byte stuffed huffman encoded JPEG scan and output is expected to be 64x1 macro blocks.
				/// </summary>
				/// <param name="pSrc">Byte-stuffed huffman encoded JPEG scan.</param>
				/// <param name="nLength">Byte length of the input.</param>
				/// <param name="restartInterval">Restart Interval, see JPEG standard.</param>
				/// <param name="Ss">Start Coefficient, see JPEG standard.</param>
				/// <param name="Se">End Coefficient, see JPEG standard.</param>
				/// <param name="Ah">Bit Approximation High, see JPEG standard.</param>
				/// <param name="Al">Bit Approximation Low, see JPEG standard.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pHuffmanTableDC">DC Huffman table.</param>
				/// <param name="pHuffmanTableAC">AC Huffman table.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// Error codes:
				///         - ::NPP_SIZE_ERROR For negative input height/width or not a multiple of
				///           8 width/height.
				///         - ::NPP_STEP_ERROR If input image width is not multiple of 8 or does not
				///           match ROI.
				///         - ::NPP_NULL_POINTER_ERROR If the destination pointer is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDecodeHuffmanScanHost_JPEG_8u16s_P1R(byte[] pSrc, int nLength, int restartInterval, int Ss, int Se, int Ah, int Al,
					short[] pDst, int nDstStep, NppiDecodeHuffmanSpec pHuffmanTableDC, NppiDecodeHuffmanSpec pHuffmanTableAC, NppiSize oSizeROI);


				/// <summary>
				/// Huffman Decoding of the JPEG decoding on the host.
				/// Input is expected in byte stuffed huffman encoded JPEG scan and output is expected to be 64x1 macro blocks.
				/// </summary>
				/// <param name="pSrc">Byte-stuffed huffman encoded JPEG scan.</param>
				/// <param name="nLength">Byte length of the input.</param>
				/// <param name="nRestartInterval">Restart Interval, see JPEG standard. </param>
				/// <param name="nSs">Start Coefficient, see JPEG standard.</param>
				/// <param name="nSe">End Coefficient, see JPEG standard.</param>
				/// <param name="nAh">Bit Approximation High, see JPEG standard.</param>
				/// <param name="nAl">Bit Approximation Low, see JPEG standard.</param>
				/// <param name="apDst">Destination-Image Pointer.</param>
				/// <param name="aDstStep">Destination-Image Line Step.</param>
				/// <param name="apHuffmanDCTable">DC Huffman tables.</param>
				/// <param name="apHuffmanACTable">AC Huffman tables.</param>
				/// <param name="aSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// Error codes:
				///         - ::NPP_SIZE_ERROR For negative input height/width or not a multiple of
				///           8 width/height.
				///         - ::NPP_STEP_ERROR If input image width is not multiple of 8 or does not
				///           match ROI.
				///         - ::NPP_NULL_POINTER_ERROR If the destination pointer is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDecodeHuffmanScanHost_JPEG_8u16s_P3R(byte[] pSrc, int nLength, int nRestartInterval, int nSs, int nSe, int nAh, int nAl, 
					[In, Out, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] IntPtr[] apDst, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aDstStep,
					[In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] NppiDecodeHuffmanSpec[] apHuffmanDCTable, 
					[In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] NppiDecodeHuffmanSpec[] apHuffmanACTable, 
					[In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] NppiSize[] aSizeROI);


				
				/// <summary>
				/// Returns the length of the NppiEncodeHuffmanSpec structure.
				/// </summary>
				/// <param name="pSize">Pointer to a variable that will receive the length of the NppiEncodeHuffmanSpec structure.</param>
				/// <returns>NPP_NULL_POINTER_ERROR If one of the pointers is 0.</returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiEncodeHuffmanSpecGetBufSize_JPEG(ref int pSize);

				/// <summary>
				/// Creates a Huffman table in a format that is suitable for the encoder.
				/// </summary>
				/// <param name="pRawHuffmanTable">Huffman table formated as specified in the JPEG standard.</param>
				/// <param name="eTableType">Enum specifying type of table (nppiDCTable or nppiACTable).</param>
				/// <param name="pHuffmanSpec">Pointer to the Huffman table for the decoder</param>
				/// <returns>NPP_NULL_POINTER_ERROR If one of the pointers is 0.</returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiEncodeHuffmanSpecInit_JPEG(byte[] pRawHuffmanTable, NppiHuffmanTableType eTableType, NppiEncodeHuffmanSpec pHuffmanSpec);

				/// <summary>
				/// Allocates memory and creates a Huffman table in a format that is suitable for the encoder.
				/// </summary>
				/// <param name="pRawHuffmanTable">Huffman table formated as specified in the JPEG standard.</param>
				/// <param name="eTableType">Enum specifying type of table (nppiDCTable or nppiACTable).</param>
				/// <param name="ppHuffmanSpec">Pointer to returned pointer to the Huffman table for the encoder</param>
				/// <returns>NPP_NULL_POINTER_ERROR If one of the pointers is 0.</returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiEncodeHuffmanSpecInitAlloc_JPEG(byte[] pRawHuffmanTable, NppiHuffmanTableType eTableType, ref NppiEncodeHuffmanSpec ppHuffmanSpec);

				/// <summary>
				/// Frees the memory allocated by nppiEncodeHuffmanSpecInitAlloc_JPEG.
				/// </summary>
				/// <param name="pHuffmanSpec">Pointer to the Huffman table for the encoder</param>
				/// <returns></returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiEncodeHuffmanSpecFree_JPEG(NppiEncodeHuffmanSpec  pHuffmanSpec);



				/// <summary>
				/// Huffman Encoding of the JPEG Encoding.
				/// Input is expected to be 64x1 macro blocks and output is expected as byte stuffed huffman encoded JPEG scan.
				/// </summary>
				/// <param name="pSrc">Destination-Image Pointer.</param>
				/// <param name="nSrcStep">Destination-Image Line Step.</param>
				/// <param name="nRestartInterval">Restart Interval, see JPEG standard. Currently only values &lt;=0 are supported.</param>
				/// <param name="nSs">Start Coefficient, see JPEG standard.</param>
				/// <param name="nSe">End Coefficient, see JPEG standard.</param>
				/// <param name="nAh">Bit Approximation High, see JPEG standard.</param>
				/// <param name="nAl">Bit Approximation Low, see JPEG standard.</param>
				/// <param name="pDst">Byte-stuffed huffman encoded JPEG scan.</param>
				/// <param name="nLength">Byte length of the huffman encoded JPEG scan.</param>
				/// <param name="pHuffmanTableDC">DC Huffman table.</param>
				/// <param name="pHuffmanTableAC">AC Huffman table.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTempStorage"></param>
				/// <returns>
				/// Error codes:
				///         - ::NPP_SIZE_ERROR For negative input height/width or not a multiple of
				///           8 width/height.
				///         - ::NPP_STEP_ERROR If input image width is not multiple of 8 or does not
				///           match ROI.
				///         - ::NPP_NULL_POINTER_ERROR If the destination pointer is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus
				nppiEncodeHuffmanScan_JPEG_8u16s_P1R(CUdeviceptr pSrc, int nSrcStep, int nRestartInterval, int nSs, int nSe, int nAh, int nAl, CUdeviceptr pDst,
					ref int nLength, NppiEncodeHuffmanSpec pHuffmanTableDC, NppiEncodeHuffmanSpec pHuffmanTableAC, NppiSize oSizeROI, CUdeviceptr pTempStorage);


				/// <summary>
				/// Huffman Encoding of the JPEG Encoding.
				/// Input is expected to be 64x1 macro blocks and output is expected as byte stuffed huffman encoded JPEG scan.
				/// </summary>
				/// <param name="apSrc">Destination-Image Pointer.</param>
				/// <param name="aSrcStep">Destination-Image Line Step.</param>
				/// <param name="nRestartInterval">Restart Interval, see JPEG standard. Currently only values &lt;=0 are supported.</param>
				/// <param name="nSs">Start Coefficient, see JPEG standard.</param>
				/// <param name="nSe">End Coefficient, see JPEG standard.</param>
				/// <param name="nAh">Bit Approximation High, see JPEG standard.</param>
				/// <param name="nAl">Bit Approximation Low, see JPEG standard.</param>
				/// <param name="pDst">Byte-stuffed huffman encoded JPEG scan.</param>
				/// <param name="nLength">Byte length of the huffman encoded JPEG scan.</param>
				/// <param name="apHuffmanDCTable">DC Huffman tables.</param>
				/// <param name="apHuffmanACTable">AC Huffman tables.</param>
				/// <param name="aSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pTempStorage"></param>
				/// <returns>
				/// Error codes:
				///         - ::NPP_SIZE_ERROR For negative input height/width or not a multiple of
				///           8 width/height.
				///         - ::NPP_STEP_ERROR If input image width is not multiple of 8 or does not
				///           match ROI.
				///         - ::NPP_NULL_POINTER_ERROR If the destination pointer is 0.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiEncodeHuffmanScan_JPEG_8u16s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] apSrc, 
					[In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] int[] aSrcStep, int nRestartInterval, int nSs, int nSe, int nAh, int nAl, CUdeviceptr pDst, ref int nLength,
					[In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] NppiEncodeHuffmanSpec[] apHuffmanDCTable,
					[In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] NppiEncodeHuffmanSpec[] apHuffmanACTable, 
					[In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] NppiSize[] aSizeROI, CUdeviceptr pTempStorage);


				/// <summary>
				/// Calculates the size of the temporary buffer for huffman encoding.
				/// \see nppiEncodeHuffmanScan_JPEG_8u16s_P1R(), nppiEncodeHuffmanScan_JPEG_8u16s_P3R().
				/// </summary>
				/// <param name="oSize">Image Dimension.</param>
				/// <param name="nChannels"></param>
				/// <param name="pBufSize">Pointer to variable that returns the size of the</param>
				/// <returns>
				/// NPP_SUCCESS Indicates no error. Any other value indicates an error
				///         or a warning
				/// \return NPP_SIZE_ERROR Indicates an error condition if any image dimension
				///         has zero or negative value
				/// \return NPP_NULL_POINTER_ERROR Indicates an error condition if pBufSize
				///         pointer is NULL
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiEncodeHuffmanGetSize(NppiSize oSize, int nChannels, ref int pBufSize);



			}

			/// <summary>
			/// Scale bit-depth up and down.<para/>
			/// To map source pixel srcPixelValue to destination pixel dstPixelValue the following equation is used:<para/>
			/// dstPixelValue = dstMinRangeValue + scaleFactor * (srcPixelValue - srcMinRangeValue)<para/>
			/// where scaleFactor = (dstMaxRangeValue - dstMinRangeValue) / (srcMaxRangeValue - srcMinRangeValue).<para/>
			/// For conversions between integer data types, the entire integer numeric range of the input data type is mapped onto 
			/// the entire integer numeric range of the output data type.<para/>
			/// For conversions to floating point data types the floating point data range is defined by the user supplied floating point values
			/// of nMax and nMin which are used as the dstMaxRangeValue and dstMinRangeValue respectively in the scaleFactor and dstPixelValue 
			/// calculations and also as the saturation values to which output data is clamped.<para/>
			/// When converting from floating-point values to integer values, nMax and nMin are used as the srcMaxRangeValue and srcMinRangeValue
			/// respectively in the scaleFactor and dstPixelValue calculations. Output values are saturated and clamped to the full output integer
			/// pixel value range.
			/// </summary>
			public static class Scale
			{
				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned to 16-bit unsigned  conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 16-bit unsigned  conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 16-bit unsigned conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned to 16-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 16-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 16-bit signed conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 32-bit signed conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned to 32-bit signed conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nMin">specifies the minimum saturation value to which every output value will be clamped.</param>
				/// <param name="nMax">specifies the maximum saturation value to which every output value will be clamped.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, ::NPP_SCALE_RANGE_ERROR indicates an error condition if nMax &lt;= nMin.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nMin, float nMax);


				/// <summary>
				/// Three channel 8-bit unsigned to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nMin">specifies the minimum saturation value to which every output value will be clamped.</param>
				/// <param name="nMax">specifies the maximum saturation value to which every output value will be clamped.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, ::NPP_SCALE_RANGE_ERROR indicates an error condition if nMax &lt;= nMin.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nMin, float nMax);


				/// <summary>
				/// Four channel 8-bit unsigned to 32-bit floating-point conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nMin">specifies the minimum saturation value to which every output value will be clamped.</param>
				/// <param name="nMax">specifies the maximum saturation value to which every output value will be clamped.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, ::NPP_SCALE_RANGE_ERROR indicates an error condition if nMax &lt;= nMin.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nMin, float nMax);


				/// <summary>
				/// Four channel 8-bit unsigned to 32-bit floating-point conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nMin">specifies the minimum saturation value to which every output value will be clamped.</param>
				/// <param name="nMax">specifies the maximum saturation value to which every output value will be clamped.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, ::NPP_SCALE_RANGE_ERROR indicates an error condition if nMax &lt;= nMin.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nMin, float nMax);


				/// <summary>
				/// Single channel 16-bit unsigned to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_16u8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Three channel 16-bit unsigned to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_16u8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Four channel 16-bit unsigned to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_16u8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Four channel 16-bit unsigned to 8-bit unsigned conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_16u8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Single channel 16-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_16s8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Three channel 16-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_16s8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Four channel 16-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_16s8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Four channel 16-bit signed to 8-bit unsigned conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_16s8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Single channel 32-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_32s8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Three channel 32-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_32s8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Four channel 32-bit signed to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_32s8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Four channel 32-bit signed to 8-bit unsigned conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hint">algorithm performance or accuracy selector, currently ignored</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_32s8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppHintAlgorithm hint);


				/// <summary>
				/// Single channel 32-bit floating point to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nMin">specifies the minimum saturation value to which every output value will be clamped.</param>
				/// <param name="nMax">specifies the maximum saturation value to which every output value will be clamped.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, ::NPP_SCALE_RANGE_ERROR indicates an error condition if nMax &lt;= nMin.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_32f8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nMin, float nMax);


				/// <summary>
				/// Three channel 32-bit floating point to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nMin">specifies the minimum saturation value to which every output value will be clamped.</param>
				/// <param name="nMax">specifies the maximum saturation value to which every output value will be clamped.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, ::NPP_SCALE_RANGE_ERROR indicates an error condition if nMax &lt;= nMin.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_32f8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nMin, float nMax);


				/// <summary>
				/// Four channel 32-bit floating point to 8-bit unsigned conversion.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nMin">specifies the minimum saturation value to which every output value will be clamped.</param>
				/// <param name="nMax">specifies the maximum saturation value to which every output value will be clamped.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, ::NPP_SCALE_RANGE_ERROR indicates an error condition if nMax &lt;= nMin.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_32f8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nMin, float nMax);


				/// <summary>
				/// Four channel 32-bit floating point to 8-bit unsigned conversion, not affecting Alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nMin">specifies the minimum saturation value to which every output value will be clamped.</param>
				/// <param name="nMax">specifies the maximum saturation value to which every output value will be clamped.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, ::NPP_SCALE_RANGE_ERROR indicates an error condition if nMax &lt;= nMin.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiScale_32f8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, float nMin, float nMax);



			}

			/// <summary>
			/// Methods for copying images and padding borders with a replicates of the nearest source image pixel color.
			/// </summary>
			public static class CopyReplicateBorder
			{
				/// <summary>
				/// 1 channel 8-bit unsigned integer image copy with nearest source image pixel color.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 8-bit unsigned integer image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 8-bit unsigned integer image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 8-bit unsigned integer image copy with nearest source image pixel color with alpha channel unaffected.
				/// See nppiCopyReplicateBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 1 channel 16-bit unsigned integer image copy with nearest source image pixel color.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 16-bit unsigned integer image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 16-bit unsigned integer image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 16-bit unsigned image copy with nearest source image pixel color with alpha channel unaffected.
				/// See nppiCopyReplicateBorder_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 1 channel 16-bit signed integer image copy with nearest source image pixel color.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 16-bit signed integer image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 16-bit signed integer image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 16-bit signed integer image copy with nearest source image pixel color with alpha channel unaffected.
				/// See nppiCopyReplicateBorder_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 1 channel 32-bit signed integer image copy with nearest source image pixel color.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_32s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 32-bit signed image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_32s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 32-bit signed integer image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_32s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 32-bit signed integer image copy with nearest source image pixel color with alpha channel unaffected.
				/// See nppiCopyReplicateBorder_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_32s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 1 channel 32-bit floating point image copy with nearest source image pixel color.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 32-bit floating point image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 32-bit floating point image copy with nearest source image pixel color.
				/// See nppiCopyReplicateBorder_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 32-bit floating point image copy with nearest source image pixel color with alpha channel unaffected.
				/// See nppiCopyReplicateBorder_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyReplicateBorder_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);



			}

			/// <summary>
			/// Methods for copying images and padding borders with wrapped replications of the source image pixel colors.
			/// </summary>
			public static class CopyWrapBorder
			{
				/// <summary>
				/// 1 channel 8-bit unsigned integer image copy with the borders wrapped by replication of source image pixel colors.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 8-bit unsigned integer image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 8-bit unsigned integer image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_8u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 8-bit unsigned integer image copy with the borders wrapped by replication of source image pixel colors with alpha channel unaffected.
				/// See nppiCopyWrapBorder_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 1 channel 16-bit unsigned integer image copy with the borders wrapped by replication of source image pixel colors.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_16u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 16-bit unsigned integer image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_16u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 16-bit unsigned integer image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_16u_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 16-bit unsigned integer image copy with the borders wrapped by replication of source image pixel colors with alpha channel unaffected.
				/// See nppiCopyWrapBorder_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 1 channel 16-bit signed integer image copy with the borders wrapped by replication of source image pixel colors.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_16s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 16-bit signed integer image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_16s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 16-bit signed integer image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_16s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 16-bit signed integer image copy with the borders wrapped by replication of source image pixel colors with alpha channel unaffected.
				/// See nppiCopyWrapBorder_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 1 channel 32-bit signed integer image copy with the borders wrapped by replication of source image pixel colors.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_32s_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 32-bit signed integer image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_32s_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 32-bit signed integer image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_32s_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 32-bit signed integer image copy with the borders wrapped by replication of source image pixel colors with alpha channel unaffected.
				/// See nppiCopyWrapBorder_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_32s_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 1 channel 32-bit floating point image copy with the borders wrapped by replication of source image pixel colors.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region of pixels.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nTopBorderHeight">Height (in pixels) of the top border. The number of pixel rows at the top of the</param>
				/// <param name="nLeftBorderWidth">Width (in pixels) of the left border. The width of the border at the right side of the</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 3 channel 32-bit floating point image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 4 channel 32-bit floating point image copy with the borders wrapped by replication of source image pixel colors.
				/// See nppiCopyWrapBorder_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);


				/// <summary>
				/// 1 channel 32-bit floating point image copy with the borders wrapped by replication of source image pixel colors with alpha channel unaffected.
				/// See nppiCopyWrapBorder_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSizeROI">Size of the source region-of-interest.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nTopBorderHeight">Height of top border.</param>
				/// <param name="nLeftBorderWidth">Width of left border.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopyWrapBorder_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcSizeROI, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, int nTopBorderHeight, int nLeftBorderWidth);



			}

			/// <summary>
			/// Functions for copying linearly interpolated images using source image subpixel coordinates
			/// </summary>
			public static class CopySubpix
			{
				/// <summary>
				/// 1 channel 8-bit unsigned integer linearly interpolated source image subpixel coordinate color copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 3 channel 8-bit unsigned integer linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 8-bit unsigned integer linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 8-bit unsigned integer linearly interpolated source image subpixel coordinate color copy with alpha channel unaffected.
				/// See nppiCopySubpix_8u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 1 channel 16-bit unsigned integer linearly interpolated source image subpixel coordinate color copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 3 channel 16-bit unsigned integer linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 16-bit unsigned integer linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 16-bit unsigned linearly interpolated source image subpixel coordinate color copy with alpha channel unaffected.
				/// See nppiCopySubpix_16u_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 1 channel 16-bit signed integer linearly interpolated source image subpixel coordinate color copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 3 channel 16-bit signed integer linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 16-bit signed integer linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 16-bit signed integer linearly interpolated source image subpixel coordinate color copy with alpha channel unaffected.
				/// See nppiCopySubpix_16s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 1 channel 32-bit signed integer linearly interpolated source image subpixel coordinate color copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 3 channel 32-bit signed linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_32s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 32-bit signed integer linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_32s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 32-bit signed integer linearly interpolated source image subpixel coordinate color copy with alpha channel unaffected.
				/// See nppiCopySubpix_32s_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_32s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 1 channel 32-bit floating point linearly interpolated source image subpixel coordinate color copy.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 3 channel 32-bit floating point linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 32-bit floating point linearly interpolated source image subpixel coordinate color copy.
				/// See nppiCopySubpix_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);


				/// <summary>
				/// 4 channel 32-bit floating point linearly interpolated source image subpixel coordinate color copy with alpha channel unaffected.
				/// See nppiCopySubpix_32f_C1R() for detailed documentation.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <param name="nDx">Fractional part of source image X coordinate.</param>
				/// <param name="nDy">Fractional part of source image Y coordinate.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCopySubpix_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, float nDx, float nDy);



			}

			/// <summary>
			/// Functions for duplicating a single channel image in a multiple channel image.
			/// </summary>
			public static class Dup
			{
				/// <summary>
				/// 1 channel 8-bit unsigned integer source image duplicated in all 3 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_8u_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 8-bit unsigned integer source image duplicated in all 4 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_8u_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 8-bit unsigned integer source image duplicated in 3 channels of 4 channel destination image with alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_8u_C1AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 16-bit unsigned integer source image duplicated in all 3 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_16u_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 16-bit unsigned integer source image duplicated in all 4 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_16u_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 16-bit unsigned integer source image duplicated in 3 channels of 4 channel destination image with alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_16u_C1AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 16-bit signed integer source image duplicated in all 3 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_16s_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 16-bit signed integer source image duplicated in all 4 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_16s_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 16-bit signed integer source image duplicated in 3 channels of 4 channel destination image with alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_16s_C1AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 32-bit signed integer source image duplicated in all 3 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_32s_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 32-bit signed integer source image duplicated in all 4 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_32s_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 32-bit signed integer source image duplicated in 3 channels of 4 channel destination image with alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_32s_C1AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 32-bit floating point source image duplicated in all 3 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size (width, height) of the destination region, i.e. the region that gets filled with</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_32f_C1C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 32-bit floating point source image duplicated in all 4 channels of destination image.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_32f_C1C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);


				/// <summary>
				/// 1 channel 32-bit floating point source image duplicated in 3 channels of 4 channel destination image with alpha channel unaffected.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Size of the destination region-of-interest.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDup_32f_C1AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI);



			}

			/// <summary>
			/// Fixed filters perform linear filtering operations (i.e. convolutions) with predefined kernels of fixed sizes.
			/// </summary>
			public static class FixedFilters
			{
				/// <summary>
				/// Single channel 8-bit unsigned horizontal Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned horizontal Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned horizontal Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned horizontal Prewitt filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit signed horizontal Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed horizontal Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed horizontal Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed horizontal Prewitt filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit floating-point horizontal Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit floating-point horizontal Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point horizontal Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point horizontal Prewitt filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittHoriz_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned vertical Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned vertical Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned vertical Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned vertical Prewitt filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit signed vertical Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed vertical Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed vertical Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed vertical Prewitt filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit floating-point vertical Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit floating-point vertical Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point vertical Prewitt filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point vertical Prewitt filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterPrewittVert_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed horizontal Scharr filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterScharrHoriz_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit signed horizontal Scharr filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterScharrHoriz_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit floating-point horizontal Scharr filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterScharrHoriz_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed vertical Scharr filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterScharrVert_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit signed vertical Scharr filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterScharrVert_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit floating-point vertical Scharr filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterScharrVert_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed horizontal Sobel filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit signed horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned horizontal Sobel filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit floating-point horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit floating-point horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point horizontal Sobel filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit signed horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHoriz_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 32-bit floating-point horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHorizMask_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit unsigned vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed vertical Sobel filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit signed vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned vertical Sobel filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit floating-point vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit floating-point vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point vertical Sobel filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit signed vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVert_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 32-bit floating-point vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVertMask_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed second derivative, horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHorizSecond_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit signed second derivative, horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHorizSecond_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 32-bit floating-point second derivative, horizontal Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelHorizSecond_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed second derivative, vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVertSecond_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit signed second derivative, vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVertSecond_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 32-bit floating-point second derivative, vertical Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelVertSecond_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed second cross derivative Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelCross_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit signed second cross derivative Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelCross_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 32-bit floating-point second cross derivative Sobel filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSobelCross_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit unsigned horizontal Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned horizontal Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned horizontal Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned horizontal Roberts filter, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit signed horizontal Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed horizontal Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed horizontal Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed horizontal Roberts filter, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit floating-point horizontal Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit floating-point horizontal Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point horizontal Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point horizontal Roberts filter, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsDown_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 8-bit unsigned vertical Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned vertical Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned vertical Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned vertical Roberts filter, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit signed vertical Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed vertical Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed vertical Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed vertical Roberts filter, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit floating-point vertical Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit floating-point vertical Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point vertical Roberts filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point vertical Roberts filter, ignoring alpha-channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterRobertsUp_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);

				/// <summary>
				/// Single channel 8-bit unsigned Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 8-bit unsigned Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 8-bit unsigned Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 8-bit unsigned Laplace filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 16-bit signed Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 16-bit signed Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit signed Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit signed Laplace filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 32-bit floating-point Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 32-bit floating-point Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 32-bit floating-point Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 32-bit floating-point Laplace filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit unsigned to 16-bit signed Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_8u16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit signed to 16-bit signed Laplace filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLaplace_8s16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);



				/// <summary>
				/// Single channel 8-bit unsigned Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 8-bit unsigned Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 8-bit unsigned Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 8-bit unsigned Gauss filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 16-bit unsigned Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 16-bit unsigned Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit unsigned Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit unsigned Gauss filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 16-bit signed Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 16-bit signed Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit signed Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit signed Gauss filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 32-bit floating-point Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 32-bit floating-point Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 32-bit floating-point Gauss filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 32-bit floating-point Gauss filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterGauss_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit unsigned high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 8-bit unsigned high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 8-bit unsigned high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 8-bit unsigned high-pass filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 16-bit unsigned high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 16-bit unsigned high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit unsigned high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit unsigned high-pass filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 16-bit signed high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 16-bit signed high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit signed high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit signed high-pass filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 32-bit floating-point high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 32-bit floating-point high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 32-bit floating-point high-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 32-bit floating-point high-pass filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterHighPass_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit unsigned low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 8-bit unsigned low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 8-bit unsigned low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 8-bit unsigned low-pass filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 16-bit unsigned low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 16-bit unsigned low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit unsigned low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit unsigned low-pass filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 16-bit signed low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 16-bit signed low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit signed low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 16-bit signed low-pass filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 32-bit floating-point low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Three channel 32-bit floating-point low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 32-bit floating-point low-pass filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Four channel 32-bit floating-point high-pass filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="eMaskSize">Enumeration value specifying the mask size.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterLowPass_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, MaskSize eMaskSize);


				/// <summary>
				/// Single channel 8-bit unsigned sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_8u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 8-bit unsigned sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_8u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_8u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 8-bit unsigned sharpening filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit unsigned sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_16u_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit unsigned sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_16u_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit unsigned sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_16u_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit unsigned sharpening filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_16u_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 16-bit signed sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_16s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 16-bit signed sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_16s_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_16s_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 16-bit signed sharpening filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_16s_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Single channel 32-bit floating-point sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three channel 32-bit floating-point sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_32f_C3R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point sharpening filter.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_32f_C4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four channel 32-bit floating-point sharpening filter, ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFilterSharpen_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI);




			}

			/// <summary>
			/// Returns NppiRect which represents the offset and size of the destination rectangle that would be generated by
			/// resizing the source NppiRect by the requested scale factors and shifts.
			/// </summary>
			public static class GetResizeRect
			{
				/// <summary>
				/// Returns NppiRect which represents the offset and size of the destination rectangle that would be generated by
				/// resizing the source NppiRect by the requested scale factors and shifts.
				/// </summary>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDstRect">User supplied host memory pointer to an NppiRect structure that will be filled in by this function with the region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>\ref image_data_error_codes, \ref roi_error_codes, \ref resize_error_codes</returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiGetResizeRect(NppiRect oSrcROI, ref NppiRect pDstRect, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);

			}

			/// <summary>
			/// Resizes images.
			/// </summary>
			public static class ResizeSqrPixel
			{
				/// <summary>
				/// 1 channel 8-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 8-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned image resize not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 8-bit unsigned planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_8u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 16-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned image resize not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit unsigned planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 16-bit signed image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16s_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit signed image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16s_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit signed image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16s_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit signed image resize not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16s_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit signed planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit signed planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_16s_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 32-bit floating point image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 32-bit floating point image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit floating point image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit floating point image resize not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 32-bit floating point planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit floating point planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_32f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 64-bit floating point image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_64f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 64-bit floating point image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_64f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 64-bit floating point image resize.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_64f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 64-bit floating point image resize not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_64f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 64-bit floating point planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_64f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 64-bit floating point planar image resize.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array (host memory array containing device memory image plane pointers).</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstROI">Region of interest in the destination image.</param>
				/// <param name="nXFactor">Factor by which x dimension is changed. </param>
				/// <param name="nYFactor">Factor by which y dimension is changed. </param>
				/// <param name="nXShift">Source pixel shift in x-direction.</param>
				/// <param name="nYShift">Source pixel shift in y-direction.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref resize_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiResizeSqrPixel_64f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiRect oDstROI, double nXFactor, double nYFactor, double nXShift, double nYShift, InterpolationMode eInterpolation);



			}

			/// <summary>
			/// Remap chooses source pixels using pixel coordinates explicitely supplied in two 2D device memory image arrays pointed to by the pXMap and pYMap pointers.<para/>
			/// The pXMap array contains the X coordinated and the pYMap array contains the Y coordinate of the corresponding source image pixel to
			/// use as input. These coordinates are in floating point format so fraction pixel positions can be used. The coordinates of the source
			/// pixel to sample are determined as follows:<para/>
			/// nSrcX = pxMap[nDstX, nDstY]<para/>
			/// nSrcY = pyMap[nDstX, nDstY]<para/>
			/// In the Remap functions below source image clip checking is handled as follows:<para/>
			/// If the source pixel fractional x and y coordinates are greater than or equal to oSizeROI.x and less than oSizeROI.x + oSizeROI.width and
			/// greater than or equal to oSizeROI.y and less than oSizeROI.y + oSizeROI.height then the source pixel is considered to be within
			/// the source image clip rectangle and the source image is sampled. Otherwise the source image is not sampled and a destination pixel is not
			/// written to the destination image. 
			/// </summary>
			public static class Remap
			{
				/// <summary>
				/// 1 channel 8-bit unsigned image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_8u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 8-bit unsigned image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_8u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_8u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned image remap not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_8u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 8-bit unsigned planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_8u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 8-bit unsigned planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_8u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 16-bit unsigned image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16u_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit unsigned image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16u_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16u_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned image remap not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16u_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit unsigned planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16u_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit unsigned planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16u_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 16-bit signed image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16s_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit signed image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16s_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit signed image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16s_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit signed image remap not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16s_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 16-bit signed planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16s_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 16-bit signed planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_16s_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 32-bit floating point image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_32f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 32-bit floating point image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_32f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit floating point image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_32f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit floating point image remap not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_32f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 32-bit floating point planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_32f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 32-bit floating point planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_32f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 1 channel 64-bit floating point image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_64f_C1R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 64-bit floating point image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_64f_C3R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 64-bit floating point image remap.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_64f_C4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 64-bit floating point image remap not affecting alpha.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of interpolation to perform resampling</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_64f_AC4R(CUdeviceptr pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, CUdeviceptr pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 3 channel 64-bit floating point planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_64f_P3R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);


				/// <summary>
				/// 4 channel 64-bit floating point planar image remap.
				/// </summary>
				/// <param name="pSrc">\ref source_planar_image_pointer_array.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcSize">Size in pixels of the source image.</param>
				/// <param name="oSrcROI">Region of interest in the source image.</param>
				/// <param name="pXMap">Device memory pointer to 2D image array of X coordinate values to be used when sampling source image. </param>
				/// <param name="nXMapStep">pXMap image array line step in bytes.</param>
				/// <param name="pYMap">Device memory pointer to 2D image array of Y coordinate values to be used when sampling source image. </param>
				/// <param name="nYMapStep">pYMap image array line step in bytes.</param>
				/// <param name="pDst">\ref destination_planar_image_pointer_array.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oDstSizeROI">Region of interest size in the destination image.</param>
				/// <param name="eInterpolation">The type of eInterpolation to perform resampling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, \ref remap_error_codes
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRemap_64f_P4R([In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pSrc, NppiSize oSrcSize, int nSrcStep, NppiRect oSrcROI, CUdeviceptr pXMap, int nXMapStep, CUdeviceptr pYMap, int nYMapStep, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 4)] CUdeviceptr[] pDst, int nDstStep, NppiSize oDstSizeROI, InterpolationMode eInterpolation);



			}

			/// <summary>
			/// Primitives for computing the relative error between two images.
			/// </summary>
			public static class NormRel
			{
				/// <summary>
				/// One-channel 8-bit unsigned image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit signed image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an 
				/// invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit signed image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormRel_Inf ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormRel_Inf ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormRel_Inf ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormRel_Inf ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit unsigned image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_8u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit signed image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_8s_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 16-bit unsigned image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 32-bit floating point image NormRel_Inf.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_32f_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit unsigned image NormRel_Inf affecting only signle channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_COI_ERROR if an 
				/// invalid channel of interest is specified, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_8u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit signed image NormRel_Inf affecting only signle channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_COI_ERROR if an invalid 
				/// channel of interest is specified, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_8s_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 16-bit unsigned image NormRel_Inf affecting only signle channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_COI_ERROR if an invalid 
				/// channel of interest is specified, or NPP_DIVISOR_ERROR if the infinity
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_16u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 32-bit floating point image NormRel_Inf affecting only signle channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the infinity norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_COI_ERROR if an invalid channel 
				/// of interest is specified, or NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point 
				/// image is specified, or NPP_DIVISOR_ERROR if the infinity norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_Inf_32f_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_32s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_32s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_Inf_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelInfGetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// One-channel 8-bit unsigned image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit signed image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an 
				/// invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit signed image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit signed image NormRel_L1 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormRel_L1 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormRel_L1 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormRel_L1 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 8-bit unsigned image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_8u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 8-bit signed image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_8s_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image NormRel_L1.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L1
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_32f_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit unsigned image NormRel_L1 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid 
				/// channel of interest is specified, or NPP_DIVISOR_ERROR if the L1 norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_8u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit signed image NormRel_L1 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid 
				/// channel of interest is specified, or NPP_DIVISOR_ERROR if the L1 norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_8s_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 16-bit unsigned image NormRel_L1 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid 
				/// channel of interest is specified, or NPP_DIVISOR_ERROR if the L1 norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_16u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 32-bit floating point image NormRel_L1 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L1 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid channel 
				/// of interest is specified, NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point image is specified, 
				/// or NPP_DIVISOR_ERROR if the L1 norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L1_32f_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L1_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL1GetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// One-channel 8-bit unsigned image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_8u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16u_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit signed image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16s_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if an 
				/// invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_8u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16u_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit signed image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16s_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormRel_L2 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_8u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormRel_L2 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16u_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormRel_L2 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16s_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormRel_L2 ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_8u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16u_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16s_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aNormRel">Array that contains the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_32f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit unsigned image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_8u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 8-bit signed image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_8s_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 16-bit unsigned image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16u_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked one-channel 32-bit floating point image NormRel_L2.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_NOT_EVEN_STEP_ERROR if 
				/// an invalid floating-point image is specified, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_32f_C1MR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit unsigned image NormRel_L2 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid 
				/// channel of interest is specified, or NPP_DIVISOR_ERROR if the L2 norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_8u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 8-bit signed image NormRel_L2 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_COI_ERROR if an invalid 
				/// channel of interest is specified, or NPP_DIVISOR_ERROR if the L2 norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_8s_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 16-bit unsigned image NormRel_L2 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_COI_ERROR if an 
				/// invalid channel of interest is specified, or NPP_DIVISOR_ERROR if the L2
				/// norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_16u_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Masked three-channel 32-bit floating point image NormRel_L2 affecting only single channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="pMask">\ref mask_image_pointer.</param>
				/// <param name="nMaskStep">\ref mask_image_line_step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="nCOI">\ref channel_of_interest_number.</param>
				/// <param name="pNormRel">Pointer to the computed relative error for the L2 norm of two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, NPP_COI_ERROR if an invalid channel 
				/// of interest is specified, or NPP_NOT_EVEN_STEP_ERROR if an invalid floating-point image is specified, 
				/// or NPP_DIVISOR_ERROR if the L2 norm of the second image is zero.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRel_L2_32f_C3CMR(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, CUdeviceptr pMask, int nMaskStep, NppiSize oSizeROI, int nCOI, CUdeviceptr pNormRel, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16s_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16s_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_8u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_8u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_8s_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_8s_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16u_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16u_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_32f_C1MR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_32f_C1MR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16s_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16s_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_8u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_8u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16u_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16u_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16s_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16s_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16s_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16s_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_8u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_8u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_8s_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_8s_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_16u_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_16u_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Computes the device scratch buffer size (in bytes) for nppiNormRel_L2_32f_C3CMR.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiNormRelL2GetBufferHostSize_32f_C3CMR(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);



			}

			/// <summary>
			/// Primitives for computing the dot product of two images.
			/// </summary>
			public static class DotProd
			{
				/// <summary>
				/// One-channel 8-bit unsigned image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDp">Pointer to the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_8u64f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 8-bit signed image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDp">Pointer to the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_8s64f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDp">Pointer to the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_16u64f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit signed image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDp">Pointer to the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_16s64f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit unsigned image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDp">Pointer to the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32u64f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit signed image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDp">Pointer to the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32s64f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pDp">Pointer to the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32f64f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_8u64f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit signed image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_8s64f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_16u64f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit signed image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_16s64f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit unsigned image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32u64f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit signed image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32s64f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32f64f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_8u64f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit signed image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_8s64f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_16u64f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_16s64f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit unsigned image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32u64f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit signed image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32s64f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image DotProd.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32f64f_C4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image DotProd ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_8u64f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit signed image DotProd ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_8s64f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image DotProd ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed Inf-norm of differences. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_16u64f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit signed image DotProd ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_16s64f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit unsigned image DotProd ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32u64f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit signed image DotProd ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32s64f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image DotProd ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aDp">Array that contains the computed dot product of the two images. </param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProd_32f64f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oSizeROI, CUdeviceptr aDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_8u64f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_8u64f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_8s64f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_8s64f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_16u64f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_16u64f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_16s64f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_16s64f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32u64f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32u64f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32s64f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32s64f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32f64f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32f64f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_8u64f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_8u64f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_8s64f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_8s64f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_16u64f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_16u64f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_16s64f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_16s64f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32u64f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32u64f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32s64f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32s64f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32f64f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32f64f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_8u64f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_8u64f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_8s64f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_8s64f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_16u64f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_16u64f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_16s64f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_16s64f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32u64f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32u64f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32s64f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32s64f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32f64f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32f64f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_8u64f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_8u64f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_8s64f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_8s64f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_16u64f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_16u64f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_16s64f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_16s64f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32u64f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32u64f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32s64f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32s64f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiDotProd_32f64f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiDotProdGetBufferHostSize_32f64f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);



			}

			/// <summary>
			/// Primitives for computing the amount of pixels that fall into the specified intensity range.
			/// The lower bound and the upper bound are inclusive.
			/// </summary>
			public static class CountInRange
			{
				/// <summary>
				/// One-channel 8-bit unsigned image CountInRange.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pCounts">Pointer to the number of pixels that fall into the specified range.</param>
				/// <param name="nLowerBound">Lower bound of the specified range.</param>
				/// <param name="nUpperBound">Upper bound of the specified range.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_RANGE_ERROR if the lower bound
				/// is larger than the upper bound.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRange_8u_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pCounts, byte nLowerBound, byte nUpperBound, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image CountInRange.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="pCounts">Pointer to the number of pixels that fall into the specified range.</param>
				/// <param name="nLowerBound">Lower bound of the specified range.</param>
				/// <param name="nUpperBound">Upper bound of the specified range.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_RANGE_ERROR if the lower bound
				/// is larger than the upper bound.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRange_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr pCounts, float nLowerBound, float nUpperBound, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image CountInRange.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCounts">Array that contains the number of pixels that fall into the specified range for Three-channels.</param>
				/// <param name="aLowerBound">Fixed size array of the lower bound of the specified range, one per channel.</param>
				/// <param name="aUpperBound">Fixed size array of the upper bound of the specified range, one per channel.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_RANGE_ERROR if the lower bound
				/// is larger than the upper bound.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRange_8u_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aCounts, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aLowerBound, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aUpperBound, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image CountInRange.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCounts">Array that contains the number of pixels that fall into the specified range for Three-channels.</param>
				/// <param name="aLowerBound">Fixed size array of the lower bound of the specified range, one per channel.</param>
				/// <param name="aUpperBound">Fixed size array of the upper bound of the specified range, one per channel.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_RANGE_ERROR if the lower bound
				/// is larger than the upper bound.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRange_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aCounts, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aLowerBound, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aUpperBound, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CountInRange ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCounts">Array that contains the number of pixels that fall into the specified range for Three-channels.</param>
				/// <param name="aLowerBound">Fixed size array of the lower bound of the specified range, one per channel.</param>
				/// <param name="aUpperBound">Fixed size array of the upper bound of the specified range, one per channel.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_RANGE_ERROR if the lower bound
				/// is larger than the upper bound.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRange_8u_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aCounts, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aLowerBound, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] byte[] aUpperBound, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image CountInRange ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="aCounts">Array that contains the number of pixels that fall into the specified range for Three-channels.</param>
				/// <param name="aLowerBound">Fixed size array of the lower bound of the specified range, one per channel.</param>
				/// <param name="aUpperBound">Fixed size array of the upper bound of the specified range, one per channel.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_RANGE_ERROR if the lower bound
				/// is larger than the upper bound.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRange_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSizeROI, CUdeviceptr aCounts, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aLowerBound, [In, MarshalAs(UnmanagedType.LPArray, SizeConst = 3)] float[] aUpperBound, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiCountInRange_8u_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRangeGetBufferHostSize_8u_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiCountInRange_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRangeGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiCountInRange_8u_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRangeGetBufferHostSize_8u_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiCountInRange_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRangeGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiCountInRange_8u_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRangeGetBufferHostSize_8u_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppiCountInRange_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCountInRangeGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);



			}

			/// <summary>
			/// Primitives for computing the minimal/maximal value of the pixel pair from two images.
			/// </summary>
			public static class MinMaxEvery
			{
				/// <summary>
				/// One-channel 8-bit unsigned image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_8u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One-channel 16-bit unsigned image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_16u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One-channel 16-bit signed image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_16s_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One-channel 32-bit floating point image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 8-bit unsigned image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_8u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit unsigned image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_16u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit signed image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_16s_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit floating point image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_32f_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_8u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_16u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit signed image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_16s_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating point image MaxEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_32f_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned image MaxEvery ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_8u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned image MaxEvery ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_16u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit signed image MaxEvery ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_16s_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating point image MaxEvery ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMaxEvery_32f_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One-channel 8-bit unsigned image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_8u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One-channel 16-bit unsigned image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_16u_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One-channel 16-bit signed image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_16s_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// One-channel 32-bit floating point image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_32f_C1IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 8-bit unsigned image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_8u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit unsigned image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_16u_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 16-bit signed image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_16s_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Three-channel 32-bit floating point image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_32f_C3IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_8u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_16u_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit signed image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_16s_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating point image MinEvery.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_32f_C4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 8-bit unsigned image MinEvery ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_8u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit unsigned image MinEvery ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_16u_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 16-bit signed image MinEvery ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_16s_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);


				/// <summary>
				/// Four-channel 32-bit floating point image MinEvery ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSrcDst">In-Place Image Pointer.</param>
				/// <param name="nSrcDstStep">Source-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiMinEvery_32f_AC4IR(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSrcDst, int nSrcDstStep, NppiSize oSizeROI);



			}

			/// <summary>
			/// Primitives for computing the integral image of a given image.
			/// </summary>
			public static class Integral
			{
				/// <summary>
				/// One-channel 8-bit unsigned image Integral with 32-bit signed output.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="nVal">The value to add to pDst image pixels</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiIntegral_8u32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, int nVal);


				/// <summary>
				/// One-channel 8-bit unsigned image Integral with 32-bit floating point output.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oROI">Region-of-Interest (ROI).</param>
				/// <param name="nVal">The value to add to pDst image pixels</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiIntegral_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, NppiSize oROI, float nVal);


				/// <summary>
				/// One-channel 8-bit unsigned image SqrIntegral. 
				/// Destination integral image and square integral image are 32-bit signed int.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pSqr">Destination-Image Pointer.</param>
				/// <param name="nSqrStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <param name="nVal">The value to add to pDst image pixels</param>
				/// <param name="nValSqr">The value to add to pSqr image pixels</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrIntegral_8u32s_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, CUdeviceptr pSqr, int nSqrStep, NppiSize oSrcROI, int nVal, int nValSqr);


				/// <summary>
				/// One-channel 8-bit unsigned image SqrIntegral. 
				/// Destination integral image is 32-bit signed int. 
				/// Destination square integral image is 64-bit double floating point.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pSqr">Destination-Image Pointer.</param>
				/// <param name="nSqrStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <param name="nVal">The value to add to pDst image pixels</param>
				/// <param name="nValSqr">The value to add to pSqr image pixels</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrIntegral_8u32s64f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, CUdeviceptr pSqr, int nSqrStep, NppiSize oSrcROI, int nVal, double nValSqr);


				/// <summary>
				/// One-channel 8-bit unsigned image SqrIntegral. 
				/// Destination integral image is 32-bit floating point.
				/// Destination square integral image is 64-bit double floating point.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pSqr">Destination-Image Pointer.</param>
				/// <param name="nSqrStep">Destination-Image Line Step.</param>
				/// <param name="oSrcROI">Region-of-Interest (ROI).</param>
				/// <param name="nVal">The value to add to pDst image pixels</param>
				/// <param name="nValSqr">The value to add to pSqr image pixels</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrIntegral_8u32f64f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pDst, int nDstStep, CUdeviceptr pSqr, int nSqrStep, NppiSize oSrcROI, float nVal, double nValSqr);


				/// <summary>
				/// One-channel 32-bit floating point image RectStdDev.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSqr">Destination-Image Pointer.</param>
				/// <param name="nSqrStep">Destination-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oRect">rectangular window</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRectStdDev_32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSqr, int nSqrStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiRect oRect);


				/// <summary>
				/// One-channel 32-bit signed image RectStdDev, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSqr">Destination-Image Pointer.</param>
				/// <param name="nSqrStep">Destination-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oRect">rectangular window</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRectStdDev_32s_C1RSfs(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSqr, int nSqrStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiRect oRect, int nScaleFactor);


				/// <summary>
				/// One-channel 32-bit signed image RectStdDev. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="pSqr">Destination-Image Pointer.</param>
				/// <param name="nSqrStep">Destination-Image Line Step.</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="oRect">rectangular window</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiRectStdDev_32s32f_C1R(CUdeviceptr pSrc, int nSrcStep, CUdeviceptr pSqr, int nSqrStep, CUdeviceptr pDst, int nDstStep, NppiSize oSizeROI, NppiRect oRect);



			}

			/// <summary>
			/// Primitives for computing the proximity measure between a source image and a template image.
			/// </summary>
			public static class ImageProximity
			{
				/// <summary>
				/// One-channel 8-bit unsigned image SqrDistanceFull_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Three-channel 8-bit unsigned image SqrDistanceFull_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceFull_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceFull_Norm ignoring alpha channel, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// One-channel 32-bit floating point image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 32-bit floating point image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image SqrDistanceFull_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit unsigned image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceFull_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit signed image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit signed image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image SqrDistanceFull_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 16-bit unsigned image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 16-bit unsigned image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image SqrDistanceFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image SqrDistanceFull_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceFull_Norm_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image SqrDistanceSame_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Three-channel 8-bit unsigned image SqrDistanceSame_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceSame_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceSame_Norm ignoring alpha channel, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// One-channel 32-bit floating point image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 32-bit floating point image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image SqrDistanceSame_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit unsigned image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceSame_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit signed image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit signed image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image SqrDistanceSame_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 16-bit unsigned image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 16-bit unsigned image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image SqrDistanceSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image SqrDistanceSame_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceSame_Norm_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image SqrDistanceValid_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Three-channel 8-bit unsigned image SqrDistanceValid_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceValid_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceValid_Norm ignoring alpha channel, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// One-channel 32-bit floating point image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 32-bit floating point image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image SqrDistanceValid_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit unsigned image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image SqrDistanceValid_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit signed image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit signed image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image SqrDistanceValid_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 16-bit unsigned image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 16-bit unsigned image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image SqrDistanceValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image SqrDistanceValid_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSqrDistanceValid_Norm_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrFull_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrFull_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrFull_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrFull_Norm ignoring alpha channel, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// One-channel 32-bit floating point image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 32-bit floating point image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrFull_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrFull_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit signed image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit signed image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrFull_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 16-bit unsigned image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 16-bit unsigned image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrFull_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrFull_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_Norm_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrSame_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrSame_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrSame_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrSame_Norm ignoring alpha channel, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// One-channel 32-bit floating point image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 32-bit floating point image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrSame_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrSame_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit signed image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit signed image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrSame_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 16-bit unsigned image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 16-bit unsigned image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrSame_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrSame_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_Norm_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrValid_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrValid_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrValid_Norm, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrValid_Norm ignoring alpha channel, scaled by \f$2^(-nScaleFactor)\f$. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor);


				/// <summary>
				/// One-channel 32-bit floating point image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 32-bit floating point image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrValid_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrValid_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit signed image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 8-bit signed image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrValid_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 16-bit unsigned image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Three-channel 16-bit unsigned image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrValid_Norm. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrValid_Norm ignoring alpha channel. 
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_Norm_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 32-bit floating point images CrossCorrValid.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned images CrossCorrValid.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit signed images CrossCorrValid.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 16-bit unsigned images CrossCorrValid.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrFull_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrFull_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrFull_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 8-bit signed image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit signed image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrFull_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrFull_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrFull_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrFull_NormLevel_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Buffer size (in bytes) for nppiCrossCorrFull_NormLevel_8u_C1RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8u_C1RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8u_C3RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8u_C3RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8u_C4RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8u_C4RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8u_AC4RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8u_AC4RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8u32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8u32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8u32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8u32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8u32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8u32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8u32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8u32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8s32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8s32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8s32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8s32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8s32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8s32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_8s32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_8s32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_16u32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_16u32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_16u32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_16u32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_16u32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_16u32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrFull_NormLevel_16u32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiFullNormLevelGetBufferHostSize_16u32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrSame_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrSame_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrSame_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 8-bit signed image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit signed image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrSame_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrSame_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrSame_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrSame_NormLevel_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Buffer size (in bytes) for nppiCrossCorrSame_NormLevel_8u_C1RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8u_C1RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8u_C3RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8u_C3RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8u_C4RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8u_C4RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8u_AC4RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8u_AC4RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8u32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8u32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8u32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8u32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8u32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8u32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8u32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8u32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8s32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8s32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8s32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8s32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8s32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8s32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_8s32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_8s32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_16u32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_16u32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_16u32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_16u32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_16u32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_16u32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrSame_NormLevel_16u32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiSameNormLevelGetBufferHostSize_16u32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8u_C1RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8u_C3RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8u_C4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrValid_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8u_AC4RSfs(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image CrossCorrValid_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 8-bit unsigned image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image CrossCorrValid_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 8-bit signed image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8s32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit signed image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8s32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8s32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit signed image CrossCorrValid_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_8s32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_16u32f_C1R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_16u32f_C3R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrValid_NormLevel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_16u32f_C4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image CrossCorrValid_NormLevel ignoring alpha channel.
				/// </summary>
				/// <param name="pSrc">Source-Image Pointer.</param>
				/// <param name="nSrcStep">Source-Image Line Step.</param>
				/// <param name="oSrcRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pTpl">Pointer to the template image.</param>
				/// <param name="nTplStep">Number of bytes between successive rows in the template image.</param>
				/// <param name="oTplRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Destination-Image Pointer.</param>
				/// <param name="nDstStep">Destination-Image Line Step.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiCrossCorrValid_NormLevel_16u32f_AC4R(CUdeviceptr pSrc, int nSrcStep, NppiSize oSrcRoiSize, CUdeviceptr pTpl, int nTplStep, NppiSize oTplRoiSize, CUdeviceptr pDst, int nDstStep, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Buffer size (in bytes) for nppiCrossCorrValid_NormLevel_8u_C1RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8u_C1RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8u_C3RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8u_C3RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8u_C4RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8u_C4RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8u_AC4RSfs.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8u_AC4RSfs(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8u32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8u32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8u32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8u32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8u32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8u32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8u32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8u32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8s32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8s32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8s32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8s32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8s32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8s32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_8s32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_8s32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_16u32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_16u32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_16u32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_16u32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_16u32f_C4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_16u32f_C4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiCrossCorrValid_NormLevel_16u32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiValidNormLevelGetBufferHostSize_16u32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);



			}

			/// <summary>
			/// Primitives for computing the image quality index of two images.
			/// </summary>
			public static class QualityIndex
			{
				/// <summary>
				/// One-channel 8-bit unsigned image QualityIndex.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Pointer to the quality index.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_QUALITY_INDEX_ERROR if pixels of 
				/// either image are constant numberse.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndex_8u32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oRoiSize, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 16-bit unsigned image QualityIndex.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Pointer to the quality index.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_QUALITY_INDEX_ERROR if pixels of 
				/// either image are constant numberse.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndex_16u32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oRoiSize, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// One-channel 32-bit floating point image QualityIndex.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Pointer to the quality index.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_QUALITY_INDEX_ERROR if pixels of 
				/// either image are constant numberse.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndex_32f_C1R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oRoiSize, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 8-bit unsigned image QualityIndex.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Pointer to the quality index.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_QUALITY_INDEX_ERROR if pixels of 
				/// either image are constant numberse.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndex_8u32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oRoiSize, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 16-bit unsigned image QualityIndex.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Pointer to the quality index.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_QUALITY_INDEX_ERROR if pixels of 
				/// either image are constant numberse.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndex_16u32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oRoiSize, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Three-channel 32-bit floating point image QualityIndex.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Pointer to the quality index.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_QUALITY_INDEX_ERROR if pixels of 
				/// either image are constant numberse.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndex_32f_C3R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oRoiSize, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 8-bit unsigned image QualityIndex.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Pointer to the quality index.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_QUALITY_INDEX_ERROR if pixels of 
				/// either image are constant numberse.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndex_8u32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oRoiSize, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 16-bit unsigned image QualityIndex.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Pointer to the quality index.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_QUALITY_INDEX_ERROR if pixels of 
				/// either image are constant numberse.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndex_16u32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oRoiSize, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Four-channel 32-bit floating point image QualityIndex.
				/// </summary>
				/// <param name="pSrc1">Source-Image Pointer.</param>
				/// <param name="nSrc1Step">Source-Image Line Step.</param>
				/// <param name="pSrc2">Source-Image Pointer.</param>
				/// <param name="nSrc2Step">Source-Image Line Step.</param>
				/// <param name="oRoiSize">Region-of-Interest (ROI).</param>
				/// <param name="pDst">Pointer to the quality index.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.StepError"/>, <see cref="NppStatus.NotEvenStepError"/>, <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>, or NPP_QUALITY_INDEX_ERROR if pixels of 
				/// either image are constant numberse.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndex_32f_AC4R(CUdeviceptr pSrc1, int nSrc1Step, CUdeviceptr pSrc2, int nSrc2Step, NppiSize oRoiSize, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiQualityIndex_8u32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndexGetBufferHostSize_8u32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiQualityIndex_16u32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndexGetBufferHostSize_16u32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiQualityIndex_32f_C1R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndexGetBufferHostSize_32f_C1R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiQualityIndex_8u32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndexGetBufferHostSize_8u32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiQualityIndex_16u32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndexGetBufferHostSize_16u32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiQualityIndex_32f_C3R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndexGetBufferHostSize_32f_C3R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiQualityIndex_8u32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndexGetBufferHostSize_8u32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiQualityIndex_16u32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndexGetBufferHostSize_16u32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);


				/// <summary>
				/// Buffer size (in bytes) for \ref nppiQualityIndex_32f_AC4R.
				/// </summary>
				/// <param name="oSizeROI">Region-of-Interest (ROI).</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_NULL_POINTER_ERROR if hpBufferSize is 0 (NULL),  <see cref="NppStatus.StepError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPI_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppiQualityIndexGetBufferHostSize_32f_AC4R(NppiSize oSizeROI, ref int hpBufferSize/* host pointer */);



			}

			
		}

		/// <summary>
		/// npps.h
		/// </summary>
		public static class NPPs
		{
			/// <summary>
			/// Signal-allocator methods for allocating 1D arrays of data in device memory.
			/// All allocators have size parameters to specify the size of the signal (1D array)
			/// being allocated.
			/// <para/>
			/// The allocator methods return a pointer to the newly allocated memory of appropriate
			/// type. If device-memory allocation is not possible due to resource constaints
			/// the allocators return 0 (i.e. NULL pointer). 
			/// <para/>
			/// All signal allocators allocate memory aligned such that it is  beneficial to the 
			/// performance of the majority of the signal-processing primitives. 
			/// It is no mandatory however to use these allocators. Any valid
			/// CUDA device-memory pointers can be passed to NPP primitives. 
			/// </summary>
			public static class MemAlloc
			{
				/// <summary>
				/// 8-bit unsigned signal allocator.
				/// </summary>
				/// <param name="nSize">Number of unsigned chars in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_8u(int nSize);

				/// <summary>
				/// 8-bit signed signal allocator.
				/// </summary>
				/// <param name="nSize">Number of (signed) chars in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_8s(int nSize);


				/// <summary>
				/// 16-bit unsigned signal allocator.
				/// </summary>
				/// <param name="nSize">Number of unsigned shorts in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_16u(int nSize);


				/// <summary>
				/// 16-bit signal allocator.
				/// </summary>
				/// <param name="nSize">Number of shorts in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_16s(int nSize);


				/// <summary>
				/// 16-bit complex-value signal allocator.
				/// </summary>
				/// <param name="nSize">Number of 16-bit complex numbers in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_16sc(int nSize);


				/// <summary>
				/// 32-bit unsigned signal allocator.
				/// </summary>
				/// <param name="nSize">Number of unsigned ints in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_32u(int nSize);


				/// <summary>
				/// 32-bit integer signal allocator.
				/// </summary>
				/// <param name="nSize">Number of ints in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_32s(int nSize);


				/// <summary>
				/// 32-bit complex integer signal allocator.
				/// </summary>
				/// <param name="nSize">Number of complex integner values in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_32sc(int nSize);


				/// <summary>
				/// 32-bit float signal allocator.
				/// </summary>
				/// <param name="nSize">Number of floats in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_32f(int nSize);


				/// <summary>
				/// 32-bit complex float signal allocator.
				/// </summary>
				/// <param name="nSize">Number of complex float values in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_32fc(int nSize);


				/// <summary>
				/// 64-bit long integer signal allocator.
				/// </summary>
				/// <param name="nSize">Number of long ints in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_64s(int nSize);


				/// <summary>
				/// 64-bit complex long integer signal allocator.
				/// </summary>
				/// <param name="nSize">Number of complex long int values in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_64sc(int nSize);


				/// <summary>
				/// 64-bit float (double) signal allocator.
				/// </summary>
				/// <param name="nSize">Number of doubles in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_64f(int nSize);


				/// <summary>
				/// 64-bit complex complex signal allocator.
				/// </summary>
				/// <param name="nSize">Number of complex double valuess in the new signal.</param>
				/// <returns>
				/// A pointer to the new signal. 0 (NULL-pointer) indicates
				///         that an error occurred during allocation.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern CUdeviceptr nppsMalloc_64fc(int nSize);


				/// <summary>
				/// Free method for any 2D allocated memory.
				/// This method should be used to free memory allocated with any of the
				/// nppiMalloc_<![CDATA[<modifier>]]> methods.
				/// </summary>
				/// <param name="pValues">A pointer to memory allocated using nppiMalloc_<![CDATA[<modifier>]]>.</param>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern void nppsFree(CUdeviceptr pValues);




			}

			/// <summary>
			/// Set methods for 1D vectors of various types. The copy methods operate on vector data given
			/// as a pointer to the underlying data-type (e.g. 8-bit vectors would
			/// be passed as pointers to Npp8u type) and length of the vectors, i.e. the number of items.
			/// </summary>
			public static class Set
			{
				/// <summary>
				/// 8-bit unsigned char, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_8u(byte nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit integer, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_16s(short nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit integer complex, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_16sc(Npp16sc nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit integer, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_32s(int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit integer complex, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_32sc(Npp32sc nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit float, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_32f(float nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit float complex, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_32fc(Npp32fc nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit long long integer, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_64s(long nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit long long integer complex, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_64sc(Npp64sc nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit double, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_64f(double nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit double complex, vector set method.
				/// </summary>
				/// <param name="nValue">Value used to initialize the vector pDst.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSet_64fc(Npp64fc nValue, CUdeviceptr pDst, int nLength);



			}

			/// <summary>
			/// Set signals to zero.
			/// </summary>
			public static class Zero
			{
				/// <summary>
				/// 8-bit unsigned char, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_8u(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit integer, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_16s(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit integer complex, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_16sc(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit integer, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_32s(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit integer complex, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_32sc(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit float, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_32f(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit float complex, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_32fc(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit long long integer, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_64s(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit long long integer complex, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_64sc(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit double, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_64f(CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit double complex, vector zero method.
				/// </summary>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZero_64fc(CUdeviceptr pDst, int nLength);



			}

			/// <summary>
			/// Copy methods for various type signals. Copy methods operate on
			/// signal data given as a pointer to the underlying data-type (e.g. 8-bit
			/// vectors would be passed as pointers to Npp8u type) and length of the
			/// vectors, i.e. the number of items. 
			/// </summary>
			public static class Copy
			{
				/// <summary>
				/// 8-bit unsigned char, vector copy method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_8u(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit signed short, vector copy method.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_16s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit signed integer, vector copy method.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_32s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit float, vector copy method.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit signed integer, vector copy method.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_64s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit complex short, vector copy method.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_16sc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit complex signed integer, vector copy method.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_32sc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit complex float, vector copy method.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_32fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit complex signed integer, vector copy method.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_64sc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit complex double, vector copy method.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCopy_64fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);



			}

			/// <summary>
			/// Adds a constant value to each sample of a signal.
			/// </summary>
			public static class AddC
			{
				/// <summary>
				/// 8-bit unsigned char in place signal add constant,
				/// scale, then clamp to saturated value
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_8u_ISfs(byte nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 8-bit unsigned charvector add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_8u_Sfs(CUdeviceptr pSrc, byte nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_16u_ISfs(ushort nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short vector add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_16u_Sfs(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place  signal add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_16s_ISfs(short nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal add constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_16s_Sfs(CUdeviceptr pSrc, short nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary)signal add constant, 
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_16sc_ISfs(Npp16sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary) signal add constant,
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_16sc_Sfs(CUdeviceptr pSrc, Npp16sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer in place signal add constant and scale.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_32s_ISfs(int nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integersignal add constant and scale.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_32s_Sfs(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit integer complex number (32 bit real, 32 bit imaginary) in place signal
				/// add constant and scale.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_32sc_ISfs(Npp32sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit integer complex number (32 bit real, 32 bit imaginary) signal add constant
				/// and scale.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_32sc_Sfs(CUdeviceptr pSrc, Npp32sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit floating point in place signal add constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_32f_I(float nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal add constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_32f(CUdeviceptr pSrc, float nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) in
				/// place signal add constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_32fc_I(Npp32fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) signal
				/// add constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_32fc(CUdeviceptr pSrc, Npp32fc nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point, in place signal add constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Length of the vectors, number of items.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_64f_I(double nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating pointsignal add constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_64f(CUdeviceptr pSrc, double nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) in
				/// place signal add constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_64fc_I(Npp64fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) signal
				/// add constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be added to each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddC_64fc(CUdeviceptr pSrc, Npp64fc nValue, CUdeviceptr pDst, int nLength);



			}

			/// <summary>
			/// Adds product of a constant and each sample of a source signal to the each sample of destination signal.
			/// </summary>
			public static class AddProductC
			{
				///// <summary>
				///// 8-bit unsigned char in place signal add product of signal times constant to destination signal,
				///// scale, then clamp to saturated value
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_8u_ISfs(byte nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 8-bit unsigned char add product of signal times constant to destination signal, scale, then clamp to saturated value.
				///// </summary>
				///// <param name="pSrc">Source signal pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="pDst">Destination signal pointer.</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_8u_Sfs(CUdeviceptr pSrc, byte nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 16-bit unsigned short in place signal add product of signal times constant to destination signal, scale, then clamp to saturated value.
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_16u_ISfs(ushort nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 16-bit unsigned short add product of signal times constant to destination signal, scale, then clamp to saturated value.
				///// </summary>
				///// <param name="pSrc">Source signal pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="pDst">Destination signal pointer.</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_16u_Sfs(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 16-bit signed short in place  signal add product of signal times constant to destination signal, scale, then clamp to saturated value.
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_16s_ISfs(short nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 16-bit signed short signal add product of signal times constant to destination signal, scale, then clamp to saturated value.
				///// </summary>
				///// <param name="pSrc">Source signal pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="pDst">Destination signal pointer.</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_16s_Sfs(CUdeviceptr pSrc, short nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 16-bit integer complex number (16 bit real, 16 bit imaginary)signal add product of signal times constant to destination signal, 
				///// scale, then clamp to saturated value.
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_16sc_ISfs(Npp16sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 16-bit integer complex number (16 bit real, 16 bit imaginary)signal add product of signal times constant to destination signal,
				///// scale, then clamp to saturated value.
				///// </summary>
				///// <param name="pSrc">Source signal pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="pDst">Destination signal pointer.</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_16sc_Sfs(CUdeviceptr pSrc, Npp16sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 32-bit signed integer in place signal add product of signal times constant to destination signal and scale.
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_32s_ISfs(int nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 32-bit signed integer signal add product of signal times constant to destination signal and scale.
				///// </summary>
				///// <param name="pSrc">Source signal pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="pDst">Destination signal pointer.</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_32s_Sfs(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 32-bit integer complex number (32 bit real, 32 bit imaginary) in place signal
				///// add product of signal times constant to destination signal and scale.
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_32sc_ISfs(Npp32sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 32-bit integer complex number (32 bit real, 32 bit imaginary)signal add product of signal times constant to destination signal
				///// and scale.
				///// </summary>
				///// <param name="pSrc">Source signal pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="pDst">Destination signal pointer.</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <param name="nScaleFactor">Integer Result Scaling.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_32sc_Sfs(CUdeviceptr pSrc, Npp32sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				///// <summary>
				///// 32-bit floating point in place signal add product of signal times constant to destination signal.
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_32f_I(float nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal add product of signal times constant to destination signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddProductC_32f(CUdeviceptr pSrc, float nValue, CUdeviceptr pDst, int nLength);


				///// <summary>
				///// 32-bit floating point complex number (32 bit real, 32 bit imaginary) in
				///// place signal add product of signal times constant to destination signal.
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_32fc_I(Npp32fc nValue, CUdeviceptr pSrcDst, int nLength);


				///// <summary>
				///// 32-bit floating point complex number (32 bit real, 32 bit imaginary)signal
				///// add product of signal times constant to destination signal.
				///// </summary>
				///// <param name="pSrc">Source signal pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="pDst">Destination signal pointer.</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_32fc(CUdeviceptr pSrc, Npp32fc nValue, CUdeviceptr pDst, int nLength);


				///// <summary>
				///// 64-bit floating point, in place signal add product of signal times constant to destination signal.
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Length of the vectors, number of items.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_64f_I(double nValue, CUdeviceptr pSrcDst, int nLength);


				///// <summary>
				///// 64-bit floating point signal add product of signal times constant to destination signal.
				///// </summary>
				///// <param name="pSrc">Source signal pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="pDst">Destination signal pointer.</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_64f(CUdeviceptr pSrc, double nValue, CUdeviceptr pDst, int nLength);


				///// <summary>
				///// 64-bit floating point complex number (64 bit real, 64 bit imaginary) in
				///// place signal add product of signal times constant to destination signal.
				///// </summary>
				///// <param name="pSrcDst">In-Place Signal Pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_64fc_I(Npp64fc nValue, CUdeviceptr pSrcDst, int nLength);


				///// <summary>
				///// 64-bit floating point complex number (64 bit real, 64 bit imaginary)signal
				///// add product of signal times constant to destination signal.
				///// </summary>
				///// <param name="pSrc">Source signal pointer.</param>
				///// <param name="nValue">Constant value to be multiplied by each vector element</param>
				///// <param name="pDst">Destination signal pointer.</param>
				///// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				///// <returns>
				///// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				///// </returns>
				//[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				//public static extern NppStatus nppsAddProductC_64fc(CUdeviceptr pSrc, Npp64fc nValue, CUdeviceptr pDst, int nLength);
			}

			/// <summary>
			/// Multiplies each sample of a signal by a constant value.
			/// </summary>
			public static class MulC
			{
				/// <summary>
				/// 8-bit unsigned char in place signal times constant,
				/// scale, then clamp to saturated value
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_8u_ISfs(byte nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 8-bit unsigned char signal times constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_8u_Sfs(CUdeviceptr pSrc, byte nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal times constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_16u_ISfs(ushort nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal times constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_16u_Sfs(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal times constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_16s_ISfs(short nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal times constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_16s_Sfs(CUdeviceptr pSrc, short nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary)signal times constant, 
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_16sc_ISfs(Npp16sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary)signal times constant,
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_16sc_Sfs(CUdeviceptr pSrc, Npp16sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer in place signal times constant and scale.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_32s_ISfs(int nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal times constant and scale.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_32s_Sfs(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit integer complex number (32 bit real, 32 bit imaginary) in place signal
				/// times constant and scale.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_32sc_ISfs(Npp32sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit integer complex number (32 bit real, 32 bit imaginary) signal times constant
				/// and scale.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_32sc_Sfs(CUdeviceptr pSrc, Npp32sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit floating point in place signal times constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_32f_I(float nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal times constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_32f(CUdeviceptr pSrc, float nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal times constant with output converted to 16-bit signed integer.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_Low_32f16s(CUdeviceptr pSrc, float nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal times constant with output converted to 16-bit signed integer
				/// with scaling and saturation of output result.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_32f16s_Sfs(CUdeviceptr pSrc, float nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) in
				/// place signal times constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_32fc_I(Npp32fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) signal
				/// times constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_32fc(CUdeviceptr pSrc, Npp32fc nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point, in place signal times constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Length of the vectors, number of items.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_64f_I(double nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal times constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_64f(CUdeviceptr pSrc, double nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal times constant with in place conversion to 64-bit signed integer
				/// and with scaling and saturation of output result.
				/// </summary>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_64f64s_ISfs(double nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) in
				/// place signal times constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_64fc_I(Npp64fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) signal
				/// times constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be multiplied by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMulC_64fc(CUdeviceptr pSrc, Npp64fc nValue, CUdeviceptr pDst, int nLength);



			}

			/// <summary>
			/// Subtracts a constant from each sample of a signal.
			/// </summary>
			public static class SubC
			{		/// <summary>
				/// 8-bit unsigned char in place signal subtract constant,
				/// scale, then clamp to saturated value
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_8u_ISfs(byte nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 8-bit unsigned char signal subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_8u_Sfs(CUdeviceptr pSrc, byte nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_16u_ISfs(ushort nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_16u_Sfs(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_16s_ISfs(short nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal subtract constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_16s_Sfs(CUdeviceptr pSrc, short nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary) signal subtract constant, 
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_16sc_ISfs(Npp16sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary) signal subtract constant,
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_16sc_Sfs(CUdeviceptr pSrc, Npp16sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer in place signal subtract constant and scale.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_32s_ISfs(int nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal subtract constant and scale.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_32s_Sfs(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit integer complex number (32 bit real, 32 bit imaginary) in place signal
				/// subtract constant and scale.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_32sc_ISfs(Npp32sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit integer complex number (32 bit real, 32 bit imaginary)signal subtract constant
				/// and scale.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_32sc_Sfs(CUdeviceptr pSrc, Npp32sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit floating point in place signal subtract constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_32f_I(float nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal subtract constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_32f(CUdeviceptr pSrc, float nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) in
				/// place signal subtract constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_32fc_I(Npp32fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) signal
				/// subtract constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_32fc(CUdeviceptr pSrc, Npp32fc nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point, in place signal subtract constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Length of the vectors, number of items.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_64f_I(double nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal subtract constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_64f(CUdeviceptr pSrc, double nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) in
				/// place signal subtract constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_64fc_I(Npp64fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) signal
				/// subtract constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be subtracted from each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubC_64fc(CUdeviceptr pSrc, Npp64fc nValue, CUdeviceptr pDst, int nLength);



			}

			/// <summary>
			/// Subtracts each sample of a signal from a constant.
			/// </summary>
			public static class SubCRev
			{		/// <summary>
				/// 8-bit unsigned char in place signal subtract from constant,
				/// scale, then clamp to saturated value
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_8u_ISfs(byte nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 8-bit unsigned char signal subtract from constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_8u_Sfs(CUdeviceptr pSrc, byte nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal subtract from constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_16u_ISfs(ushort nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal subtract from constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_16u_Sfs(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal subtract from constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_16s_ISfs(short nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal subtract from constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_16s_Sfs(CUdeviceptr pSrc, short nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary) signal subtract from constant, 
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_16sc_ISfs(Npp16sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary) signal subtract from constant,
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_16sc_Sfs(CUdeviceptr pSrc, Npp16sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer in place signal subtract from constant and scale.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_32s_ISfs(int nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integersignal subtract from constant and scale.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_32s_Sfs(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit integer complex number (32 bit real, 32 bit imaginary) in place signal
				/// subtract from constant and scale.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_32sc_ISfs(Npp32sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit integer complex number (32 bit real, 32 bit imaginary) signal subtract from constant
				/// and scale.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_32sc_Sfs(CUdeviceptr pSrc, Npp32sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit floating point in place signal subtract from constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_32f_I(float nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal subtract from constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_32f(CUdeviceptr pSrc, float nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) in
				/// place signal subtract from constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_32fc_I(Npp32fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) signal
				/// subtract from constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_32fc(CUdeviceptr pSrc, Npp32fc nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point, in place signal subtract from constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Length of the vectors, number of items.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_64f_I(double nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal subtract from constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_64f(CUdeviceptr pSrc, double nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) in
				/// place signal subtract from constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_64fc_I(Npp64fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) signal
				/// subtract from constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value each vector element is to be subtracted from</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSubCRev_64fc(CUdeviceptr pSrc, Npp64fc nValue, CUdeviceptr pDst, int nLength);



			}

			/// <summary>
			/// Divides each sample of a signal by a constant.
			/// </summary>
			public static class DivC
			{		/// <summary>
				/// 8-bit unsigned char in place signal divided by constant,
				/// scale, then clamp to saturated value
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_8u_ISfs(byte nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 8-bit unsigned char signal divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_8u_Sfs(CUdeviceptr pSrc, byte nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_16u_ISfs(ushort nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_16u_Sfs(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_16s_ISfs(short nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal divided by constant, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_16s_Sfs(CUdeviceptr pSrc, short nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary)signal divided by constant, 
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_16sc_ISfs(Npp16sc nValue, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit integer complex number (16 bit real, 16 bit imaginary) signal divided by constant,
				/// scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_16sc_Sfs(CUdeviceptr pSrc, Npp16sc nValue, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit floating point in place signal divided by constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_32f_I(float nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal divided by constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_32f(CUdeviceptr pSrc, float nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) in
				/// place signal divided by constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_32fc_I(Npp32fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point complex number (32 bit real, 32 bit imaginary) signal
				/// divided by constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_32fc(CUdeviceptr pSrc, Npp32fc nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point in place signal divided by constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="nLength">Length of the vectors, number of items.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_64f_I(double nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal divided by constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_64f(CUdeviceptr pSrc, double nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) in
				/// place signal divided by constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_64fc_I(Npp64fc nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point complex number (64 bit real, 64 bit imaginary) signal
				/// divided by constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided into each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivC_64fc(CUdeviceptr pSrc, Npp64fc nValue, CUdeviceptr pDst, int nLength);



			}

			/// <summary>
			/// Divides a constant by each sample of a signal.
			/// </summary>
			public static class DivCRev
			{		/// <summary>
				/// 16-bit unsigned short in place constant divided by signal, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivCRev_16u_I(ushort nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal divided by constant, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivCRev_16u(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point in place constant divided by signal.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be divided by each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivCRev_32f_I(float nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point constant divided by signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be divided by each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDivCRev_32f(CUdeviceptr pSrc, float nValue, CUdeviceptr pDst, int nLength);



			}

			/// <summary>
			/// Sample by sample addition of two signals.
			/// </summary>
			public static class AddSignal
			{		/// <summary>
				/// 16-bit signed short signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16s(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned int signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_32u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_64fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char signal add signal with 16-bit unsigned result,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_8u16u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit signed short signal add signal with 32-bit floating point result,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16s32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char add signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be added to signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_8u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short add signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be added to signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short add signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be added to signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer add signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be added to signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 64-bit signed integer add signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be added to signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_64s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed complex short add signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be added to signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed complex integer add signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be added to signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_32sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16s_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point in place signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_32f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point in place signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_64f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point in place signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_32fc_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point in place signal add signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_64fc_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16/32-bit signed short in place signal add signal with 32-bit signed integer results,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16s32s_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal add signal, with scaling,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_8u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal add signal, with scaling,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal add signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16s_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer in place signal add signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_32s_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit complex signed short in place signal add signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_16sc_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit complex signed integer in place signal add signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be added to signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAdd_32sc_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// Adds sample by sample product of two signals to the destination signal.
			/// </summary>
			public static class AddProductSignal
			{		/// <summary>
				/// 32-bit floating point signal add product of source signal times destination signal to destination signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer. product of source1 and source2 signal elements to be added to destination elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddProduct_32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal add product of source signal times destination signal to destination signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer. product of source1 and source2 signal elements to be added to destination elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddProduct_64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point signal add product of source signal times destination signal to destination signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer. product of source1 and source2 signal elements to be added to destination elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddProduct_32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point signal add product of source signal times destination signal to destination signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer. product of source1 and source2 signal elements to be added to destination elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddProduct_64fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit signed short signal add product of source signal1 times source signal2 to destination signal,
				/// with scaling, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer. product of source1 and source2 signal elements to be added to destination elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddProduct_16s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed short signal add product of source signal1 times source signal2 to destination signal,
				/// with scaling, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer. product of source1 and source2 signal elements to be added to destination elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddProduct_32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal add product of source signal1 times source signal2 to 32-bit signed integer destination signal,
				/// with scaling, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer. product of source1 and source2 signal elements to be added to destination elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAddProduct_16s32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// Sample by sample multiplication the samples of two signals.
			/// </summary>
			public static class MulSignal
			{		/// <summary>
				/// 16-bit signed short signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16s(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_64fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char signal times signal with 16-bit unsigned result,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_8u16u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit signed short signal times signal with 32-bit floating point result,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16s32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal times 32-bit complex floating point signal with complex 32-bit floating point result,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32f32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char signal times signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_8u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal time signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal times signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal times signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed complex short signal times signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed complex integer signal times signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal times 16-bit signed short signal, scale, then clamp to 16-bit signed saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16u16s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal times signal, scale, then clamp to 32-bit signed saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16s32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal times 32-bit complex signed integer signal, scale, then clamp to 32-bit complex integer saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32s32sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal times signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal2 elements to be multiplied by signal1 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_Low_32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16s_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point in place signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point in place signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_64f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point in place signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32fc_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point in place signal times signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_64fc_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point in place signal times 32-bit floating point signal,
				/// then clamp to 32-bit complex floating point saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32f32fc_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal times signal, with scaling,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_8u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal times signal, with scaling,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal times signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16s_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer in place signal times signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32s_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit complex signed short in place signal times signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_16sc_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit complex signed integer in place signal times signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32sc_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit complex signed integer in place signal times 32-bit signed integer signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be multiplied by signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMul_32s32sc_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// Sample by sample subtraction of the samples of two signals.
			/// </summary>
			public static class SubSignal
			{		/// <summary>
				/// 16-bit signed short signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_16s(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_64fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit signed short signal subtract 16-bit signed short signal,
				/// then clamp and convert to 32-bit floating point saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_16s32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char signal subtract signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 elements to be subtracted from signal2 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_8u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal subtract signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 elements to be subtracted from signal2 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_16u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal subtract signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 elements to be subtracted from signal2 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_16s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal subtract signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 elements to be subtracted from signal2 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed complex short signal subtract signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 elements to be subtracted from signal2 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_16sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed complex integer signal subtract signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 elements to be subtracted from signal2 elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_32sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_16s_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point in place signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_32f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point in place signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_64f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point in place signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_32fc_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point in place signal subtract signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_64fc_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal subtract signal, with scaling,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_8u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal subtract signal, with scaling,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_16u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal subtract signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_16s_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer in place signal subtract signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_32s_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit complex signed short in place signal subtract signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_16sc_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit complex signed integer in place signal subtract signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 elements to be subtracted from signal2 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSub_32sc_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// Sample by sample division of the samples of two signals.
			/// </summary>
			public static class DivSignal
			{		/// <summary>
				/// 8-bit unsigned char signal divide signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_8u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal divide signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_16u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal divide signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_16s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal divide signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed complex short signal divide signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_16sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal divided by 16-bit signed short signal, scale, then clamp to 16-bit signed short saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_32s16s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit floating point signal divide signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal divide signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point signal divide signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point signal divide signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_64fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal divide signal, with scaling,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_8u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal divide signal, with scaling,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_16u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal divide signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_16s_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit complex signed short in place signal divide signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_16sc_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer in place signal divide signal, with scaling, 
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_32s_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit floating point in place signal divide signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_32f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point in place signal divide signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_64f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point in place signal divide signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_32fc_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point in place signal divide signal,
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_64fc_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Sample by sample division of the samples of two signals with rounding.
			/// </summary>
			public static class DivRoundSignal
			{		/// <summary>
				/// 8-bit unsigned char signal divide signal, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nRndMode">various rounding modes.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_Round_8u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, NppRoundMode nRndMode, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal divide signal, scale, round, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nRndMode">various rounding modes.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_Round_16u_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, NppRoundMode nRndMode, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal divide signal, scale, round, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer, signal1 divisor elements to be divided into signal2 dividend elements.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nRndMode">various rounding modes.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_Round_16s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength, NppRoundMode nRndMode, int nScaleFactor);


				/// <summary>
				/// 8-bit unsigned char in place signal divide signal, with scaling, rounding
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nRndMode">various rounding modes.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_Round_8u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, NppRoundMode nRndMode, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short in place signal divide signal, with scaling, rounding
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nRndMode">various rounding modes.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_Round_16u_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, NppRoundMode nRndMode, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short in place signal divide signal, with scaling, rounding
				/// then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal1 divisor elements to be divided into signal2 dividend elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nRndMode">various rounding modes.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDiv_Round_16s_ISfs(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength, NppRoundMode nRndMode, int nScaleFactor);



			}

			/// <summary>
			/// Absolute value of each sample of a signal.
			/// </summary>
			public static class AbsoluteValueSignal
			{		/// <summary>
				/// 16-bit signed short signal absolute value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAbs_16s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit signed integer signal absolute value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAbs_32s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal absolute value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAbs_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal absolute value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAbs_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit signed short signal absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAbs_16s_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit signed integer signal absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAbs_32s_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAbs_32f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal absolute value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAbs_64f_I(CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Squares each sample of a signal.
			/// </summary>
			public static class SquareSignal
			{		/// <summary>
				/// 32-bit floating point signal squared.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal squared.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point signal squared.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_32fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point signal squared.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_64fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal squared.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_32f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal squared.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_64f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point signal squared.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_32fc_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point signal squared.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_64fc_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char signal squared, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_8u_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal squared, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_16u_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal squared, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit complex signed short signal squared, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_16sc_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 8-bit unsigned char signal squared, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_8u_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal squared, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_16u_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal squared, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_16s_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit complex signed short signal squared, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqr_16sc_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// Square root of each sample of a signal.
			/// </summary>
			public static class SquareRootSignal
			{		/// <summary>
				/// 32-bit floating point signal square root.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal square root.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point signal square root.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_32fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point signal square root.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_64fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal square root.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_32f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal square root.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_64f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit complex floating point signal square root.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_32fc_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit complex floating point signal square root.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_64fc_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_8u_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_16u_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit complex signed short signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_16sc_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 64-bit signed integer signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_64s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal square root, scale, then clamp to 16-bit signed integer saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_32s16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 64-bit signed integer signal square root, scale, then clamp to 16-bit signed integer saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_64s16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 8-bit unsigned char signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_8u_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit unsigned short signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_16u_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_16s_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit complex signed short signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_16sc_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 64-bit signed integer signal square root, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSqrt_64s_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// Cube root of each sample of a signal.
			/// </summary>
			public static class CubeRootSignal
			{		/// <summary>
				/// 32-bit floating point signal cube root.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCubrt_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit signed integer signal cube root, scale, then clamp to 16-bit signed integer saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCubrt_32s16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// e raised to the power of each sample of a signal.
			/// </summary>
			public static class ExponentSignal
			{		/// <summary>
				/// 32-bit floating point signal exponent.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal exponent.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal exponent with 64-bit floating point result.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_32f64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal exponent.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_32f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal exponent.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_64f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit signed short signal exponent, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal exponent, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_32s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 64-bit signed integer signal exponent, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_64s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal exponent, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_16s_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal exponent, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_32s_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 64-bit signed integer signal exponent, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsExp_64s_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// Natural logarithm of each sample of a signal.
			/// </summary>
			public static class NaturalLogarithmSignal
			{		/// <summary>
				/// 32-bit floating point signal natural logarithm.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal natural logarithm.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal natural logarithm with 32-bit floating point result.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_64f32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal natural logarithm.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_32f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal natural logarithm.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_64f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit signed short signal natural logarithm, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal natural logarithm, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_32s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal natural logarithm, scale, then clamp to 16-bit signed short saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_32s16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 16-bit signed short signal natural logarithm, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_16s_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal natural logarithm, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLn_32s_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// Ten times the decimal logarithm of each sample of a signal.
			/// </summary>
			public static class TenTimesBaseTenLogarithmSignal
			{		/// <summary>
				/// 32-bit signed integer signal 10 times base 10 logarithm, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus npps10Log10_32s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);


				/// <summary>
				/// 32-bit signed integer signal 10 times base 10 logarithm, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus npps10Log10_32s_ISfs(CUdeviceptr pSrcDst, int nLength, int nScaleFactor);



			}

			/// <summary>
			/// Inverse tangent of each sample of a signal.
			/// </summary>
			public static class InverseTangentSignal
			{		/// <summary>
				/// 32-bit floating point signal inverse tangent.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsArctan_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal inverse tangent.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsArctan_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit floating point signal inverse tangent.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsArctan_32f_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point signal inverse tangent.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsArctan_64f_I(CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Normalize each sample of a real or complex signal using offset and division operations.
			/// </summary>
			public static class NormalizeSignal
			{		/// <summary>
				/// 32-bit floating point signal normalize.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="vSub">value subtracted from each signal element before division</param>
				/// <param name="vDiv">divisor of post-subtracted signal element dividend</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormalize_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float vSub, float vDiv);


				/// <summary>
				/// 32-bit complex floating point signal normalize.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="vSub">value subtracted from each signal element before division</param>
				/// <param name="vDiv">divisor of post-subtracted signal element dividend</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormalize_32fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, Npp32fc vSub, float vDiv);


				/// <summary>
				/// 64-bit floating point signal normalize.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="vSub">value subtracted from each signal element before division</param>
				/// <param name="vDiv">divisor of post-subtracted signal element dividend</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormalize_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double vSub, double vDiv);


				/// <summary>
				/// 64-bit complex floating point signal normalize.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="vSub">value subtracted from each signal element before division</param>
				/// <param name="vDiv">divisor of post-subtracted signal element dividend</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormalize_64fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, Npp64fc vSub, double vDiv);


				/// <summary>
				/// 16-bit signed short signal normalize, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="vSub">value subtracted from each signal element before division</param>
				/// <param name="vDiv">divisor of post-subtracted signal element dividend</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormalize_16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short vSub, int vDiv, int nScaleFactor);


				/// <summary>
				/// 16-bit complex signed short signal normalize, scale, then clamp to saturated value.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="vSub">value subtracted from each signal element before division</param>
				/// <param name="vDiv">divisor of post-subtracted signal element dividend</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormalize_16sc_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, Npp16sc vSub, int vDiv, int nScaleFactor);



			}

			/// <summary>
			/// Determine Cauchy robust error function and its first and second derivatives for each sample of a signal.
			/// </summary>
			public static class Cauchy
			{		/// <summary>
				/// 32-bit floating point signal Cauchy error calculation.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nParam">constant used in Cauchy formula</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCauchy_32f_I(CUdeviceptr pSrcDst, int nLength, float nParam);


				/// <summary>
				/// 32-bit floating point signal Cauchy first derivative.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nParam">constant used in Cauchy formula</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCauchyD_32f_I(CUdeviceptr pSrcDst, int nLength, float nParam);


				/// <summary>
				/// 32-bit floating point signal Cauchy first and second derivatives.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="pD2FVal">Source signal pointer. This signal contains the second derivative</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="nParam">constant used in Cauchy formula</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCauchyDD2_32f_I(CUdeviceptr pSrcDst, CUdeviceptr pD2FVal, int nLength, float nParam);



			}

			/// <summary>
			/// Bitwise AND of a constant and each sample of a signal.
			/// </summary>
			public static class AndC
			{		/// <summary>
				/// 8-bit unsigned char signal and with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be anded with each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAndC_8u(CUdeviceptr pSrc, byte nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal and with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be anded with each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAndC_16u(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer signal and with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be anded with each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAndC_32u(CUdeviceptr pSrc, uint nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal and with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be anded with each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAndC_8u_I(byte nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short in place signal and with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be anded with each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAndC_16u_I(ushort nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit unsigned signed integer in place signal and with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be anded with each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAndC_32u_I(uint nValue, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Sample by sample bitwise AND of samples from two signals.
			/// </summary>
			public static class And
			{		/// <summary>
				/// 8-bit unsigned char signal and with signal.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be anded with signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAnd_8u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal and with signal.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be anded with signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAnd_16u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer signal and with signal.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be anded with signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAnd_32u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal and with signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be anded with signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAnd_8u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short in place signal and with signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be anded with signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAnd_16u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer in place signal and with signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be anded with signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsAnd_32u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Bitwise OR of a constant and each sample of a signal.
			/// </summary>
			public static class OrC
			{		/// <summary>
				/// 8-bit unsigned char signal or with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be ored with each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOrC_8u(CUdeviceptr pSrc, byte nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal or with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be ored with each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOrC_16u(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer signal or with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be ored with each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOrC_32u(CUdeviceptr pSrc, uint nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal or with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be ored with each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOrC_8u_I(byte nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short in place signal or with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be ored with each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOrC_16u_I(ushort nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit unsigned signed integer in place signal or with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be ored with each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOrC_32u_I(uint nValue, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Sample by sample bitwise OR of samples from two signals.
			/// </summary>
			public static class Or
			{		/// <summary>
				/// 8-bit unsigned char signal or with signal.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be ored with signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOr_8u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal or with signal.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be ored with signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOr_16u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer signal or with signal.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be ored with signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOr_32u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal or with signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be ored with signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOr_8u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short in place signal or with signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be ored with signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOr_16u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer in place signal or with signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be ored with signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsOr_32u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Bitwise XOR of a constant and each sample of a signal.
			/// </summary>
			public static class XorC
			{		/// <summary>
				/// 8-bit unsigned char signal exclusive or with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be exclusive ored with each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXorC_8u(CUdeviceptr pSrc, byte nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal exclusive or with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be exclusive ored with each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXorC_16u(CUdeviceptr pSrc, ushort nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer signal exclusive or with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be exclusive ored with each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXorC_32u(CUdeviceptr pSrc, uint nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal exclusive or with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be exclusive ored with each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXorC_8u_I(byte nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short in place signal exclusive or with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be exclusive ored with each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXorC_16u_I(ushort nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit unsigned signed integer in place signal exclusive or with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be exclusive ored with each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXorC_32u_I(uint nValue, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Sample by sample bitwise XOR of samples from two signals.
			/// </summary>
			public static class Xor
			{		/// <summary>
				/// 8-bit unsigned char signal exclusive or with signal.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be exclusive ored with signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXor_8u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal exclusive or with signal.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be exclusive ored with signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXor_16u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer signal exclusive or with signal.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer. signal2 elements to be exclusive ored with signal1 elements</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXor_32u(CUdeviceptr pSrc1, CUdeviceptr pSrc2, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal exclusive or with signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be exclusive ored with signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXor_8u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short in place signal exclusive or with signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be exclusive ored with signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXor_16u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer in place signal exclusive or with signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer. signal2 elements to be exclusive ored with signal1 elements</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsXor_32u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Bitwise NOT of each sample of a signal.
			/// </summary>
			public static class Not
			{		/// <summary>
				/// 8-bit unsigned char not signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNot_8u(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short not signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNot_16u(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer not signal.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNot_32u(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place not signal.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNot_8u_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short in place not signal.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNot_16u_I(CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit unsigned signed integer in place not signal.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNot_32u_I(CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Left shifts the bits of each sample of a signal by a constant amount.
			/// </summary>
			public static class LShiftC
			{		/// <summary>
				/// 8-bit unsigned char signal left shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_8u(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal left shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_16u(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit signed short signal left shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_16s(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer signal left shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_32u(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit signed integer signal left shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_32s(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal left shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_8u_I(int nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short in place signal left shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_16u_I(int nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit signed short in place signal left shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_16s_I(int nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit unsigned signed integer in place signal left shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_32u_I(int nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit signed signed integer in place signal left shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to left shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsLShiftC_32s_I(int nValue, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Right shifts the bits of each sample of a signal by a constant amount.
			/// </summary>
			public static class RShiftC
			{
				/// <summary>
				/// 8-bit unsigned char signal right shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_8u(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short signal right shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_16u(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 16-bit signed short signal right shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_16s(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit unsigned integer signal right shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_32u(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 32-bit signed integer signal right shift with constant.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_32s(CUdeviceptr pSrc, int nValue, CUdeviceptr pDst, int nLength);


				/// <summary>
				/// 8-bit unsigned char in place signal right shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_8u_I(int nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short in place signal right shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_16u_I(int nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit signed short in place signal right shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_16s_I(int nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit unsigned signed integer in place signal right shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_32u_I(int nValue, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit signed signed integer in place signal right shift with constant.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nValue">Constant value to be used to right shift each vector element</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsRShiftC_32s_I(int nValue, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Functions that provide global signal statistics like: average, standard deviation, minimum, etc.
			/// </summary>
			public static class Sum
			{
				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsSum_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumGetBufferSize_32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsSum_32fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumGetBufferSize_32fc(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsSum_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumGetBufferSize_64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsSum_64fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumGetBufferSize_64fc(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsSum_16s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumGetBufferSize_16s_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsSum_16sc_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumGetBufferSize_16sc_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsSum_16sc32sc_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumGetBufferSize_16sc32sc_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsSum_32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumGetBufferSize_32s_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsSum_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumGetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 32-bit float vector sum method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pSum">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSum_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pSum, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit float complex vector sum method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pSum">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSum_32fc(CUdeviceptr pSrc, int nLength, CUdeviceptr pSum, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit double vector sum method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pSum">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSum_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pSum, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit double complex vector sum method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pSum">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSum_64fc(CUdeviceptr pSrc, int nLength, CUdeviceptr pSum, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit short vector sum with integer scaling method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pSum">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSum_16s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pSum, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector sum with integer scaling method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pSum">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSum_32s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pSum, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit short complex vector sum with integer scaling method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pSum">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSum_16sc_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pSum, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit short complex vector sum (32bit int complex) with integer scaling
				/// method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pSum">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSum_16sc32sc_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pSum, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit integer vector sum (32bit) with integer scaling method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pSum">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSum_16s32s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pSum, int nScaleFactor, CUdeviceptr pDeviceBuffer);

			}

			/// <summary>
			/// Functions that provide global signal statistics like: average, standard deviation, minimum, etc.
			/// </summary>
			public static class Max
			{
				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMax_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxGetBufferSize_16s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMax_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxGetBufferSize_32s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMax_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxGetBufferSize_32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMax_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxGetBufferSize_64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 16-bit integer vector max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMax">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMax_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMax">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMax_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit float vector max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMax">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMax_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit float vector max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMax">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMax_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMaxIndx_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxIndxGetBufferSize_16s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMaxIndx_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxIndxGetBufferSize_32s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMaxIndx_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxIndxGetBufferSize_32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMaxIndx_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxIndxGetBufferSize_64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 16-bit integer vector max index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMax">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first maximum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxIndx_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMax, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector max index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMax">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first maximum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxIndx_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMax, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit float vector max index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMax">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first maximum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxIndx_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMax, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit float vector max index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMax">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first maximum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxIndx_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMax, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMaxAbs_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxAbsGetBufferSize_16s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMaxAbs_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxAbsGetBufferSize_32s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 16-bit integer vector max absolute method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMaxAbs">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxAbs_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMaxAbs, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector max absolute method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMaxAbs">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxAbs_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMaxAbs, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMaxAbsIndx_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxAbsIndxGetBufferSize_16s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMaxAbsIndx_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxAbsIndxGetBufferSize_32s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 16-bit integer vector max absolute index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMaxAbs">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first maximum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxAbsIndx_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMaxAbs, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector max absolute index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMaxAbs">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first maximum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxAbsIndx_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMaxAbs, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);

				
			}

			/// <summary>
			/// Functions that provide global signal statistics like: average, standard deviation, minimum, etc.
			/// </summary>
			public static class Min
			{
				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMin_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinGetBufferSize_16s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMin_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinGetBufferSize_32s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMin_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinGetBufferSize_32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMin_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinGetBufferSize_64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 16-bit integer vector min method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMin_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector min method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMin_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector min method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMin_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit integer vector min method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMin_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMinIndx_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinIndxGetBufferSize_16s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMinIndx_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinIndxGetBufferSize_32s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMinIndx_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinIndxGetBufferSize_32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMinIndx_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinIndxGetBufferSize_64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 16-bit integer vector min index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first minimum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinIndx_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector min index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first minimum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinIndx_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit float vector min index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first minimum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinIndx_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit float vector min index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first minimum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinIndx_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMinAbs_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinAbsGetBufferSize_16s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMinAbs_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinAbsGetBufferSize_32s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 16-bit integer vector min absolute method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMinAbs">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinAbs_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMinAbs, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector min absolute method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMinAbs">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinAbs_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMinAbs, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMinAbsIndx_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinAbsIndxGetBufferSize_16s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMinAbsIndx_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinAbsIndxGetBufferSize_32s(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 16-bit integer vector min absolute index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMinAbs">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first minimum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinAbsIndx_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMinAbs, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector min absolute index method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMinAbs">Pointer to the output result.</param>
				/// <param name="pIndx">Pointer to the index value of the first minimum element.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinAbsIndx_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMinAbs, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);



			}

			/// <summary>
			/// Functions that provide functionality of generating output signal
			/// based on the input signal like signal integral, etc.
			/// </summary>
			public static class FilteringFunctions
			{ 
				/// <summary>
				/// 
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="hpBufferSize">Required buffer size.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsIntegralGetBufferSize_32s(int nLength,  ref int hpBufferSize);

				/// <summary>
				/// 
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsIntegral_32s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, CUdeviceptr pDeviceBuffer);
			}

			/// <summary>
			/// Sums up the natural logarithm of each sample of a signal.
			/// </summary>
			public static class SumLn
			{		/// <summary>
				/// Device scratch buffer size (in bytes) for 32f SumLn.
				/// This primitive provides the correct buffer size for nppsSumLn_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumLnGetBufferSize_32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 32-bit floating point signal sum natural logarithm.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDst">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumLn_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for 64f SumLn.
				/// This primitive provides the correct buffer size for nppsSumLn_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumLnGetBufferSize_64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 64-bit floating point signal sum natural logarithm.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDst">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumLn_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for 32f64f SumLn.
				/// This primitive provides the correct buffer size for nppsSumLn_32f64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumLnGetBufferSize_32f64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 32-bit flaoting point input, 64-bit floating point output signal sum natural logarithm.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDst">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumLn_32f64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for 16s32f SumLn.
				/// This primitive provides the correct buffer size for nppsSumLn_16s32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumLnGetBufferSize_16s32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 16-bit signed short integer input, 32-bit floating point output signal sum natural logarithm.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDst">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsSumLn_16s32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pDst, CUdeviceptr pDeviceBuffer);



			}

			/// <summary>
			/// Routines for converting the sample-data type of signals.
			/// </summary>
			public static class Convert
			{
				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_8s16s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_8s32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_8u32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_16s8s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, uint nLength, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_16s32s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_16s32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_16u32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32s16s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32s32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32s64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32f64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_64s64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_64f32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_16s32f_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_16s64f_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32s16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32s32f_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32s64f_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32f8s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32f8u_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32f16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32f16u_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_32f32s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_64s32s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_64f16s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_64f32s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, NppRoundMode eRoundMode, int nScaleFactor);

				/// <summary/>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsConvert_64f64s_Sfs(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, NppRoundMode eRoundMode, int nScaleFactor);


			}

			/// <summary>
			/// Performs the threshold operation on the samples of a signal by limiting the sample values by a specified constant value.
			/// </summary>
			public static class Threshold
			{
				/// <summary>
				/// 16-bit signed short signal threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_16s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 16-bit in place signed short signal threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_16s_I(CUdeviceptr pSrcDst, int nLength, short nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 16-bit signed short complex number signal threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_16sc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 16-bit in place signed short complex number signal threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_16sc_I(CUdeviceptr pSrcDst, int nLength, short nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 32-bit floating point signal threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 32-bit in place floating point signal threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_32f_I(CUdeviceptr pSrcDst, int nLength, float nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 32-bit floating point complex number signal threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_32fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 32-bit in place floating point complex number signal threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_32fc_I(CUdeviceptr pSrcDst, int nLength, float nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 64-bit floating point signal threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel, NppCmpOp nRelOp);
				
				/// <summary>
				/// 64-bit in place floating point signal threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_64f_I(CUdeviceptr pSrcDst, int nLength, double nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 64-bit floating point complex number signal threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_64fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 64-bit in place floating point complex number signal threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nRelOp">NppCmpOp type of thresholding operation (NPP_CMP_LESS or NPP_CMP_GREATER only)</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_64fc_I(CUdeviceptr pSrcDst, int nLength, double nLevel, NppCmpOp nRelOp);

				/// <summary>
				/// 16-bit signed short signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_16s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel);

				/// <summary>
				/// 16-bit in place signed short signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_16s_I(CUdeviceptr pSrcDst, int nLength, short nLevel);

				/// <summary>
				/// 16-bit signed short complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_16sc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel);

				/// <summary>
				/// 16-bit in place signed short complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_16sc_I(CUdeviceptr pSrcDst, int nLength, short nLevel);

				/// <summary>
				/// 32-bit floating point signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel);

				/// <summary>
				/// 32-bit in place floating point signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_32f_I(CUdeviceptr pSrcDst, int nLength, float nLevel);

				/// <summary>
				/// 32-bit floating point complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_32fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel);

				/// <summary>
				/// 32-bit in place floating point complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_32fc_I(CUdeviceptr pSrcDst, int nLength, float nLevel);

				/// <summary>
				/// 64-bit floating point signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel);

				/// <summary>
				/// 64-bit in place floating point signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_64f_I(CUdeviceptr pSrcDst, int nLength, double nLevel);

				/// <summary>
				/// 64-bit floating point complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_64fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel);

				/// <summary>
				/// 64-bit in place floating point complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LT_64fc_I(CUdeviceptr pSrcDst, int nLength, double nLevel);



				/// <summary>
				/// 16-bit signed short signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_16s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel);

				/// <summary>
				/// 16-bit in place signed short signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_16s_I(CUdeviceptr pSrcDst, int nLength, short nLevel);

				/// <summary>
				/// 16-bit signed short complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_16sc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel);

				/// <summary>
				/// 16-bit in place signed short complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_16sc_I(CUdeviceptr pSrcDst, int nLength, short nLevel);

				/// <summary>
				/// 32-bit floating point signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel);

				/// <summary>
				/// 32-bit in place floating point signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_32f_I(CUdeviceptr pSrcDst, int nLength, float nLevel);

				/// <summary>
				/// 32-bit floating point complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_32fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel);

				/// <summary>
				/// 32-bit in place floating point complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_32fc_I(CUdeviceptr pSrcDst, int nLength, float nLevel);

				/// <summary>
				/// 64-bit floating point signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel);

				/// <summary>
				/// 64-bit in place floating point signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_64f_I(CUdeviceptr pSrcDst, int nLength, double nLevel);

				/// <summary>
				/// 64-bit floating point complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_64fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel);

				/// <summary>
				/// 64-bit in place floating point complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GT_64fc_I(CUdeviceptr pSrcDst, int nLength, double nLevel);



				/// <summary>
				/// 16-bit signed short signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_16s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel, short nValue);

				/// <summary>
				/// 16-bit in place signed short signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_16s_I(CUdeviceptr pSrcDst, int nLength, short nLevel, short nValue);

				/// <summary>
				/// 16-bit signed short complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_16sc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel, Npp16sc nValue);

				/// <summary>
				/// 16-bit in place signed short complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_16sc_I(CUdeviceptr pSrcDst, int nLength, short nLevel, Npp16sc nValue);

				/// <summary>
				/// 32-bit floating point signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel, float nValue);

				/// <summary>
				/// 32-bit in place floating point signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_32f_I(CUdeviceptr pSrcDst, int nLength, float nLevel, float nValue);

				/// <summary>
				/// 32-bit floating point complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_32fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel, Npp32fc nValue);

				/// <summary>
				/// 32-bit in place floating point complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_32fc_I(CUdeviceptr pSrcDst, int nLength, float nLevel, Npp32fc nValue);

				/// <summary>
				/// 64-bit floating point signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel, double nValue);

				/// <summary>
				/// 64-bit in place floating point signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_64f_I(CUdeviceptr pSrcDst, int nLength, double nLevel, double nValue);

				/// <summary>
				/// 64-bit floating point complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_64fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel, Npp64fc nValue);

				/// <summary>
				/// 64-bit in place floating point complex number signal NPP_CMP_LESS threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_LTVal_64fc_I(CUdeviceptr pSrcDst, int nLength, double nLevel, Npp64fc nValue);

















				/// <summary>
				/// 16-bit signed short signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_16s(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel, short nValue);

				/// <summary>
				/// 16-bit in place signed short signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_16s_I(CUdeviceptr pSrcDst, int nLength, short nLevel, short nValue);

				/// <summary>
				/// 16-bit signed short complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_16sc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, short nLevel, Npp16sc nValue);

				/// <summary>
				/// 16-bit in place signed short complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_16sc_I(CUdeviceptr pSrcDst, int nLength, short nLevel, Npp16sc nValue);

				/// <summary>
				/// 32-bit floating point signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_32f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel, float nValue);

				/// <summary>
				/// 32-bit in place floating point signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_32f_I(CUdeviceptr pSrcDst, int nLength, float nLevel, float nValue);

				/// <summary>
				/// 32-bit floating point complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_32fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, float nLevel, Npp32fc nValue);

				/// <summary>
				/// 32-bit in place floating point complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_32fc_I(CUdeviceptr pSrcDst, int nLength, float nLevel, Npp32fc nValue);

				/// <summary>
				/// 64-bit floating point signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_64f(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel, double nValue);

				/// <summary>
				/// 64-bit in place floating point signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_64f_I(CUdeviceptr pSrcDst, int nLength, double nLevel, double nValue);

				/// <summary>
				/// 64-bit floating point complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pDst">Destination signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_64fc(CUdeviceptr pSrc, CUdeviceptr pDst, int nLength, double nLevel, Npp64fc nValue);

				/// <summary>
				/// 64-bit in place floating point complex number signal NPP_CMP_GREATER threshold with constant level.
				/// </summary>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="nLevel">Constant threshold value (real part only and must be greater than 0) to be used to limit each signal sample</param>
				/// <param name="nValue">Constant value to replace source value when threshold test is true.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsThreshold_GTVal_64fc_I(CUdeviceptr pSrcDst, int nLength, double nLevel, Npp64fc nValue);


			}

			/// <summary>
			/// Performs the min or max operation on the samples of a signal.
			/// </summary>
			public static class MinMaxEvery
			{
				/// <summary>
				/// 8-bit in place min value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinEvery_8u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short integer in place min value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinEvery_16u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit signed short integer in place min value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinEvery_16s_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit signed integer in place min value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinEvery_32s_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point in place min value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinEvery_32f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 64-bit floating point in place min value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinEvery_64f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);

				/// <summary>
				/// 8-bit in place max value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxEvery_8u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit unsigned short integer in place max value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxEvery_16u_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 16-bit signed short integer in place max value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxEvery_16s_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit signed integer in place max value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxEvery_32s_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);


				/// <summary>
				/// 32-bit floating point in place max value for each pair of elements.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="pSrcDst">In-Place Signal Pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMaxEvery_32f_I(CUdeviceptr pSrc, CUdeviceptr pSrcDst, int nLength);



			}

			/// <summary>
			/// Minimum / Maximum values / indices
			/// </summary>
			public static class MinMaxIndex
			{
				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMax_8u.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxGetBufferSize_8u(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMax_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxGetBufferSize_16s(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMax_16u.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxGetBufferSize_16u(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMax_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxGetBufferSize_32s(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMax_32u.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxGetBufferSize_32u(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMax_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxGetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMax_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxGetBufferSize_64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 8-bit char vector min and max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMax_8u(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit signed short vector min and max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMax_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit unsigned short vector min and max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMax_16u(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit unsigned int vector min and max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMax_32u(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit signed int vector min and max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMax_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit float vector min and max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMax_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit double vector min and max method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMax_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMax, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMaxIndx_8u.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndxGetBufferSize_8u(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMaxIndx_16s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndxGetBufferSize_16s(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMaxIndx_16u.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndxGetBufferSize_16u(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMaxIndx_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndxGetBufferSize_32s(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMaxIndx_32u.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndxGetBufferSize_32u(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMaxIndx_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndxGetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsMinMaxIndx_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndxGetBufferSize_64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 8-bit char vector min and max with indices method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMinIndx">Pointer to the index of the first min value.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pMaxIndx">Pointer to the index of the first max value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndx_8u(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMinIndx, CUdeviceptr pMax, CUdeviceptr pMaxIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit signed short vector min and max with indices method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMinIndx">Pointer to the index of the first min value.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pMaxIndx">Pointer to the index of the first max value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndx_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMinIndx, CUdeviceptr pMax, CUdeviceptr pMaxIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit unsigned short vector min and max with indices method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMinIndx">Pointer to the index of the first min value.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pMaxIndx">Pointer to the index of the first max value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndx_16u(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMinIndx, CUdeviceptr pMax, CUdeviceptr pMaxIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit signed short vector min and max with indices method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMinIndx">Pointer to the index of the first min value.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pMaxIndx">Pointer to the index of the first max value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndx_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMinIndx, CUdeviceptr pMax, CUdeviceptr pMaxIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit unsigned short vector min and max with indices method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMinIndx">Pointer to the index of the first min value.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pMaxIndx">Pointer to the index of the first max value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndx_32u(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMinIndx, CUdeviceptr pMax, CUdeviceptr pMaxIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit float vector min and max with indices method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMinIndx">Pointer to the index of the first min value.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pMaxIndx">Pointer to the index of the first max value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndx_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMinIndx, CUdeviceptr pMax, CUdeviceptr pMaxIndx, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit float vector min and max with indices method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMin">Pointer to the min output result.</param>
				/// <param name="pMinIndx">Pointer to the index of the first min value.</param>
				/// <param name="pMax">Pointer to the max output result.</param>
				/// <param name="pMaxIndx">Pointer to the index of the first max value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMinMaxIndx_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMin, CUdeviceptr pMinIndx, CUdeviceptr pMax, CUdeviceptr pMaxIndx, CUdeviceptr pDeviceBuffer);



			}

			///// <summary>
			///// min / max absolute values / indices
			///// </summary>
			//public static class MinMaxAbs
			//{
			//    /// <summary>
			//    /// Device scratch buffer size (in bytes) for nppsMaxAbs_16s.
			//    /// </summary>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="hpBufferSize">Required buffer size.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMaxAbsGetBufferSize_16s(int nLength, ref int hpBufferSize);

			//    /// <summary>
			//    /// Device scratch buffer size (in bytes) for nppsMaxAbs_32s.
			//    /// </summary>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="hpBufferSize">Required buffer size.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMaxAbsGetBufferSize_32s(int nLength, ref int hpBufferSize);

			//    /// <summary>
			//    /// 16-bit integer vector max absolute method
			//    /// </summary>
			//    /// <param name="pSrc">Source signal pointer.</param>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="pMaxAbs">Pointer to the output result.</param>
			//    /// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMaxAbs_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMaxAbs, CUdeviceptr pDeviceBuffer);

			//    /// <summary>
			//    /// 32-bit integer vector max absolute method
			//    /// </summary>
			//    /// <param name="pSrc">Source signal pointer.</param>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="pMaxAbs">Pointer to the output result.</param>
			//    /// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMaxAbs_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMaxAbs, CUdeviceptr pDeviceBuffer);

			//    /// <summary>
			//    /// Device scratch buffer size (in bytes) for nppsMaxAbsIndx_16s.
			//    /// </summary>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="hpBufferSize">Required buffer size.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMaxAbsIndxGetBufferSize_16s(int nLength, ref int hpBufferSize);

			//    /// <summary>
			//    /// Device scratch buffer size (in bytes) for nppsMaxAbsIndx_32s.
			//    /// </summary>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="hpBufferSize">Required buffer size.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMaxAbsIndxGetBufferSize_32s(int nLength, ref int hpBufferSize);

			//    /// <summary>
			//    /// 16-bit integer vector max absolute index method
			//    /// </summary>
			//    /// <param name="pSrc">Source signal pointer.</param>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="pMaxAbs">Pointer to the output result.</param>
			//    /// <param name="pIndx">Pointer to the index value of the first maximum element.</param>
			//    /// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMaxAbsIndx_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMaxAbs, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);

			//    /// <summary>
			//    /// 32-bit integer vector max absolute index method
			//    /// </summary>
			//    /// <param name="pSrc">Source signal pointer.</param>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="pMaxAbs">Pointer to the output result.</param>
			//    /// <param name="pIndx">Pointer to the index value of the first maximum element.</param>
			//    /// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMaxAbsIndx_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMaxAbs, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);

			//    /// <summary>
			//    /// Device scratch buffer size (in bytes) for nppsMinAbs_16s.
			//    /// </summary>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="hpBufferSize">Required buffer size.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMinAbsGetBufferSize_16s(int nLength, ref int hpBufferSize);

			//    /// <summary>
			//    /// Device scratch buffer size (in bytes) for nppsMinAbs_32s.
			//    /// </summary>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="hpBufferSize">Required buffer size.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMinAbsGetBufferSize_32s(int nLength, ref int hpBufferSize);

			//    /// <summary>
			//    /// 16-bit integer vector min absolute method
			//    /// </summary>
			//    /// <param name="pSrc">Source signal pointer.</param>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="pMinAbs">Pointer to the output result.</param>
			//    /// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMinAbs_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMinAbs, CUdeviceptr pDeviceBuffer);

			//    /// <summary>
			//    /// 32-bit integer vector min absolute method
			//    /// </summary>
			//    /// <param name="pSrc">Source signal pointer.</param>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="pMinAbs">Pointer to the output result.</param>
			//    /// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMinAbs_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMinAbs, CUdeviceptr pDeviceBuffer);

			//    /// <summary>
			//    /// Device scratch buffer size (in bytes) for nppsMinAbsIndx_16s.
			//    /// </summary>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="hpBufferSize">Required buffer size.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMinAbsIndxGetBufferSize_16s(int nLength, ref int hpBufferSize);

			//    /// <summary>
			//    /// Device scratch buffer size (in bytes) for nppsMinAbsIndx_32s.
			//    /// </summary>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="hpBufferSize">Required buffer size.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMinAbsIndxGetBufferSize_32s(int nLength, ref int hpBufferSize);

			//    /// <summary>
			//    /// 16-bit integer vector min absolute index method
			//    /// </summary>
			//    /// <param name="pSrc">Source signal pointer.</param>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="pMinAbs">Pointer to the output result.</param>
			//    /// <param name="pIndx">Pointer to the index value of the first minimum element.</param>
			//    /// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMinAbsIndx_16s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMinAbs, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);

			//    /// <summary>
			//    /// 32-bit integer vector min absolute index method
			//    /// </summary>
			//    /// <param name="pSrc">Source signal pointer.</param>
			//    /// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
			//    /// <param name="pMinAbs">Pointer to the output result.</param>
			//    /// <param name="pIndx">Pointer to the index value of the first minimum element.</param>
			//    /// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
			//    /// <returns></returns>
			//    [DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
			//    public static extern NppStatus nppsMinAbsIndx_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pMinAbs, CUdeviceptr pIndx, CUdeviceptr pDeviceBuffer);


			//}

			/// <summary>
			/// Mean and StdDev
			/// </summary>
			public static class MeanStdDev
			{
				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMean_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanGetBufferSize_32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMean_32fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanGetBufferSize_32fc(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMean_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanGetBufferSize_64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMean_64fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanGetBufferSize_64fc(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMean_16s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanGetBufferSize_16s_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMean_32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanGetBufferSize_32s_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMean_16sc_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanGetBufferSize_16sc_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 32-bit float vector mean method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMean_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit float complex vector mean method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMean_32fc(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit double vector mean method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMean_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit double complex vector mean method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMean_64fc(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit short vector mean with integer scaling method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMean_16s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 32-bit integer vector mean with integer scaling method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMean_32s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit short complex vector mean with integer scaling method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMean_16sc_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, int nScaleFactor, CUdeviceptr pDeviceBuffer);






				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsStdDev_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsStdDevGetBufferSize_32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsStdDev_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsStdDevGetBufferSize_64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsStdDev_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsStdDevGetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsStdDev_16s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsStdDevGetBufferSize_16s_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 32-bit float vector standard deviation method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pStdDev">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsStdDev_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pStdDev, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit float vector standard deviation method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pStdDev">Pointer to the output result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsStdDev_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pStdDev, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit float vector standard deviation method (return value is 32-bit)
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pStdDev">Pointer to the output result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsStdDev_16s32s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pStdDev, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit float vector standard deviation method (return value is also 16-bit)
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pStdDev">Pointer to the output result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsStdDev_16s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pStdDev, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMeanStdDev_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanStdDevGetBufferSize_32f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMeanStdDev_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanStdDevGetBufferSize_64f(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMeanStdDev_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanStdDevGetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// Device scratch buffer size (in bytes) for nppsMeanStdDev_16s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size. Important: hpBufferSize is a </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanStdDevGetBufferSize_16s_Sfs(int nLength, ref int hpBufferSize /* host pointer */);


				/// <summary>
				/// 32-bit float vector mean and standard deviation method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output mean value.</param>
				/// <param name="pStdDev">Pointer to the output standard deviation value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanStdDev_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, CUdeviceptr pStdDev, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 64-bit float vector mean and standard deviation method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output mean value.</param>
				/// <param name="pStdDev">Pointer to the output standard deviation value.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanStdDev_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, CUdeviceptr pStdDev, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit float vector mean and standard deviation method (return values are 32-bit)
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output mean value.</param>
				/// <param name="pStdDev">Pointer to the output standard deviation value.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanStdDev_16s32s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, CUdeviceptr pStdDev, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// 16-bit float vector mean and standard deviation method (return values are also 16-bit)
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pMean">Pointer to the output mean value.</param>
				/// <param name="pStdDev">Pointer to the output standard deviation value.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsMeanStdDev_16s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pMean, CUdeviceptr pStdDev, int nScaleFactor, CUdeviceptr pDeviceBuffer);




			}

			/// <summary>
			/// Infinity Norm, L1 Norm, L2 Norm
			/// </summary>
			public static class Norm
			{
				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_Inf_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormInfGetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float vector C norm method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_Inf_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_Inf_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormInfGetBufferSize_64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float vector C norm method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_Inf_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_Inf_16s32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormInfGetBufferSize_16s32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer vector C norm method, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_Inf_16s32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_Inf_32fc32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormInfGetBufferSize_32fc32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float complex vector C norm method, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_Inf_32fc32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_Inf_64fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormInfGetBufferSize_64fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float complex vector C norm method, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_Inf_64fc64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_Inf_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormInfGetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer vector C norm method, return value is 32-bit signed integer.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_Inf_16s32s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L1_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL1GetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float vector L1 norm method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L1_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L1_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL1GetBufferSize_64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float vector L1 norm method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L1_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L1_16s32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL1GetBufferSize_16s32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer vector L1 norm method, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the L1 norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L1_16s32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L1_32fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL1GetBufferSize_32fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float complex vector L1 norm method, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L1_32fc64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L1_64fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL1GetBufferSize_64fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float complex vector L1 norm method, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L1_64fc64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L1_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL1GetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer vector L1 norm method, return value is 32-bit signed integer.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L1_16s32s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L1_16s64s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL1GetBufferSize_16s64s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer vector L1 norm method, return value is 64-bit signed integer.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L1_16s64s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);



				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L2_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL2GetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float vector L2 norm method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L2_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L2_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL2GetBufferSize_64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float vector L2 norm method
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L2_64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L2_16s32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL2GetBufferSize_16s32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer vector L2 norm method, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L2_16s32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L2_32fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL2GetBufferSize_32fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float complex vector L2 norm method, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L2_32fc64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L2_64fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL2GetBufferSize_64fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float complex vector L2 norm method, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L2_64fc64f(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L2_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL2GetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer vector L2 norm method, return value is 32-bit signed integer.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L2_16s32s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNorm_L2Sqr_16s64s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormL2SqrGetBufferSize_16s64s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer vector L2 Square norm method, return value is 64-bit signed integer.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNorm_L2Sqr_16s64s_Sfs(CUdeviceptr pSrc, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);



			}

			/// <summary>
			/// Infinity Norm Diff, L1 Norm Diff, L2 Norm Diff
			/// </summary>
			public static class NormDiff
			{
				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_Inf_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffInfGetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float C norm method on two vectors' difference
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_Inf_32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_Inf_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffInfGetBufferSize_64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float C norm method on two vectors' difference
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_Inf_64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_Inf_16s32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffInfGetBufferSize_16s32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer C norm method on two vectors' difference, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_Inf_16s32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_Inf_32fc32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffInfGetBufferSize_32fc32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float complex C norm method on two vectors' difference, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_Inf_32fc32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_Inf_64fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffInfGetBufferSize_64fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float complex C norm method on two vectors' difference, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_Inf_64fc64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_Inf_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffInfGetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer C norm method on two vectors' difference, return value is 32-bit signed integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_Inf_16s32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);

				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L1_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL1GetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float L1 norm method on two vectors' difference
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L1_32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L1_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL1GetBufferSize_64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float L1 norm method on two vectors' difference
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L1_64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L1_16s32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL1GetBufferSize_16s32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer L1 norm method on two vectors' difference, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the L1 norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L1_16s32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L1_32fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL1GetBufferSize_32fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float complex L1 norm method on two vectors' difference, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L1_32fc64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L1_64fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL1GetBufferSize_64fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float complex L1 norm method on two vectors' difference, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L1_64fc64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L1_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL1GetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer L1 norm method on two vectors' difference, return value is 32-bit signed integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer..</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L1_16s32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L1_16s64s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL1GetBufferSize_16s64s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer L1 norm method on two vectors' difference, return value is 64-bit signed integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L1_16s64s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L2_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL2GetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float L2 norm method on two vectors' difference
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L2_32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L2_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL2GetBufferSize_64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float L2 norm method on two vectors' difference
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L2_64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L2_16s32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL2GetBufferSize_16s32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer L2 norm method on two vectors' difference, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L2_16s32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L2_32fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL2GetBufferSize_32fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float complex L2 norm method on two vectors' difference, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L2_32fc64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L2_64fc64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL2GetBufferSize_64fc64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float complex L2 norm method on two vectors' difference, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L2_64fc64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L2_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL2GetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer L2 norm method on two vectors' difference, return value is 32-bit signed integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L2_16s32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsNormDiff_L2Sqr_16s64s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiffL2SqrGetBufferSize_16s64s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer L2 Square norm method on two vectors' difference, return value is 64-bit signed integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pNorm">Pointer to the norm result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsNormDiff_L2Sqr_16s64s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pNorm, int nScaleFactor, CUdeviceptr pDeviceBuffer);



			}

			/// <summary>
			/// Dot Product
			/// </summary>
			public static class DotProduct
			{
				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float dot product method, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_32fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_32fc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float complex dot product method, return value is 32-bit float complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_32f32fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_32f32fc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float and 32-bit float complex dot product method, return value is 32-bit float complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_32f32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_32f64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_32f64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float dot product method, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_32f64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_32fc64fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_32fc64fc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float complex dot product method, return value is 64-bit float complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_32fc64fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_32f32fc64fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_32f32fc64fc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit float and 32-bit float complex dot product method, return value is 64-bit float complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_32f32fc64fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_64f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_64f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float dot product method, return value is 64-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_64f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_64fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_64fc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float complex dot product method, return value is 64-bit float complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_64fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_64f64fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_64f64fc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 64-bit float and 64-bit float complex dot product method, return value is 64-bit float complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_64f64fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16s64s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16s64s(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer dot product method, return value is 64-bit signed integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16s64s(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16sc64sc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16sc64sc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer complex dot product method, return value is 64-bit signed integer complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16sc64sc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16s16sc64sc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16s16sc64sc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer and 16-bit signed short integer short dot product method, return value is 64-bit signed integer complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16s16sc64sc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16s32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16s32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer dot product method, return value is 32-bit float.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16s32f(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16sc32fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16sc32fc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer complex dot product method, return value is 32-bit float complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16sc32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16s16sc32fc.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16s16sc32fc(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer and 16-bit signed short integer complex dot product method, return value is 32-bit float complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16s16sc32fc(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer dot product method, return value is 16-bit signed short integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16sc_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16sc_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer complex dot product method, return value is 16-bit signed short integer complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_32s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit signed integer dot product method, return value is 32-bit signed integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_32sc_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_32sc_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit signed integer complex dot product method, return value is 32-bit signed integer complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_32sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16s32s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer dot product method, return value is 32-bit signed integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result. </param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16s32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16s16sc32sc_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16s16sc32sc_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer and 16-bit signed short integer complex dot product method, return value is 32-bit signed integer complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16s16sc32sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16s32s32s_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16s32s32s_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer and 32-bit signed integer dot product method, return value is 32-bit signed integer.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16s32s32s_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16s16sc_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16s16sc_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer and 16-bit signed short integer complex dot product method, return value is 16-bit signed short integer complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16s16sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_16sc32sc_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_16sc32sc_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer complex dot product method, return value is 32-bit signed integer complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_16sc32sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsDotProd_32s32sc_Sfs.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProdGetBufferSize_32s32sc_Sfs(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit signed short integer and 32-bit signed short integer complex dot product method, return value is 32-bit signed integer complex.
				/// </summary>
				/// <param name="pSrc1">Source signal pointer.</param>
				/// <param name="pSrc2">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pDp">Pointer to the dot product result.</param>
				/// <param name="nScaleFactor">Integer Result Scaling.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsDotProd_32s32sc_Sfs(CUdeviceptr pSrc1, CUdeviceptr pSrc2, int nLength, CUdeviceptr pDp, int nScaleFactor, CUdeviceptr pDeviceBuffer);



			}

			/// <summary>
			/// Count In Range
			/// </summary>
			public static class CountInRange
			{
				/// <summary>
				/// Device-buffer size (in bytes) for nppsCountInRange_32s.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="hpBufferSize">Required buffer size.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCountInRangeGetBufferSize_32s(int nLength, ref int hpBufferSize);

				/// <summary>
				/// Computes the number of elements whose values fall into the specified range on a 32-bit signed integer array.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed.</param>
				/// <param name="pCounts">Pointer to the number of elements.</param>
				/// <param name="nLowerBound">Lower bound of the specified range.</param>
				/// <param name="nUpperBound">Upper bound of the specified range.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation.</param>
				/// <returns></returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsCountInRange_32s(CUdeviceptr pSrc, int nLength, CUdeviceptr pCounts, int nLowerBound, int nUpperBound, CUdeviceptr pDeviceBuffer);
			}

			/// <summary>
			/// Count Zero Crossings
			/// </summary>
			public static class ZeroCrossing
			{
				/// <summary>
				/// Device-buffer size (in bytes) for nppsZeroCrossing_16s32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZeroCrossingGetBufferSize_16s32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 16-bit signed short integer zero crossing method, return value is 32-bit floating point.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pValZC">Pointer to the output result.</param>
				/// <param name="tZCType">Type of the zero crossing measure: nppZCR, nppZCXor or nppZCC.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZeroCrossing_16s32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pValZC, NppsZCType tZCType, CUdeviceptr pDeviceBuffer);


				/// <summary>
				/// Device-buffer size (in bytes) for nppsZeroCrossing_32f.
				/// </summary>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="hpBufferSize">Required buffer size.  Important: </param>
				/// <returns>
				/// NPP_SUCCESS
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZeroCrossingGetBufferSize_32f(int nLength, ref int hpBufferSize);


				/// <summary>
				/// 32-bit floating-point zero crossing method, return value is 32-bit floating point.
				/// </summary>
				/// <param name="pSrc">Source signal pointer.</param>
				/// <param name="nLength">Signal length. Tells the primitive how many of the signal’s samples starting from the given data pointer are to be processed..</param>
				/// <param name="pValZC">Pointer to the output result.</param>
				/// <param name="tZCType">Type of the zero crossing measure: nppZCR, nppZCXor or nppZCC.</param>
				/// <param name="pDeviceBuffer">Pointer to the required device memory allocation, Scratch Buffer and Host Pointer. </param>
				/// <returns>
				/// <see cref="NppStatus.NullPointerError"/>, <see cref="NppStatus.AlignmentError"/>, <see cref="NppStatus.SizeError"/>.
				/// </returns>
				[DllImport(NPPS_API_DLL_NAME, CallingConvention = CallingConvention.Cdecl)]
				public static extern NppStatus nppsZeroCrossing_32f(CUdeviceptr pSrc, int nLength, CUdeviceptr pValZC, NppsZCType tZCType, CUdeviceptr pDeviceBuffer);



			}
		}
	}
}
